#!/usr/bin/env python
# vim: set tabstop=8 softtabstop=8 shiftwidth=8 noexpandtab
# Version 1.0  April-28-2020
#############################################################
def is_float(float_str):
	# Developer: Maosheng Zhao, 2/26/2020
	# return False or True
	TF = False
	try:
		flt = float(float_str)
		TF = True
	except ValueError:
		TF = False

	return TF
#############################################################

def is_int(int_str):
	# Developer: Maosheng Zhao, 2/26/2020
	# return False or True
	TF = False
	try:
		int_n = int(int_str)
		TF = True
	except ValueError:
		TF = False

	return TF
#############################################################

def is_exe_file(program):
	# Developer: Maosheng Zhao, 2/18/2020
	# check if the executable file exist
	# return None if program doesn't exists or it is not executible
	# return executable file name if it is executible
	import os

	fpath, fname = os.path.split(program)

	if fpath == '' and (os.path.isfile(fname) and os.access(fname, os.X_OK)):
		return program     # may also fit the else case below
	elif fpath == '.' and (os.path.isfile(fname) and os.access(fname, os.X_OK)):
		return program
	else:    # full path file name for the executable file
		path=os.getenv('PATH')
		yes = False
		for p in path.split(os.path.pathsep):
			exe_file=os.path.join(p, program)
			if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
				yes = True
				return exe_file
		if not yes:
			return None
#############################################################

def run_command(command_line):
	# Developer: Maosheng Zhao, 2/18/2020
	# please note that "The run() function was added in Python 3.5"
	# cmd_return is Exit status of the child process. Typically, an exit status of 0
	#  indicates that it ran successfully.
	import subprocess
	import os

	# get the command filename
	if " " not in command_line:
		cmd = command_line
	else:
		cmd = command_line.split(" ")[0]

	# run the command string
	stdout_str = None
	stderr_str = None

	try:
		cmd_return = subprocess.run(command_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,)
	except subprocess.CalledProcessError as err:
		print('ERROR:', err)
	else:
		#print('returncode for %s:%d' % (cmd, cmd_return.returncode))
    		#print('Have {} bytes in stdout: {!r}'.format(len(cmd_return.stdout), cmd_return.stdout.decode('utf-8')))
    		#print('Have {} bytes in stderr: {!r}'.format(len(cmd_return.stderr), cmd_return.stderr.decode('utf-8')))
		stdout_str = cmd_return.stdout.decode('utf-8')
		stderr_str = cmd_return.stderr.decode('utf-8')

	return cmd_return, stdout_str, stderr_str
#############################################################

def is_HDF_NC_fmt(HDF_file):
	# Developer: Maosheng Zhao, 2/26/2020
	# check if the input HDF file an HDF4/HDF5 or NetCDF4
	# the key tool used here to do the work is read_sds_attributes
	# it requires function run_command()
	import os

	# is the command files executable files?
	command = "read_sds_attributes"
	exe_fname = is_exe_file(command)

	exe_cmd = None
	if exe_fname is not None:
		exe_cmd = command

	TF = False

	if exe_cmd is None:   # exit or abort
		print("There is no executable file read_sds_attributes, make sure you put it in a path defined in you Shell PATH. Exit...")
		sys.exit(0)
	else:  # run the command
		# check if the HDF_file exists
		if not (os.path.isfile(HDF_file) and os.access(HDF_file, os.R_OK)):
			print("input file %s does not exists or is not readable, exit..." % HDF_file)
			sys.exit(0)
		else:
			# make command line
			cmd_line_list = [command, "-h", HDF_file]
			cmd_line  = " ".join(cmd_line_list)
			# run and get the results.
			cmd_return, stdout_str, stderr_str = run_command(cmd_line)

			# if there is stderr_str and it says there is no StructMetadata.0, projected is False
			if type(stderr_str) == str:
				if stderr_str.find("is neither HDF4 nor HDF5, exit") > -1:
					TF = False
				elif stderr_str.find("Cannot open the HDF file") > -1:
					TF = False
				else:
					TF = True

	return TF
#############################################################

def get_ESDT_from_HDFname(HDF_filename):
	# Developer: Maosheng Zhao, 3/20/2020
	# added granuleID and file extension name ext_name on 4/10/2020
	# get the ESDT name from an HDF_filename
	import os
	import re

	ESDT     = None
	DatTime  = None
	ext_name = None
	granuleID= None

	HDFbname = os.path.basename(HDF_filename)
	# get the ESDT
	m = re.search('^(.+)\.A(\d{7})\.(.+)\.\d{3}\..+$', HDFbname)
	if m:
		ESDT     = m.group(1)
		DatTime  = m.group(2)
		granuleID= m.group(3)

	# get the extension name for the files with .hdf. .h5 or .nc
	m=re.search('\.(.+)', HDFbname[-4:])
	if m:
		ext_name = m.group(1)

	return ESDT, DatTime, ext_name, granuleID
#############################################################

def grid_or_swath_HDF(HDF_file):
	# Developer: Maosheng Zhao, 2/21/2020
	# determine if the HDF_file is a grid HDF (i.e., with map projection in StructMetadata.0)
	# or not (i.e., without map projection in StructMetadata.0 or no StructMetadata.0 at all)
	# Note that it requires Python functions of is_exe_file() and run_command() and is_HDF_NC_fmt()
	# The tool can deal with either HDF4/HDF5 or NetCDF4.
	# just return False or True
	import os
	import sys

	# initialize Boolean and case
	Boolean = False
	case = 0
	
	#STEP 1. is the command files executable files?
	command = "read_meta"
	exe_fname = is_exe_file(command)

	exe_cmd = None
	if exe_fname is not None:
		exe_cmd = command

	if exe_cmd is None:   # exit or abort
		print("There is no executable file read_meta, make sure you put it in a path defined in you Shell PATH. Exit...")
		sys.exit(0)
	else:  # run the command
		if not is_HDF_NC_fmt(HDF_file):
			print("input file %s does not exists or is not readable or is not HDF4/HDF5/NetCDF4 format, exit..." % HDF_file)
			sys.exit(0)
		else:
			# make command line
			cmd_line_list = [command, "-struct", HDF_file]
			cmd_line  = " ".join(cmd_line_list)

			# run and get the results.
			cmd_return, stdout_str, stderr_str = run_command(cmd_line)

			# if there is stderr_str and it says there is no StructMetadata.0, projected is False
			if type(stderr_str) == str:
				if stderr_str.find("There is no metadata StructMetadata.0") != -1:
					case = 0
					Boolean = False

			if type(stdout_str) == str:
				# There is StructMetadata but if there is map projection information?
				if stdout_str.find("Projection=") > 0:
					Boolean = True
					case = 2
				else:
					Boolean = False
					case = 1
	# case 0: No StructMetadata.0
	# case 1: with StructMetadata.0 but no map projection
	# case 2: with StructMetadata.0 and has defined map projection
	return case
#############################################################

def DDDMMMSSS_to_Decimal_DEG(HDF_DMS_npflt64):
	# Developer: Maosheng Zhao, 3/18/2020
	# convert HDFEOS DDDMMMSSS.SS to Decimal Degree
	#   of the Geographic map projection
	import numpy as np

	int_deg = np.float64(np.int32(np.floor(HDF_DMS_npflt64*1.e-6 + 0.5)))
	fraction_min= np.float64((np.float64(np.int32(np.floor(HDF_DMS_npflt64*1.e-3+0.5))) - int_deg*1.e+3)/60.)
	fraction_sec= np.float64((HDF_DMS_npflt64 - np.float64(np.int32(np.floor(HDF_DMS_npflt64*1.e-3+0.5)))*1.e+3)/(60.*60.))

	Angle_DDG = int_deg+fraction_min+fraction_sec

	# np.float64 data type
	return Angle_DDG
#############################################################

def parse_struct_metadata(HDF_with_struct_metadata):
	# Developer: Maosheng Zhao, 3/18/2020
	# Purpose: parse struct metadata in HDF4/HDF5 files
	# I didn't use some HDFEOS APIs because QA Tools
	# sometimes copy structmetadata to an HDF file and thus
	# the HDF with the structmetadata may not be an HDFEOS file
	# requires function DDDMMMSSS_to_Decimal_DEG()
	import re
	import numpy as np

	# number, list tuple, list tuple, list, list, list of list
	num_grids = 0
	upleft_coor = []
	lowright_coor = []
	proj_geo =[]
	field_names_num=[]
	field_names =[]

	metadata = ''
	if grid_or_swath_HDF(HDF_with_struct_metadata) == 2:
		# make command line. Here I assume read_meta exists otherwise,
		# grid_or_swath_HDF() will report error and exist
		command = 'read_meta'
		cmd_line_list = [command, "-struct", HDF_with_struct_metadata]
		cmd_line  = " ".join(cmd_line_list)

		# run and get the results.
		_, metadata, _ = run_command(cmd_line)
	else:
		metadata = None

	if type(metadata) == str:
		# step 1, split the grid
		num_grids = metadata.count('END_GROUP=GRID_')

		if num_grids > 0:
			grids_str_list=[]
			for i in range(num_grids):
				grid_start_str = "GROUP=GRID_" + str(i+1)
				grid_end_str   = "END_GROUP=GRID_" + str(i+1)
				start_loc      = metadata.find(grid_start_str)
				end_loc        = metadata.find(grid_end_str) + len(grid_end_str)
				grids_str_list.append(metadata[start_loc : end_loc])

		# step 2, get the upleft and lowright of the grid
		#         and map projection (Geographic or non-geographic)
		if num_grids > 0:
			for grid in grids_str_list:
				start_s       = "UpperLeftPointMtrs=("
				tmp_loc       = grid.find(start_s)+len(start_s)
				upleft_str    = grid[tmp_loc : grid[tmp_loc :].find(')') + tmp_loc]
				upleft_tmp    = list(map(np.float64, upleft_str.split(',')))

				start_s       = "LowerRightMtrs=("
				tmp_loc       = grid.find(start_s)+len(start_s)
				lowright_str  = grid[tmp_loc : grid[tmp_loc :].find(')') + tmp_loc]
				lowright_tmp  = list(map(np.float64, lowright_str.split(',')))

				# convert DDDMMMSSS.SS to decimal degree if the map projection is Geographic
				proj_code = 0   # non_geographic
				if re.search('GCTP_GEO', grid):
					upleft_tmp   = list(map(DDDMMMSSS_to_Decimal_DEG, upleft_tmp))
					lowright_tmp = list(map(DDDMMMSSS_to_Decimal_DEG, lowright_tmp))
					proj_code = 1   # geographic
				
				upleft_coor.append(tuple(upleft_tmp))
				lowright_coor.append(tuple(lowright_tmp))
				proj_geo.append(proj_code)

				# get number of data fields
				num_fields = grid.count('DataFieldName="')
				field_names_num.append(num_fields)

				# get the data field names
				start_s    = 'DataFieldName="'
				tmp_list=[]
				shift_loc = 0
				start_loc = 0
				tmp_beg_loc   = 0
				while tmp_beg_loc != -1:
					tmp_beg_loc = grid[shift_loc :].find(start_s) 
					start_loc   = shift_loc + tmp_beg_loc + len(start_s)
					tmp_end_loc = grid[start_loc :].find('"')
					end_loc     = start_loc + tmp_end_loc
					if tmp_beg_loc != -1:
						field_name = grid[start_loc : end_loc]
						tmp_list.append(field_name)
						# update shift_loc
						shift_loc = end_loc
					else:
						break	

				field_names.append(tmp_list)
	else:
		num_grids       = 0
		upleft_coor     = None
		lowright_coor   = None
		proj_geo        = None
		field_names_num = None
		field_names     = None

	# number, list tuple, list tuple, list, list, list of list
	return num_grids, upleft_coor, lowright_coor, proj_geo, field_names_num, field_names
#############################################################

def get_native_pixsz_grid(SDS_shape, Proj_Geo, upleft_coor, lowright_coor):
	# Developer: Maosheng Zhao, 3/19/2020
	# Get the native spatial resolution for a specified
	# SDS_shape from a gridded HDFEOS, its upleft_coor
	# & lowright_coor, and if it is Geographic map
	# projection. Note that Proj_Geo (0 or 1) refers
	# to non-geographic or geographic map projection
	# with geometrics of upleft_coor and lowright_coor. 
	# Return resolution in KM
	import numpy as np
	# Step 1:  get X and Y dimensions from the SDS_shape
	Xdim = 0
	Ydim = 0
	if len(SDS_shape) <= 1:
		print("This is for >=2D SDS not 1D, exit...")
		sys.exit(0)
	elif len(SDS_shape) == 2:
		Ydim = SDS_shape[0]
		Xdim = SDS_shape[1]
	else:
		if SDS_shape[0] > SDS_shape[-1]:  # BIP
			Ydim = SDS_shape[0]
			Xdim = SDS_shape[1]
		else:                             # BSQ
			Ydim = SDS_shape[-2]
			Xdim = SDS_shape[-1]

	# Step 2: get the X and Y dimensional distance
	X_dis = lowright_coor[0] - upleft_coor[0]
	Y_dis = upleft_coor[1]   - lowright_coor[1]

	# Step 3: get the resolution of X and Y dimensions
	X_res = X_dis / np.float64(Xdim)
	Y_res = Y_dis / np.float64(Ydim)

	# Step 4: get the relative differeces between X_res and Y_res
	relative_diff_percent = 100. * np.absolute(X_res - Y_res) / min(X_res, Y_res)
	if relative_diff_percent > 1:  # >%1 difference 
		print("The spatial resolution between X and Y dimensions are greater than 1%")

	# Step 5: get the pixesz in KM
	pixesz = 0.5 * (X_res + Y_res)

	# convert decimal degree to distance if it is Geographic map projection
	if Proj_Geo == 1:
		PI = 3.14159265359
		D2R = PI / 180.
		R_EARTH = 6371007.181000

		X_res = X_res * D2R * R_EARTH
		Y_res = Y_res * D2R * R_EARTH

	# Step 6: convert meters to KM
	X_res *= 0.001
	Y_res *= 0.001

	# Step 7: pixsz in KM
	pixsz = 0.5 * (X_res + Y_res)

	# pixsz in KM and relative_diff_percent for X and Y dim
	return pixsz, relative_diff_percent
#############################################################

def get_HDF_SDS_info(HDF_file):
	# Developer: Maosheng Zhao, 2/20/2020
	#  3/17/2020, added output tuple shape in the list and HDF version.
	#  3/22/2020, added further clean SDS list and shape by assuming that
	#            valid SDS should have same ratio of Y and X dims to the SDS
	#            with maximum Ydim * Xdim
	#  3/27/2020, added data_type output or return in case users need to specify fillvalue
	# get the list of "true" SDS names with >1D (or >=2D) data using LDOPE tool 
	# read_sds_attributes -h. It also output the shape of each SDSs and HDF version
	# The tool read_sds_attributes works for either HDF4/HDF5 but just
	# list true SDS names not the name with group path for the SDSs. It
	# also get the corresponding shape of the SDS
	# Note that it requires Python functions of is_exe_file() and run_command()
	#     and is_HDF_NC_fmt()

	# Note that just list 
	import os
	import sys
	import re
	import numpy as np

	SDS_list   = []
	shape_list = []
	data_type  = []
	HDF_ver    = None

	#STEP 1. is the command files executable files?
	command = "read_sds_attributes"
	exe_fname = is_exe_file(command)

	exe_cmd = None
	if exe_fname is not None:
		exe_cmd = command

	if exe_cmd is None:   # exit or abort
		print("There is no executable file read_sds_attributes, make sure you put it in a path defined in you Shell PATH. Exit...")
		sys.exit(0)
	else:  # run the command
		if not is_HDF_NC_fmt(HDF_file):
			print("input file %s does not exists or is not readable or is not HDF4/HDF5/NetCDF4 format, exit..." % HDF_file)
			sys.exit(0)
		else:
			# make command line
			cmd_line_list = [command, "-h", HDF_file]
			cmd_line  = " ".join(cmd_line_list)
			# run and get the results.
			cmd_return, stdout_str, stderr_str = run_command(cmd_line)

			# since I know that the output from read_sds_attributes is in stderr_str
			if stderr_str is None:
				print("Error, the output from read_sds_attributes should be in stderr, exit...")
				sys.exit(0)
			else:
				# get the valid SDS_list and shape_list
				key_output_str = "Valid SDS names, dimension and data type"
				if stderr_str.find(key_output_str) == 0:
					# get HDF version
					if re.search('in HDF4 file', stderr_str):
						HDF_ver = 4
					elif re.search('in HDF5 file', stderr_str):
						HDF_ver = 5
					else:
						HDF_ver = 0
					
					# get SDS list and shape	
					p_beg = stderr_str.find(key_output_str) + len(key_output_str) + len(HDF_file) + 1
					SDS_lines = stderr_str[p_beg :].split('\n')
					# remove the empty element. I remeber someone online said not using filter
					SDS_lines = [i for i in SDS_lines if i]
					# remove the '\t' at the beginning of each element
					SDS_lines = [x.strip("\t") for x in SDS_lines]
					# remove the leading and training spaces
					SDS_lines = [x.strip() for x in SDS_lines]
					# just get the SDS name from each element
					SDS_list_temp   = [x[:x.find(' (')] for x in SDS_lines]
					shape_list_temp = [x[x.find('(') : x.find(') ') + 1] for x in SDS_lines]
					data_type_temp  = [x[x.find(') ')+2 :] for x in SDS_lines]

					# make a dictionary for the SDS_shape for cleaning 1D SDSs
					SDS_shape_dict = dict(zip( SDS_list_temp, shape_list_temp ))
					data_type_dict = dict(zip( SDS_list_temp, data_type_temp ))

					# JUST KEEP SDS names with >=2D data not 1D !!!!!
					#newDict = dict()
					SDS_list  = []
					SDS_shape = []
					data_type = []
					for key in sorted(SDS_shape_dict.keys()):
						# just keep these SDSs with >1D dimensions
						# because the format of shape is like (ZdimxYdimxXdim)
						# for 1D data it is just like (dim) 
						if re.search('x', SDS_shape_dict[key]):
							#newDict[key] = SDS_shape_dict[key]
							SDS_list.append(key)
							shape_str = SDS_shape_dict[key]
							shape_list = list(map(int, shape_str[1:-1].split(' x ')))
							shape_tuple = tuple(shape_list)
							SDS_shape.append(shape_tuple)

							# data_type_dict same the same key as SDS_shape_dict
							data_type.append(data_type_dict[key])

					# further clean valid SDSs by assuming the finest SDSs are valid
					# others valid should have same ratios of X and Y dimensions

					# clean the SDSs get again
					SDS_list_new  = []
					SDS_shape_new = []
					data_type_new = []
					if len(SDS_shape) > 0:
						# get the finest SDS
						data_len_max = 0
						Ydim = 0
						Xdim = 0
						Ydim_max = 0
						Xdim_max = 0
						for shape in SDS_shape: # all shape >=2D
							if len(shape) == 2:
								Ydim = shape[0]
								Xdim = shape[1]
							else:
								if shape[0] > shape[-1]:  # BIP
									Ydim = shape[0]
									Xdim = shape[1]
								else:                     # BSQ
									Ydim = shape[-2]
									Xdim = shape[-1]

							# note here that Ydim_max and Xdim_max may
							# not be the real maxima but their multiplication
							if (Ydim * Xdim) > data_len_max:
								data_len_max = Ydim * Xdim
								Ydim_max = Ydim
								Xdim_max = Xdim

						Y_ratio = 0.
						X_ratio = 0.
						diff_percent = 100.
						for idx, shape in enumerate(SDS_shape): # all shape >=2D
							if len(shape) == 2:
								Ydim = shape[0]
								Xdim = shape[1]
							else:
								if shape[0] > shape[-1]:  # BIP
									Ydim = shape[0]
									Xdim = shape[1]
								else:                     # BSQ
									Ydim = shape[-2]
									Xdim = shape[-1]

							Y_ratio = np.float64(Ydim_max) / np.float64(Ydim)
							X_ratio = np.float64(Xdim_max) / np.float64(Xdim)
							diff_percent = 100. * np.absolute(Y_ratio - X_ratio
									) / (0.5*(Y_ratio + X_ratio))    
							# if the Y_ratio and X_ratio are almost the same, SDSs are valid
							if diff_percent < 1.:
								SDS_list_new.append(SDS_list[idx])
								SDS_shape_new.append(SDS_shape[idx])
								data_type_new.append(data_type[idx])

						if len(SDS_list_new) < 1:
							SDS_list = None
							SDS_shape= None
							data_type= None
							HDF_ver  = None
						else:
							SDS_list = SDS_list_new
							SDS_shape= SDS_shape_new
							data_type= data_type_new
							HDF_ver  = HDF_ver

					else:
						SDS_list = None
						SDS_shape= None
						data_type= None
						HDF_ver  = None
				else:
					SDS_list = None
					SDS_shape= None
					data_type= None
					HDF_ver  = None

	return SDS_list, SDS_shape, data_type, HDF_ver   # None or lists and HDF version (4 or 5)
#############################################################

def get_HDF_SDS_info_exclude_lat_lon(HDF_file):
	# Developer: Maosheng Zhao, 3/17/2020
	# using get_HDF_SDS_info get >=2D datasets' SDS_name,
	# shapes and HDF version but excluding lat/lon SDSs
	# so that users choose the reasonable SDS names for
	# reprojection
	import re
	SDS_list, SDS_shape, SDS_dtype, HDF_ver = get_HDF_SDS_info(HDF_file)

	# exclude lat/lon SDSs
	SDS_list_new  = []
	SDS_shape_new = []
	SDS_dtype_new = []

	if (SDS_list is not None) and grid_or_swath_HDF(HDF_file) != 2:  # for swath only
		for idx, SDS_name in enumerate(SDS_list):
			if not (re.search("latitude", SDS_name, re.IGNORECASE) or 
				re.search("longitude", SDS_name, re.IGNORECASE)):
				SDS_list_new.append(SDS_list[idx])
				SDS_shape_new.append(SDS_shape[idx])
				SDS_dtype_new.append(SDS_dtype[idx])

		if len(SDS_list_new) == 0 or len(SDS_shape_new) == 0 or SDS_dtype_new == 0:
			SDS_list_new  = None
			SDS_shape_new = None
			SDS_dtype_new = None
	else:
		SDS_list_new = SDS_list 
		SDS_shape_new= SDS_shape
		SDS_dtype_new= SDS_dtype

	return SDS_list_new, SDS_shape_new, SDS_dtype_new, HDF_ver

#############################################################

def swath_has_latlon(HDF_file):
	# Developer: Maosheng Zhao, 2/25/2020
	# check if there are lat/lon SDSs in the HDF_file
	# check if latitude* longitude* exist in the SDS list
	# requires get_HDF_SDS_info(), 
	import sys
	import re

	lat_SDS_exist = False
	lon_SDS_exist = False

	SDS_list, _, _, _ = get_HDF_SDS_info(HDF_file)

	if SDS_list is not None and type(SDS_list) == list:
		for SDS_name in SDS_list:
			if re.search("latitude", SDS_name, re.IGNORECASE):
				lat_SDS_exist = True
			if re.search("longitude", SDS_name, re.IGNORECASE):
				lon_SDS_exist = True

			if lat_SDS_exist and lon_SDS_exist:
				break

	else:
		print("Error: empty scientific data in the %s, exit..." % HDF_file)
		sys.exit(0)

	return lat_SDS_exist, lon_SDS_exist
#############################################################

def get_LAT_LON_SDS_names(geo_HDF):
	# Developer: Maosheng Zhao, 3/15/2020
	# get lat/lon SDS names
	import sys
	import re

	lat_sds_name = None
	lon_sds_name = None

	SDS_list, _, _, _ = get_HDF_SDS_info(geo_HDF)

	if SDS_list is not None and type(SDS_list) == list:
		for SDS_name in SDS_list:
			if re.search("latitude", SDS_name, re.IGNORECASE):
				lat_sds_name = SDS_name
			if re.search("longitude", SDS_name, re.IGNORECASE):
				lon_sds_name = SDS_name

			if (lat_sds_name is not None) and (lon_sds_name is not None):
				break
	else:
		print("Error: empty scientific data in the %s, exit..." % geo_HDF)
		sys.exit(0)

	if (lat_sds_name is None) or (lon_sds_name is None):
		print("Error: no Latitude and/or Longitude in %s, exit", geo_HDF)
		sys.exit(0)

	return lat_sds_name, lon_sds_name
#############################################################

def SDS_name_valid(sds_name, HDF_file):
	# Developer: Maosheng Zhao, 3/15/2020
	# Enhanced on 4/21/2020 so that sds.n can be checked and also return the original sds_name_orig
	# if the SDS_name a valid for reprojection (>=2D)
	import sys
	import re

	name_valid    = False
	# for sds_name with sds.n for >=3D, sds_name_orig is just sds or the original sds name
	sds_name_orig = None

	SDS_list, SDS_shape, _, _ = get_HDF_SDS_info(HDF_file)

	if SDS_list is not None and type(SDS_list) == list:
		if sds_name in SDS_list:
			name_valid = True
		# added SDS.n case Apr. 21 2020
		else:  # for SDS.n case
			m = re.search('(.+)\.(\d+)', sds_name)
			if m:
				sds_name_orig = m.group(1)
				if sds_name_orig in SDS_list:
					name_valid = True

					plane_number = int(m.group(2))

					SDS_shape_dict = dict(zip(SDS_list, SDS_shape))
					shape = SDS_shape_dict[sds_name_orig]
					if len(shape) < 3:
						print("%s for the argument -sds= is for >=2D data, exit." % sds_name)
						sys.exit(0)
					else:  # currently just for 3D
						if shape[0] > shape[-1]:  # BIP
							Ydim = shape[0]
							Xdim = shape[1]
							Zdim = shape[2]
						else:                     # BSQ
							Zdim = shape[-3]
							Ydim = shape[-2]
							Xdim = shape[-1]

						if plane_number > Zdim:
							print("The %d in %s for -sds= is beyond the range of the plane_number, exit." % (plane_number, sds_name))
							sys.exit(0)
	else:
		print("Error: empty scientific data in the %s, exit..." % HDF_file)
		sys.exit(0)

	return name_valid, sds_name_orig
#############################################################

def copy_geolocation(in_geo_HDF, lat_sds_name, lon_sds_name, in_HDF):
	# Developer: Maosheng Zhao, 2/25/2020
	# using copy_sds to copy the user-specified two SDSs of lat and lon in in_geo_HDF to in_HDF
	# Note that it requires 5 Python functions of is_exe_file() and run_command()
	#         get_HDF_SDS_info(), swath_has_latlon() and is_HDF_NC_fmt()
	import os
	import sys
	import re
	
	# STEP 1, check if in_geo_HDF is a readable file
	if not (os.path.isfile(in_geo_HDF) and os.access(in_geo_HDF, os.R_OK)):
		print("file %s does not exists or is not readable, exit..." % in_geo_HDF)
		sys.exit(0)

	if not (os.path.isfile(in_HDF) and os.access(in_HDF, os.W_OK)):
		print("file %s does not exists or is not writable, exit..." % in_HDF)
		sys.exit(0)

	# STEP 2, check the existance of lat_sds_name and lon_sds_name
	lat_exist = False
	lon_exist = False

	SDS_list, _, _, _ = get_HDF_SDS_info(in_geo_HDF)
	if SDS_list is not None and type(SDS_list) == list:
		if lat_sds_name in SDS_list:
			lat_exist = True
		if lon_sds_name in SDS_list:
			lon_exist = True
	else:
		print("Error: empty scientific data in the %s, exit..." % in_geo_HDF)
		sys.exit(0)

	# STEP 3, copy geolocation SDSs if both exist
	if lat_exist and lon_exist:
		command = "copy_sds"
		exe_fname = is_exe_file(command)

		exe_cmd = None
		if exe_fname is not None:
			exe_cmd = command

		if exe_cmd is None:   # exit or abort
			print("There is no executable file copy_sds, make sure you put it in a path defined in you Shell PATH. Exit...")
			sys.exit(0)
		else:  # run the command
			# check if the HDF_file exists
			if not os.path.isfile(in_HDF):
				print("There is no HDF file named %s, exit..." % in_HDF)
				sys.exit(0)
			else:
				# make command line
				lat_sds = "-sds='" + lat_sds_name + ",Latitude'"
				lon_sds = "-sds='" + lon_sds_name + ",Longitude'"
				out_f   = "-of=" + in_HDF
				in_f    = in_geo_HDF

				cmd_line_list = [command, lon_sds, lat_sds, "-ncp", out_f, in_f]
				cmd_line  = " ".join(cmd_line_list)

				# run and get the results.
				print("copying geolocation data from %s to %s..." % (in_geo_HDF, in_HDF))
				#print(cmd_line)
				cmd_return, stdout_str, stderr_str = run_command(cmd_line)

				# confirm if copy_sds works fine
				lat_exist_new, lon_exist_new = swath_has_latlon(in_HDF)

				if not (lat_exist_new and lon_exist_new):
					print("Error in calling swath_has_latlon, information is below:")
					if type(stdout_str) == str:
						print(stdout_str)
					if type(stderr_str) == str:
						print(stderr_str)
				else:
					print("%s has geolocation SDSs now." % in_HDF)
				
	else:
		print("Error: specified wrong SDS name for geolocation latitude and longitude for %s, exit..." % in_geo_HDF)
		sys.exit(0)
#############################################################

def get_MODIS_pixsz_swath(SDS_shape):
	# Developer: Maosheng Zhao, 3/27/2020
	# get the default pixsz based on SDS_shape
	# of the MODIS swath data basd on exhaustive
	# attack method.
	Xdim = 0
	Ydim = 0
	if len(SDS_shape) <= 1:
		print("This is for >=2D SDS not 1D, exit...")
		sys.exit(0)
	elif len(SDS_shape) == 2:
		Ydim = SDS_shape[0]
		Xdim = SDS_shape[1]
	else:
		if SDS_shape[0] > SDS_shape[-1]:  # BIP
			Ydim = SDS_shape[0]
			Xdim = SDS_shape[1]
		else:                             # BSQ
			Ydim = SDS_shape[-2]
			Xdim = SDS_shape[-1]

	pixsz = None
	# pixsz in KM
	if 270 <= Xdim <= 271:
		pixsz = 5.
	elif Xdim == 1354:
		pixsz = 1.
	elif Xdim == 2708:
		pixsz = 0.5
	elif Xdim == 5416:
		pixsz = 0.25
	else:
		pixsz = None

	return pixsz, Ydim, Xdim
#############################################################

def get_VNP_pixsz_swath(SDS_shape):
	# Developer: Maosheng Zhao, 4/3/2020
	# get the default pixsz based on SDS_shape
	# of the VNP or NPP swath data basd on exhaustive
	# attack method.
	Xdim = 0
	Ydim = 0
	if len(SDS_shape) <= 1:
		print("This is for >=2D SDS not 1D, exit...")
		sys.exit(0)
	elif len(SDS_shape) == 2:
		Ydim = SDS_shape[0]
		Xdim = SDS_shape[1]
	else:
		if SDS_shape[0] > SDS_shape[-1]:  # BIP
			Ydim = SDS_shape[0]
			Xdim = SDS_shape[1]
		else:                             # BSQ
			Ydim = SDS_shape[-2]
			Xdim = SDS_shape[-1]

	pixsz = None
	# pixsz in KM
	if 400 <= Xdim <= 508:      # coarse res.
		pixsz = 5.
	elif 3200 <= Xdim <= 4064:  # DNB or MOD bands
		pixsz = 0.75
	elif Xdim == 6400:  # just 6464 for IMG but expand 3248 * 2
		pixsz = 0.375
	else:
		pixsz = None

	return pixsz, Ydim, Xdim
#############################################################

def get_VJ1_pixsz_swath(SDS_shape):
	# Developer: Maosheng Zhao, 4/3/2020
	# get the default pixsz based on SDS_shape
	# of the VNP or NPP swath data basd on exhaustive
	# attack method.
	Xdim = 0
	Ydim = 0
	if len(SDS_shape) <= 1:
		print("This is for >=2D SDS not 1D, exit...")
		sys.exit(0)
	elif len(SDS_shape) == 2:
		Ydim = SDS_shape[0]
		Xdim = SDS_shape[1]
	else:
		if SDS_shape[0] > SDS_shape[-1]:  # BIP
			Ydim = SDS_shape[0]
			Xdim = SDS_shape[1]
		else:                             # BSQ
			Ydim = SDS_shape[-2]
			Xdim = SDS_shape[-1]

	pixsz = None
	# pixsz in KM
	if  400 <= Xdim <= 508: 
		pixsz = 5.
	elif 3200 <= Xdim <= 4064:  # DNB or MOD bands
		pixsz = 0.75
	elif Xdim == 6400:  # for IMG bands
		pixsz = 0.375
	else:
		pixsz = None

	return pixsz, Ydim, Xdim
#############################################################

def make_directory(dir_val):
	# Developer Maosheng Zhao: 4/9/2020
	# Note that os.makedirs is sort of "mkdir -p path" which
	#           can recursively make path with multiple layers
	import os
	import errno
	if not os.path.isdir(dir_val):
		try:
			# mkdir -p path and make the path read/write by me, readable for everone else
			os.makedirs(dir_val, 0o755)
		except OSError as e:
			if e.errno != errno.EEXIST:
				print("Error: You have to specify a different temporary directory for -tmp_path=, see below for details, exit.")
				raise
#############################################################

def HDF5_to_HDF4(in_HDF5, SDS_name_list, tmp_path):
	# Developer: Maosheng Zhao, 3/20/2020
	# using copy_sds to convert HDF5 to HDF4
	import os
	import sys
	import re
	
	# STEP 1, check if in_HDF5 is a readable file
	if not (os.path.isfile(in_HDF5) and os.access(in_HDF5, os.R_OK)):
		print("file %s does not exists or is not readable, exit..." % in_HDF5)
		sys.exit(0)

	# STEP 2, get the out_HDF4 name by substitute the input extension name from h5/nc to hdf
	out_HDF4     = in_HDF5
	basename     = os.path.basename(in_HDF5)
	basename_new = re.sub('\.(h5|nc)$', '.hdf', basename)

	# STEP 3, temporary directory for storing converted DHF4s
	path_for_HDF4 = ''
	if tmp_path is not None:  # user specified temporay path
		if not os.path.isdir(tmp_path):
			make_directory(tmp_path)
			path_for_HDF4 = tmp_path
		else:
			if not os.access(tmp_path, os.W_OK):
				print("Error: %s is not writable for output converted HDF4 files, User must specify a different path temp directory for outputing the converted HDF4 files. User must specify a different path or the argument -tmp_path=. exit." % tmp_path)
				sys.exit(0)
			else:
				path_for_HDF4 = tmp_path
	else:  # make a temp path under the current working folder
		cwd = os.getcwd()
		if not os.access(cwd, os.W_OK):
			print("Error: current working path %s is not writable to make temp directory for outputing the converted HDF4 files. User must specify a path for the argument -tmp_path=. exit." % cwd)
			sys.exit(0)
		else:
			path_for_HDF4 = cwd + '/temp_dir'	
			if not os.path.isdir(path_for_HDF4):
				make_directory(path_for_HDF4)

	# full name for the out_HDF4		
	out_HDF4 = path_for_HDF4 + "/" + basename_new

	# remove old out_HDF4 if there is 
	if os.path.isfile(out_HDF4):
		os.remove(out_HDF4)

	if in_HDF5 == out_HDF4:
		print("input HDF5 %s has the same file name as the output HDF4 %s, exit..." % (in_HDF5, out_HDF4))
		sys.exit(0)

	# STEP 4, combine all valid SDSs for the argument into a list
	sds_args = []
	if type(SDS_name_list) == list and len(SDS_name_list):
		for sds_name in SDS_name_list:
			sds_name_right = sds_name + "," + sds_name
			sds_name_right = '"' + sds_name_right + '"'
			sds_arg = "-sds=" + sds_name_right
			sds_args.append(sds_arg)
	else:
		SDS_list, _, _, _ = get_HDF_SDS_info(in_HDF5)
		if SDS_list is not None and type(SDS_list) == list:
			for sds_name in SDS_list:
				sds_name_right = sds_name + "," + sds_name
				sds_name_right = '"' + sds_name_right + '"'
				sds_arg = "-sds=" + sds_name_right
				sds_args.append(sds_arg)
				lon_exist = True
		else:
			print("Error: empty scientific data in the %s, exit..." % in_HDF5)
			sys.exit(0)

	# STEP 5, make the full -sds argument
	sds_arg_full = " ".join(sds_args)

	# STEP 6, use copy_sds to convert DHF5 to HDF4
	command = "copy_sds"
	exe_fname = is_exe_file(command)

	exe_cmd = None
	if exe_fname is not None:
		exe_cmd = command

	if exe_cmd is None:   # exit or abort
		print("There is no executable file copy_sds, make sure you put it in a path defined in you Shell PATH. Exit...")
		sys.exit(0)
	else:  # run the command
		# make command line
		out_f = "-of=" + out_HDF4
		in_f = in_HDF5

		cmd_line_list = [command, sds_arg_full, "-ncp -meta", out_f, in_f]
		cmd_line  = " ".join(cmd_line_list)

		# run and get the results.
		print("converting HDF5 %s to HDF4 to %s..." % (in_HDF5, out_HDF4))
		#print(cmd_line)
		cmd_return, stdout_str, stderr_str = run_command(cmd_line)
		if type(cmd_return) == str:
			print(cmd_return)
		if type(stdout_str) == str:
			print(stdout_str)
		if type(stderr_str) == str:
			print(stderr_str)
	return out_HDF4
#############################################################

def get_HDFs_from_a_dir(dir_name):
	# developer: Maosheng Zhao, 4/10/2020
	# purpose: for the case that user just input the path for the -if=
	# when there are too many files to manually list tthem for -if= argument
	# Note that this is not recursive but just for the dir_name layer.
	# Note that the extended name must be hdf, h5 or nc.
	import os
	import sys
	import fnmatch
	HDF_files = None
	if os.path.isdir(dir_name):
		HDF_files = [ os.path.join(dir_name,f) for f in os.listdir(dir_name) if(
			fnmatch.fnmatch(f, '*.hdf') or fnmatch.fnmatch(f, '*.h5') or fnmatch.fnmatch(f, '*.nc')
			) ]
		HDF_files.sort()

		if len(HDF_files) == 0:
			print("Could not find HDF files (.hdf, .h5, or .nc) in the input path %s from -if=, exit." % dir_name)
			sys.exit(0)

	return HDF_files
#############################################################

def check_data_range(value_str, np_dtype_str):
	# developer: Maosheng Zhao, 4/20/2020
	# purpose: to check if a value within the limits of a given data type
	# return False or True
	import numpy as np

	TF = False

	if np_dtype_str == "float32":
		if np.finfo('float32').min <= float(value_str) <= np.finfo('float32').max:
			TF = True
	elif np_dtype_str == "uint32":
		if np.iinfo('uint32').min <= int(value_str) <= np.iinfo('uint32').max:
			TF = True
	elif np_dtype_str == "int32":
		if np.iinfo('int32').min <= int(value_str) <= np.iinfo('int32').max:
			TF = True
	elif np_dtype_str == "uint16":
		if np.iinfo('uint16').min <= int(value_str) <= np.iinfo('uint16').max:
			TF = True
	elif np_dtype_str == "int16":
		if np.iinfo('int16').min <= int(value_str) <= np.iinfo('int16').max:
			TF = True
	elif np_dtype_str == "uint8":
		if np.iinfo('uint8').min <= int(value_str) <= np.iinfo('uint8').max:
			TF = True
	elif np_dtype_str == "int8":
		if np.iinfo('int8').min <= int(value_str) <= np.iinfo('int8').max:
			TF = True
	else:
		print("Warning: check_data_range() doesn't check the range for data type of %s" % np_dtype_str)

	return TF

#############################################################
def print_help_str():
	# Developer: Maosheng Zhao, 3/20/2020
	help_str = '''
Usage: projbrowse.py -if=input_HDFs -sds=SDSs_name -of=output_HDF [-proj=] [-pixsz=] 
                     [-geoif=] [-tmp_path=] [-DN=] [-vector=1] [-fill=] [-global]
	'''
	print(help_str)
#############################################################

def print_all_help_str():
	# Developer: Maosheng Zhao, 3/20/2020
	help_str = '''
Usage: projbrowse.py -if=input_HDFs -sds=SDSs_name -of=output_HDF [-proj=] [-pixsz=]
                     [-geoif=] [-tmp_path=] [-DN=] [-vector=1] [-fill=] [-global]
	'''

	# define a long string
	# copied from projbrowse -help
	help_info = '''
NAME
     projbrowse.py - project data sets into a global projection

SYNOPSIS
     projbrowse.py -help
     projbrowse.py -if=input_HDFs -sds=SDSs_name -of=output_HDF [-proj=projection_code]
                   [-pixsz=pixsz_in_KM] [-geoif=geolocation_files] [-tmp_path=temp_dir]
                   [-DN=1,...,3] [-vector=1] 
                   [-fill=specified_fill_value] [-global]

    Note that argument orders can be different from the above SYNOPSIS.
		
DESCRIPTION
     Project a list of SDS into a geolocated projection. Various output
     projections are available. The output domain is either the actual
     bounding box of the complete input data set or the full globe. 
     Also coastlines can be plotted. The output is an HDF4 file.

OPTIONS
     -help                    Display this help message.
     -if                      Input HDF4/HDF5/NetCDF4 files separated in comma, or
                              a directory containing all input files.
     -sds=SDS[.no]            Specify the SDS names separated in comma to project.
                              Apply to all input files. For a 3D SDS, a .no suffix
                              can be used to specified the plane number to project
                              (default is 1).
     -of                      Specify the output file.
     -proj=no                 Specify the output projection (see below for details)
                              Optional, default is 16, Hammer-Aitoff.
     -pixsz=pixel_size        Set the output pixel size (in km), can be single or multple.
     -geoif=geolocation_file  Corresponding Geolocation files or their path for the input
                              swath HDFs without geolocation data fields, latitude and Longitude.
     -tmp_path=temp_dir       temporary path to store the coverted HDF4 files or new swath HDFs
                              with copied geolocation data fields, latitude and Longitude.
                              Optional, default is current working path ./temp_dir.
     -DN=1,...,3              Restrict the process to products that carry the specified values
                              of the Day/Night flag
                              Default is 1,2,3 (no restriction)
                              1 = Day
                              2 = Night
                              3 = Both (mixed)
                              Note that values of -DN can be have more than one value but they will
                              be applied to all SDSs in the -sds=.  This is different from -pixsz=
                              and -fill=
     -vector=1                Overlay any combination of geographic vectors. 1 = primary coastlines.
                              Only one value which will be applied to all SDSs from -sds=
     -fill=new_fill_value     Specify a different fill value than that for the data. For multiple
                              SDS names in -sds=, fill values can be one for all SDSs or multiple
                              different fill values separated by comma.
     -global                  Extend the output domain to the whole globe.
                              Default is extended box (actual limits of the specified products)

SUPPORTED PROJECTIONS
      1. Sinusoidal
      2. Integerized Sinusoidal
      3. Interrupted Sinusoidal USGS
      4. Interrupted Sinusoidal Variant
      5. Plate Carree
      6. Robinson
      7. Mollweide
      8. Interrupted Mollweide
      9. Goode Homolosine
     10. Goode Homolosine Variant
     11. Vertical Perspective
     12. Orthographic
     13. Gnomonic
     14. Stereographic
     15. Lambert Azimuthal
     16. Hammer-Aitoff
     17. Aitoff
     18. Eckert IV
     19. Azimuthal Equidistant
     20. Albers Conical Equal Area
     21. Regular Interrupted Sinusoidal
     22. Geographic

     WARNNG: only the Hammer-Aitoff (16) and Lambert-Azimuthal (15) are 
	     supported and the rest have not been fully tested.

Version 1.0  April-28-2020
	'''

	print(help_str)
	print(help_info)
############################################################

#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
#############################################################
#######################  MAIN PROGRAM  ######################
#############################################################
# start the main program
def _main(argv):
	import os
	import sys
	import glob  # for wildcard pattern match file names
	import re
	import numpy as np
	from copy import deepcopy

	# only for help case
	if len(argv) == 2 and ( argv[1] == '-h'    or
				argv[1] == '--h'   or
				argv[1] == '-help' or
				argv[1] == '--help'):
		print_all_help_str()
		sys.exit(0)

	# not for help case
	# key arguments
	arg_in_HDFs  = 0
	arg_out_HDF  = 0
	arg_sds      = 0
	arg_proj     = 0
	arg_pixsz    = 0
	arg_fill     = 0
	arg_tmp_path = 0
	arg_geo_HDFs = 0

	# values of key arguments
	in_HDFs      = None
	out_HDF      = None
	SDS_name     = None
	proj_code    = None
	pixsz        = None
	fills        = None
	tmp_path     = None
	geo_HDFs     = None

	# other arguments
	arg_DN       = 0
	arg_vector   = 0
	arg_global   = 0

	# values of other arguments (directly the argument)
	DN_val       = ""
	vector_val   = ""
	global_val   = ""

	# for HDF5_to_HDF4_temp_path
	HDF5_to_HDF4_temp_path = None

	# input file classes
	in_files_grid_swath = None
	in_files_HDF4_HDF5  = None

	if len(argv) < 2:
		print_all_help_str()
		sys.exit(0)
	else:  # get the input arguments and also check validality
		for arg in argv[1:]:
			# key arguments
			if re.search('-if=(.+)', arg, re.IGNORECASE):
				arg_in_HDFs = 1
				m = re.search('-if=(.+)', arg, re.IGNORECASE)
				if m:
					if_val = m.group(1)
					if os.path.isdir(if_val):
						in_HDFs = get_HDFs_from_a_dir(if_val)
					else:
						# if wildcard pattern match	
						in_HDFs = glob.glob(if_val)

						# else they are separated by comma
						if len(in_HDFs) == 0:
							in_HDFs = m.group(1)
							in_HDFs = in_HDFs.split(',')
							# remove empty elements
							in_HDFs = [x for x in in_HDFs if x != '']

							# remove leading or tailing white spaces for each file name if there are
							in_HDFs = list(map(lambda elem: elem.strip(), in_HDFs))

					# check the consistency in the input files
					esdt_0, time_0, ext_0, _ = get_ESDT_from_HDFname(in_HDFs[0])
					if type(esdt_0) != str:
						print("cannot get esdt name from the input file %s, exit" % in_HDFs[0])
						sys.exit(0)
					if type(time_0) != str:
						print("cannot get time from the input file %s, exit" % in_HDFs[0])
						sys.exit(0)
					if type(ext_0) != str:
						print("cannot get extension name from the input file %s, exit" % in_HDFs[0])
						sys.exit(0)
					for in_HDF in in_HDFs:
						esdt_tmp, time_tmp, ext_tmp, _ = get_ESDT_from_HDFname(in_HDF)
						if esdt_tmp != esdt_0:
							print("The input files have inconsistent ESDT names %s and %s, exit." % (esdt_tmp, esdt_0))
							sys.exit(0)
						if time_tmp != time_0:
							print("The input files have inconsistent time %s and %s, exit." % (time_tmp, time_0))
							sys.exit(0)
						if ext_tmp  != ext_0:
							print("The input files have inconsistent extension %s and %s, exit." % (ext_tmp, ext_0))
							sys.exit(0)
						if not (os.path.isfile(in_HDF) and os.access(in_HDF, os.R_OK)):
							print("%s doesn't exist or not readble, exit." % in_HDF)
							sys.exit(0)

					# check the SDSs in the input HDF	
					SDS_list, _, _, _ = get_HDF_SDS_info(in_HDFs[0])
					if SDS_list is None:
						print("Error: %s have no >=2D scientific data sets, exit." % in_HDFs[0])
						sys.exit(0)
				else:
					print("-if= has no value, exit.")
					sys.exit(0)
					
			elif re.search('-of=(.+)', arg, re.IGNORECASE):
				arg_out_HDF = 1
				
				m = re.search('-of=(.+)', arg, re.IGNORECASE)
				if m:
					out_HDF = m.group(1)
				else:
					print("-of= has no value, exit.")
					sys.exit(0)

				# check the validity of the output file
				o_dir = os.path.dirname(out_HDF)
				if o_dir == '':
					o_dir = os.getcwd()

				if not (os.path.isdir(o_dir) and os.access(o_dir, os.W_OK)):
					print("Error: The output path %s infered from -of= does not exist or is not writable, exit." % o_dir)
					sys.exit(0)

				# check if the extend name is hdf
				out_base_temp = os.path.basename(out_HDF)
				m = re.search('\.hdf$', out_base_temp)
				if not m:
					print("Error: The output file %s is not hdf, exit." % out_HDF)
					sys.exit(0)

				# remove the output file if it exists
				if os.path.isfile(out_HDF):
					os.remove(out_HDF)

			elif re.search('-sds=(.+)', arg, re.IGNORECASE):
				arg_sds = 1
				m = re.search('-sds=(.+)', arg, re.IGNORECASE)
				if m:
					SDS_names = m.group(1)
					SDS_names = SDS_names.split(',')
					# remove empty elements
					SDS_names = [x for x in SDS_names if x != '']

					# remove leading or tailing white spaces for each file name if there are
					SDS_names = list(map(lambda x: x.strip(), SDS_names))
				else:
					print("-sds= has no value, exit")
					sys.exit(0)
			elif re.search('-proj=(.+)', arg, re.IGNORECASE):
				arg_proj = 1
				m = re.search('-proj=(.+)', arg, re.IGNORECASE)
				if m:
					proj_code = m.group(1)
					if not is_int(proj_code):
						print("Error: %s is not integer number in %s, exit..." % (proj_code, arg))
						sys.exit(0)
					if not (1 <= int(proj_code) <= 22):
						print("Error: %s in %s should be a number from 1 to 22, exit..." % (proj_code, arg))
						sys.exit(0)
				else:
					print("-proj= has no value, exit.")
					sys.exit(0)
			elif re.search('-pixsz=(.+)', arg, re.IGNORECASE):
				arg_pixsz = 1
				m = re.search('-pixsz=(.+)', arg, re.IGNORECASE)
				if m:
					pixszs = m.group(1)
					pixszs = pixszs.split(',')
					# remove empty elements
					pixszs = [x for x in pixszs if x != '']
					# are they float?
					for pixsz in pixszs:
						if not is_float(pixsz):
							print("Error: %s is not float number in %s, exit..." % (pixsz, arg))
							sys.exit(0)
						if float(pixsz) <= 0:
							print("Error: %s in %s should be in KM and greater than 0., exit..." % (pixsz, arg))
							sys.exit(0)

					# covert them to float
					pixszs = [float(x) for x in pixszs]
					if float(pixsz) <= 0:
						print("Error: %s in %s should be in KM and greater than 0., exit..." % (pixsz, arg))
						sys.exit(0)
				else:
					print("-pixsz= has no value, exit")
					sys.exit(0)
			elif re.search('-fill=(.+)', arg, re.IGNORECASE):
				arg_fill = 1
				m = re.search('-fill=(.+)', arg, re.IGNORECASE)
				if m:
					fills = m.group(1)
					fills = fills.split(',')
					# remove empty elements
					fills = [x for x in fills if x != '']
					# are they numbers?
					for fill in fills:
						if not is_float(fill):
							print("Error: %s is not number in %s, exit..." % (fill, arg))
							sys.exit(0)
				else:
					print("-fill= has no value, exit")
					sys.exit(0)
			elif re.search('-tmp_path=(.+)', arg, re.IGNORECASE):
				arg_tmp_path = 1
				m = re.search('-tmp_path=(.+)', arg, re.IGNORECASE)
				if m:
					# Note that we don't check validity of tmp_path here but in HDF5_to_HDF4()
					tmp_path = m.group(1)
				else:
					print("-tmp_path= has no value, exit.")
					sys.exit(0)
			elif re.search('-geoif=(.+)', arg, re.IGNORECASE):
				arg_geo_HDFs = 1
				m = re.search('-geoif=(.+)', arg, re.IGNORECASE)
				if m:
					geoif_val = m.group(1)
					if os.path.isdir(geoif_val):
						geo_HDFs = get_HDFs_from_a_dir(geoif_val)
					else:
						# if wildcard pattern match
						geo_HDFs = glob.glob(geoif_val)

						# else they are separated by comma
						if len(geo_HDFs) == 0:
							geo_HDFs = m.group(1)
							geo_HDFs = geo_HDFs.split(',')

							# remove empty elements
							geo_HDFs = [x for x in geo_HDFs if x != '']

							# remove leading or tailing white spaces for each file name if there are
							geo_HDFs = list(map(lambda x: x.strip(), geo_HDFs))

					# check the consistency in the geolocation files
					esdt_0, time_0, ext_0, _ = get_ESDT_from_HDFname(geo_HDFs[0])
					if type(esdt_0) != str:
						print("cannot get esdt name from the geolocation file %s, exit." % geo_HDFs[0])
						sys.exit(0)
					if type(time_0) != str:
						print("cannot get time from the geolocation file %s, exit." % geo_HDFs[0])
						sys.exit(0)
					if type(ext_0) != str: 
						print("cannot get extension name from the geolocation file %s, exit." % geo_HDFs[0])
						sys.exit(0)
					for geo_HDF in geo_HDFs:
						esdt_tmp, time_tmp, ext_tmp, _ = get_ESDT_from_HDFname(geo_HDF)
						if esdt_tmp != esdt_0:
							print("The geolocationt files have inconsistent ESDT names %s and %s, exit." % (esdt_tmp, esdt_0))
							sys.exit(0)
						if time_tmp != time_0:
							print("The geolocation files have inconsistent time %s and %s, exit." % (time_tmp, time_0))
							sys.exit(0)
						if ext_tmp  != ext_0:
							print("The geolocation files have inconsistent extension %s and %s, exit." % (ext_tmp, ext_0))
							sys.exit(0)
						if not (os.path.isfile(geo_HDF) and os.access(geo_HDF, os.R_OK)):
							print("%s doesn't exist or not readble, exit" % geo_HDF)
							sys.exit(0)
				else:
					print("-geoif= has no value, exit.")
					sys.exit(0)
			# other arguments
			elif re.search('-DN=(.+)', arg, re.IGNORECASE):
				arg_DN = 1
				m = re.search('-DN=(.+)', arg, re.IGNORECASE)
				if m:
					DN_val = " " + arg + " "
					DN_elems = m.group(1)
					DN_elems = DN_elems.split(',')
					DN_elems = [x for x in DN_elems if x != '']
					for DN_elem in DN_elems:
						if not is_int(DN_elem):
							print("Error: %s is not integer number in %s, exit..." % (DN_elem, arg))
							sys.exit(0)
						if not (1 <= int(DN_elem) <= 3):
							print("Error: %s in %s should be 1,2 or 3, exit..." % (DN_elem, arg))
							sys.exit(0)
				else:
					print("-DN= has no value, exit.")
					sys.exit(0)
			elif re.search('-vector=(.+)', arg, re.IGNORECASE):
				arg_vector = 1
				m = re.search('-vector=(.+)', arg, re.IGNORECASE)
				if m:
					vector_val = " " + arg + " "
					vector = m.group(1)
					if not is_int(vector):
						print("Error: %s is not integer number in %s, exit..." % (vector, arg))
						sys.exit(0)
					if not (1 <= int(vector) <= 8):
						print("Error: %s in %s should be any number ranging from 1 to 8, exit..." % (vector, arg))
						sys.exit(0)
				else:
					print("-vector= has no value, exit.")
					sys.exit(0)
			elif re.search('-global', arg, re.IGNORECASE):
				arg_global = 1
				global_val = " -global "     #don't remove spaces
			else:
				print("There is no such option %s for the arguments, exit" % arg)
				sys.exit(0)

	# get the summation of all_args_except_in_out_HDF
	all_args_except_in_out_HDF = arg_sds + arg_proj + arg_pixsz + arg_fill + arg_tmp_path \
				+ arg_geo_HDFs + arg_DN + arg_vector + arg_global
	#########################
	# required arguments
	#########################
	# must have in_HDF and out_HDF in the arguments
	if arg_in_HDFs == 0:
		print_help_str()
		print("There is no input file -if= argument, exit")
		sys.exit(0)

	if arg_out_HDF == 0:
		print("There is no input file -of= argument, exit")
		# note that I iintend not use sys.exit(0) here but just as a warning information,
		# the below case will exit for it so that use can get all valid SDS names.

	if (arg_in_HDFs == 1 and arg_out_HDF == 0) or (arg_in_HDFs == arg_out_HDF == 1 and all_args_except_in_out_HDF == 0):
		print("There is no input SDS name -sds= argument")
		SDS_list, _, _, _ = get_HDF_SDS_info_exclude_lat_lon(in_HDFs[0])
		if SDS_list is not None:
			print("please use any valid SDS name(s) in the below list obtained from the input file")
			print(SDS_list)
		sys.exit(0)

	if arg_out_HDF == 0:
		print_help_str()
		print("There is no output file -of= argument, exit")
		sys.exit(0)

	# user must specify one or more SDSs
	if arg_sds == 0:
		print("There is no input SDS name -sds= argument")
		if arg_in_HDFs == 1:
			SDS_list, _, _, _ = get_HDF_SDS_info_exclude_lat_lon(in_HDFs[0])
			if SDS_list is not None:
				print("please use any valid SDS name(s) in the below list obtained from the input file:")
				print(SDS_list)
		sys.exit(0)

	# check if SDS_name valid
	# SDS_names_orig removes n from sds.n
	SDS_names_orig = []
	for SDS_name in SDS_names:
		sds_valid, sds_name_orig = SDS_name_valid(SDS_name, in_HDFs[0])
		if not sds_valid:
			print("Error: input -sds= argument may be wrong, because there is no %s in the input file." % SDS_name)
			if arg_in_HDFs == 1:
				SDS_list, _, _, _ = get_HDF_SDS_info_exclude_lat_lon(in_HDFs[0])
				if SDS_list is not None:
					print("please use any valid SDS name(s) in the below list obtained from the input file:")
					print(SDS_list)
			sys.exit(0)
		else:
			if sds_name_orig is not None:
				SDS_names_orig.append(sds_name_orig)
			else:
				SDS_names_orig.append(SDS_name)
				
	################################################
	################################################
	#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	# GET KEY INFORMATION FOR THE INPUT DATA PRODUCT:
	# HDF5 or HDF4; grid or swath; MODIS or non-MODIS
	# and all valid SDS names, shape and data type. 
	#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	################################################
	################################################
	# INFORMATION 1. get HDF4/HDF5 version, and all SDS_list, shape
	SDS_list, SDS_shape, SDS_dtype, HDF_ver = get_HDF_SDS_info_exclude_lat_lon(in_HDFs[0])
	in_files_HDF4_HDF5 = HDF_ver

	# INFORMATION 2. grid or swath data products
	# using read_meta tool to determine if the input HDF is projected data
	# grid_or_swath_HDF() returns
	#  0 if no StructMetadata.0 at all
	#  1 swath with StructMetadata.0 and
	#  2 projected with StructMetadata.0
	in_files_grid_swath = grid_or_swath_HDF(in_HDFs[0])
	esdt_code = 0  # for distinguish fire and snow purpose

	# INFORMATION 3. sensor: MODIS, VNP or VJ1
	in_files_sensor = ''
	esdt_name, _, _, _ = get_ESDT_from_HDFname(in_HDFs[0])
	re_rt_MODIS = re.search('^M(O|Y|C)D(\d+)', esdt_name)
	re_rt_VNP = re.search('^(VNP|NPP)(\d+)', esdt_name)
	re_rt_VJ1 = re.search('^VJ1(\d+)', esdt_name)
	if re_rt_MODIS:
		in_files_sensor = 'MODIS'
		esdt_code       = int(re_rt_MODIS.group(2))
	elif re_rt_VNP:
		in_files_sensor = 'VNP'
		esdt_code       = int(re_rt_VNP.group(2))
	elif re_rt_VJ1:
		in_files_sensor = 'VJ1'
		esdt_code       = int(re_rt_VJ1.group(1))
	else:
		print("Don't know the sensor name for the input data product with ESDT name of %s, exit." % esdt_name)
		sys.exit(0)
	# not use the below for VNP because Min told me VNP
	# names is unlike MODIS, and they are inconsistent.
	#npp   = re.search('^VNP', esdt_name)

	# Note that the input file will be always the last in the projbrowse's argument
	# so for the basic_cmd_list, we will not list it but late for the final cmd_list
	#########################################
	# optional arguments
	#########################################
	#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
	# combined SDS_list with SDS_shape using library
	SDS_shape_dict = dict(zip(SDS_list, SDS_shape))
	SDS_type_dict  = dict(zip(SDS_list, SDS_dtype))

	fills_strs = []

	# check the consistency of number of specified SDSs and pixszs
	if arg_pixsz == 1:
		if len(pixszs) > 1:
			if len(SDS_names_orig) != len(pixszs):
				print("the number of SDSs in -sds= is different from number of pixsz in -pixsz=, exit")
				sys.exit(0)
		else:
			# expand one pixsz for all SDSs
			if len(SDS_names_orig) > 1:
				pixszs = pixszs * len(SDS_names_orig)

	# check the consistency of number of specified SDSs and fill value
	if arg_fill == 1:
		if len(fills) > 1:
			if len(SDS_names_orig) != len(fills):
				print("the number of fill values in -fill= is different from number of fill values in -fill=, exit")
				sys.exit(0)
		else:
			# expand one fill_elem for all SDSs
			if len(SDS_names_orig) > 1:
				fills = fills * len(SDS_names_orig)

		# check if the fill values are valid
		for idx, SDS_name in enumerate(SDS_names_orig):
			SDS_dtype = SDS_type_dict[SDS_name]
			SDS_dtype = SDS_dtype.lower()
			# note that fill_elem is string not number
			fill_elem = fills[idx]
			
			if not check_data_range(fill_elem, SDS_dtype):
				print("Error: %s in -fill= argument is beyond the limits of data type %s for %s, plese use a different fill value, exit" % (fill_elem, SDS_dtype, SDS_name))
				exit(0)
			else:
				fill_str = " -fill" + SDS_dtype + "=" + fill_elem + " "
				fills_strs.append(fill_str)

	# get the default or native pixsz
	pixel_size_set = []
	grid_HDF = 0
	grid_pixsz_success = 0
	if in_files_grid_swath == 2:  # gridded HDF
		grid_HDF = 1
		num_grids, upleft_coor, lowright_coor, proj_geo, field_names_num, field_names = parse_struct_metadata(in_HDFs[0])
		if num_grids > 0:
			grid_pixsz_success = 1
			for SDS_name in SDS_names_orig:
				shape = SDS_shape_dict[SDS_name]
				pixel_size, relative_diff = get_native_pixsz_grid(shape, proj_geo[0],
						 upleft_coor[0], lowright_coor[0])
				pixel_size_set.append(float(pixel_size))
				if relative_diff > 1.:
					print("Warning: The pixsz is different between X and Y dimensions (>1%), something may be wrong with the input file.")
		else:
			grid_pixsz_success = 0
		
	else: # use look-up-table to get the default pixsz for swath data products late
		if in_files_sensor == 'MODIS':
			for SDS_name in SDS_names_orig:
				pixel_size, Ydim, Xdim = get_MODIS_pixsz_swath(SDS_shape_dict[SDS_name])
				if pixel_size is None and arg_pixsz == 0:
					print("You must specify -pixsz= in the argument because the tool cannot infer it from data shape, exit.")
					sys.exit(0)
				else:
					pixel_size_set.append(float(pixel_size))  # default pixsz for swath data, update late
		elif in_files_sensor == 'VNP':
			for SDS_name in SDS_names_orig:
				pixel_size, Ydim, Xdim = get_VNP_pixsz_swath(SDS_shape_dict[SDS_name])
				if pixel_size is None and arg_pixsz == 0:
					print("You must specify -pixsz= in the argument because the tool cannot infer it from data shape, exit.")
					sys.exit(0)
				else:
					pixel_size_set.append(float(pixel_size))  # default pixsz for swath data, update late
		elif in_files_sensor == 'VJ1':
			for SDS_name in SDS_names_orig:
				pixel_size, Ydim, Xdim = get_VJ1_pixsz_swath(SDS_shape_dict[SDS_name])
				if pixel_size is None and arg_pixsz == 0:
					print("You must specify -pixsz= in the argument because the tool cannot infer it from data shape, exit.")
					sys.exit(0)
				else:
					pixel_size_set.append(float(pixel_size))  # default pixsz for swath data, update late
		else:   # VNP swath pixsz
			for SDS_name in SDS_names_orig:
				if arg_pixsz == 0:
					print("You must specify -pixsz= , exit.")
					sys.exit(0)

	# user must specify pixsz if fails to get native pixsz from the StructMetadata.0
	if grid_HDF == 1 and grid_pixsz_success == 0 and arg_pixsz == 0:
		print("cannot get the native pixsz from the StructMetadata.0, user must specify -pixsz=, exit...")
		sys.exit(0)

	# overritten the default or native pixsz
	if arg_pixsz == 1:
		for idx, pixel_size in enumerate(pixel_size_set):
			pixel_size_set[idx] = pixszs[idx]

	# default target map projection is 16 Hammer-Aitoff, or overritten if there is input
	#proj_set = 22       # default proj value
	proj_set = 16
	if arg_proj == 1:
		proj_set = int(proj_code)

	# assign the correct projbrowse tools for MODIS and VNP data products
	esdt_name, _, _, _ = get_ESDT_from_HDFname(in_HDFs[0])

	proj_cmd = ""
	if in_files_sensor == 'MODIS':
		proj_cmd = "projbrowse"
	else:
	#elif npp:
		proj_cmd = "projbrowse_NPP"
	#else:
	#	print("This version just deals with MODIS or VNP data products, exit...")
	#	sys.exit(0)

	proj_exe_cmd = None
	exe_fname = is_exe_file(proj_cmd)
	if exe_fname is None:
		print("Error: cannot use %s at any location, follow user guide to make it usable at any location, exit.")
		sys.exit(0)
	else:
		proj_exe_cmd = proj_cmd

	if proj_exe_cmd is None:
		print("Error: cannot find projbrowse or projbrowse_NPP command, exit")
		sys.exit(0)

	if in_files_grid_swath != 2:
		# if there are geolocation SDSs, i.e.: latitude and longitude ?
		lat_exist, lon_exist = swath_has_latlon(in_HDFs[0])
		if not (lat_exist and lon_exist):
			if arg_geo_HDFs == 0:
				print("Error: No geolocation lat/lon for the swath file %s, need to provide Geolocation file -geoif=, exit" % in_HDFs[0])
				sys.exit(0)
			else:
				if len(in_HDFs) != len(geo_HDFs):
					print("the number of input files %d is different from the number of geolocation files %d, exit" % (len(in_HDFs), len(geo_HDFs)))

				# if the geo_HDF has lat/lon
				lat_exist, lon_exist = swath_has_latlon(geo_HDFs[0])
				if not (lat_exist and lon_exist):
					print("The geolocation file %s has no latitude and longitude SDSs, exit" % geo_HDFs[0])
					sys.exit(0)
				else:
					print("copying the latitude and longitude from geolocation file (-geoif) to input file (-if)")
					# first make sure the two HDFs have the consistent time and granuleID
					for idx, HDF_temp in enumerate(in_HDFs):
						_, datatime, _, gID         = get_ESDT_from_HDFname(HDF_temp)
						_, datatime_geo, _, gID_geo = get_ESDT_from_HDFname(geo_HDFs[idx])
						if type(datatime) != str or type(datatime_geo) != str or type(gID) != str or type(gID_geo) != str:
							print("cannot get the data time and granuleID from the file names of %s and/or %s, exit." % (HDF_temp, geo_HDFs[idx]))
							sys.exit(0)
						else:
							if datatime != datatime_geo:
								print("input file %s and geolocation file %s have different data time, exit." % (HDF_temp, geo_HDFs[idx]))
								sys.exit(0)
							if gID != gID_geo:
								print("input file %s and geolocation file %s have different granuleID, exit." % (HDF_temp, geo_HDFs[idx]))
								sys.exit(0)

					# secondly, convert HDF5 to HDF4 for the input HDFs if they are.
					# 	   Also update the inf names to the new HDF4 filenames and in_files_HDF4_HDF5 to be 4 
					if in_files_HDF4_HDF5 == 5:
						SDS_list = SDS_names_orig
						for idx, HDF5_temp in enumerate(in_HDFs):
							# note that the returned in_HDF is different from in_HDF in the argument
							HDF4_temp = HDF5_to_HDF4(HDF5_temp, SDS_list, tmp_path)
							in_HDFs[idx] = HDF4_temp
						# the input files are in HDF4 not HDF5
						in_files_HDF4_HDF5 = 4
						HDF5_to_HDF4_temp_path = os.path.dirname(in_HDFs[0])

					# thirdly, get the SDS names for lat and lon from the first geo_HDFs list
					# note we have checked if there are lat/lon in the geo_HDFs above swath_has_latlon()
					lat_sds_name, lon_sds_name = get_LAT_LON_SDS_names(geo_HDFs[0])

					# fourthly, copy lat/lon from geoif (regardless of HDF4/HDF5) to inf (must be HDF4).
					for idx, HDF_temp in enumerate(in_HDFs):
						copy_geolocation(geo_HDFs[idx], lat_sds_name, lon_sds_name, HDF_temp)
						
	# add the in_HDF as the last argument of projbrowse
	# copy HDF5/NC to HDF4 using copy_sds
	if in_files_HDF4_HDF5 == 5:
		SDS_list = SDS_names_orig
		for idx, HDF5_temp in enumerate(in_HDFs):
			HDF4_temp = HDF5_to_HDF4(HDF5_temp, SDS_list, tmp_path)
			in_HDFs[idx] = HDF4_temp

		in_files_HDF4_HDF5 = 4
		HDF5_to_HDF4_temp_path = os.path.dirname(in_HDFs[0])

	# put all input files together into one long string
	in_HDF4s = " ".join(in_HDFs)

	# put all other parameters together
	# Note that all element value have leading and trailing spaces
	fire_val = ""
	if esdt_code == 14:
		fire_val = " -fire -magnify=1 "

	snow_val = ""
	if esdt_code == 10:
		if in_files_sensor == "MODIS":
			snow_val = " -snow_c6 "	
			#  for A2, should use -snow NOT -snow_c6
			if len(esdt_name) >= 7 and esdt_name[5:7] == "A2":
				snow_val = " -snow "
		else:
			snow_val = " -snow "

	other_parameters = " " + DN_val + vector_val + global_val + fire_val + snow_val

	# remove output file if it exists
	if os.path.isfile(out_HDF):
		os.remove(out_HDF)
			   
	for idx, SDS_name in enumerate(SDS_names): #NOTE!!!! SDS_names not SDS_names_orig should be used here
		# make the sds name have the double quotation in case there are spaces in the SDS name
		sds_name = '"' + SDS_name + '"'

		#######################
		# basic command list for all cases
		# just leave the last argument, input HDF
		#######################	
		basic_cmd_list = [
				proj_exe_cmd,
				"-append",
				"-clean",
				"-proj="+str(proj_set),
				"-pixsz="+str(pixel_size_set[idx]),
				"-sds="+sds_name,
				"-of="+out_HDF,
				]

		cmd_list = deepcopy(basic_cmd_list)
		# append other_parameters
		cmd_list.append(other_parameters)

		# append fill value if there is
		if len(fills_strs) > 0:
			cmd_list.append(fills_strs[idx])

		# append the input files which must be the last element of the argument for the projbrowse
		cmd_list.append(in_HDF4s)

		# make the command line	for reprojection
		#print(cmd_list)
		cmd_line = " ".join(cmd_list)
	
		#print(cmd_line)
		print("Reprojection...")
		cmd_return, stdout_str, stderr_str = run_command(cmd_line)

		if type(cmd_return) == str:
			print(cmd_return)
		if type(stdout_str) == str:
			print(stdout_str)
		if type(stderr_str) == str:
			print(stderr_str)

	# remove the intermediate HDF4 input files in the temporary folder
	# I don't remove the temporary folder in case users also put the output file there.
	if HDF5_to_HDF4_temp_path is not None and type(HDF5_to_HDF4_temp_path) == str:
		for temp_HDF in in_HDFs:
			os.remove(temp_HDF)

		# now if it is an empty directory, remove it
		if len(os.listdir(HDF5_to_HDF4_temp_path) ) == 0:
			os.rmdir(HDF5_to_HDF4_temp_path)
			
#$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
if __name__ == '__main__':
	import sys

	# check if Python has the right version
	if not (sys.version_info.major == 3 and sys.version_info.minor >= 5):
		print("This script requires Python 3.5 or higher!")
		print("You are using Python {}.{}.".format(sys.version_info.major, sys.version_info.minor))
		sys.exit(1)

	# run it or stop and exit
	try:
		sys.exit(_main(sys.argv))
	except KeyboardInterrupt:
		sys.exit(-1)
