/*******************************************************************************
NAME                            ECKERT IV

PURPOSE:	Transforms input longitude and latitude to Easting and
		Northing for the Eckert IV projection.  The
		longitude and latitude must be in radians.  The Easting
		and Northing values will be returned in meters.

PROGRAMMER              DATE
----------              ----
J. Descloitres, UMD	July, 1998
(Based on GCTP routines written by D. Steinwand, S. Nelson and T. Mittan)

ALGORITHM REFERENCES

1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
*******************************************************************************/
#include "cproj.h"

/* Variables common to all subroutines in this code file
  -----------------------------------------------------*/
static double lon_center;	/* Center longitude (projection center) */
static double R;		/* Radius of the earth (sphere) */
static double false_easting;	/* x offset in meters			*/
static double false_northing;	/* y offset in meters			*/

/* Initialize the Eckert IV projection
  ------------------------------------*/
eckivforint(r, center_long,false_east,false_north)
double r; 			/* (I) Radius of the earth (sphere) 	*/
double center_long;		/* (I) Center longitude 		*/
double false_east;		/* x offset in meters			*/
double false_north;		/* y offset in meters			*/
{
/* Place parameters in static storage for common use
  -------------------------------------------------*/
false_easting = false_east;
false_northing = false_north;
R = r;
lon_center = center_long;

/* Report parameters to the user
  -----------------------------*/
ptitle("ECKERT IV"); 
radius(r);
cenlon(center_long);
offsetp(false_easting,false_northing);
return(OK);
}

/* Eckert IV forward equations--mapping lat,long to x,y
  ----------------------------------------------------*/
eckivfor(lon, lat, x, y)
double lon;			/* (I) Longitude */
double lat;			/* (I) Latitude */
double *x;			/* (O) X projection coordinate */
double *y;			/* (O) Y projection coordinate */
{
double adjust_lon();	/* Function to adjust longitude to -180 - 180 */
double delta_lon;	/* Delta longitude (Given longitude - center */
double theta;
double delta_theta;
double sin_theta;
double cos_theta;
double con;
long i;

/* Forward equations
  -----------------*/
delta_lon = adjust_lon(lon - lon_center);
theta = lat / 2;
con = (2 + HALF_PI) * sin(lat);

/* Iterate using the Newton-Raphson method to find theta
  -----------------------------------------------------*/
for (i=0;;i++)
   {
   sincos(theta, &sin_theta, &cos_theta);
   delta_theta = -(theta + sin_theta * cos_theta + 2 * sin_theta - con) / (2 * cos_theta * (1.0 + cos_theta));
   theta += delta_theta;
   if (fabs(delta_theta) < EPSLN) break;
   if (i >= 50) 
     {
     p_error("Iteration failed to converge","EckertIV-forward");
     return(ERROR);
     }
   }

*x = 0.4222382003158 * R * delta_lon * (1 + cos(theta)) + false_easting;
*y = 1.3265004281770 * R * sin(theta) + false_northing;

return(OK);
}
