/*******************************************************************************
NAME                            Eckert IV

PURPOSE:	Transforms input Easting and Northing to longitude and
		latitude for the Eckert IV projection.  The
		Easting and Northing must be in meters.  The longitude
		and latitude values will be returned in radians.

PROGRAMMER              DATE
----------              ----
J. Descloitres, UMD     July, 1998
(Based on GCTP routines written by D. Steinwand, S. Nelson and T. Mittan)

ALGORITHM REFERENCES

1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
*******************************************************************************/
#include "cproj.h"

/* Variables common to all subroutines in this code file
  -----------------------------------------------------*/
static double lon_center;	/* Center longitude (projection center) */
static double R;		/* Radius of the earth (sphere) */
static double false_easting;	/* x offset in meters			*/
static double false_northing;	/* y offset in meters			*/

/* Initialize the Eckert IV projection
  ------------------------------------*/
eckivinvint(r, center_long, false_east, false_north) 
double r; 			/* (I) Radius of the earth (sphere) */
double center_long;		/* (I) Center longitude */
double false_east;		/* x offset in meters			*/
double false_north;		/* y offset in meters			*/
{
/* Place parameters in static storage for common use
  -------------------------------------------------*/
false_easting = false_east;
false_northing = false_north;
R = r;
lon_center = center_long;

/* Report parameters to the user
  -----------------------------*/
ptitle("ECKERT IV"); 
radius(r);
cenlon(center_long);
offsetp(false_easting,false_northing);
return(OK);
}

/* Eckert IV inverse equations--mapping x,y to lat,long 
  ----------------------------------------------------*/
eckivinv(x, y, lon, lat)
double x;		/* (I) X projection coordinate */
double y;		/* (I) Y projection coordinate */
double *lon;		/* (O) Longitude */
double *lat;		/* (O) Latitude */
{
double adjust_lon();    /* Function to adjust longitude to -180 - 180 */
double theta;
double sin_theta;
double cos_theta;
double arg;

/* Inverse equations
  -----------------*/

x -= false_easting;
y -= false_northing;
arg = y /  (1.3265004281770 * R);

/* Because of division by zero problems, 'arg' can not be 1.0.  Therefore
   a number very close to one is used instead.
   -------------------------------------------------------------------*/
if(fabs(arg) > 0.999999999999) arg=0.999999999999;
theta = asin(arg);
sincos(theta, &sin_theta, &cos_theta);

*lon = adjust_lon(lon_center + (x / (0.4222382003158 * R * (1 + cos_theta))));
if (fabs(*lon) > PI) return(ERROR);

arg = (theta + sin_theta * cos_theta + 2 * sin_theta) / (2 + HALF_PI);
if(fabs(arg) > 1.0) return(ERROR);
*lat = asin(arg);

return(OK);
}
