/*******************************************************************************
NAME                        INTERRUPTED MOLLWEIDE

PURPOSE:	Transforms input longitude and latitude to Easting and
		Northing for the Interrupted Mollweide projection.  The
		longitude and latitude must be in radians.  The Easting
		and Northing values will be returned in meters.

PROGRAMMER              DATE
----------              ----
J. Descloitres, UMD	July, 1998
(Based on GCTP routines written by D. Steinwand, S. Nelson and T. Mittan)

ALGORITHM REFERENCES

1.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
*******************************************************************************/
#include "cproj.h"

/* Variables common to all subroutines in this code file
  -----------------------------------------------------*/
static double R;		/* Radius of the earth (sphere) */
static double false_easting;	/* x offset in meters			*/
static double false_northing;	/* y offset in meters			*/
static double lon_center[6];	/* Central meridians, one for each region */
static double feast[6];		/* False easting, one for each region */

/* Initialize the Interrupted Mollweide projection
  --------------------------------------------*/
imolwforint(r, false_east,false_north) 
double r; 			/* (I) Radius of the earth (sphere) */
double false_east;		/* x offset in meters                   */
double false_north;		/* y offset in meters                   */
{
/* Place parameters in static storage for common use
  -------------------------------------------------*/
false_easting = false_east;
false_northing = false_north;
R = r;

/* Initialize central meridians for each of the 6 regions
  -------------------------------------------------------*/
lon_center[0] = -100 * D2R;	/* -100.0 degrees */
lon_center[1] =   30 * D2R;	/*   30.0 degrees */
lon_center[2] =  -80 * D2R;	/*  -80.0 degrees */
lon_center[3] =  -80 * D2R;	/*  -80.0 degrees */
lon_center[4] =   20 * D2R;	/*   20.0 degrees */
lon_center[5] =  140 * D2R;	/*  140.0 degrees */

/* Initialize false eastings for each of the 6 regions
  ----------------------------------------------------*/
feast[0] = 0.900316316158 * R * -100 * D2R;
feast[1] = 0.900316316158 * R *   30 * D2R;
feast[2] = 0.900316316158 * R *  -80 * D2R;
feast[3] = 0.900316316158 * R *  -80 * D2R;
feast[4] = 0.900316316158 * R *   20 * D2R;
feast[5] = 0.900316316158 * R *  140 * D2R;

/* Report parameters to the user
  -----------------------------*/
ptitle("INTERRUPTED MOLLWEIDE EQUAL-AREA"); 
radius(r);
offsetp(false_easting,false_northing);
return(OK);
}

/* Interrupted Mollweide forward equations--mapping lat,long to x,y
  -------------------------------------------------------------*/
imolwfor(lon, lat, x, y)
double lon;			/* (I) Longitude */
double lat;			/* (I) Latitude */
double *x;			/* (O) X projection coordinate */
double *y;			/* (O) Y projection coordinate */
{
double adjust_lon();	/* Function to adjust longitude to -180 - 180 */
double delta_lon;	/* Delta longitude (Given longitude - center */
double theta;
double delta_theta;
double constant;
long i;
long region;

/* Forward equations
  -----------------*/
if (lat >= 0.0)
   {
   if (lon <= -0.698131700798)			/* West of -40 deg */
     if (lon <= -2.94925737002 &&			/* West of -168.98 deg in Bering Strait */
         lat - 1.047197551197 >= (lon + PI) / 2.) {	/* North of Saint Lawrence and Saint Matthew Islands */
         region = 1;					/* Affect this region (Eastern Siberia) to region 1 */
         lon += TWO_PI;					/* Add TWO_PI to be to place this region at the right of region 1 */ 
     } else region = 0;					/* All other cases West of -40 deg are in region 0 */
   else						/* East of -40 deg */
     if (lon <= -0.174532925199 &&				/* West of -10 deg */
         lat >= 1.047197551197 &&				/* North of 60 deg */
         lat - 1.0471975512 >= (lon + 0.698131700798) / 2.)	/* North of Iceland in Denmark Strait */
           region = 0;						/* Affect this region (Eastern Greenland) to region 0 */
     else  region = 1;						/* All other cases East of -40 deg are in region 1 */
   }
else
   {
   if (lon <= -1.74532925199) region = 2;		/* If between -180 and -100 */
   else if (lon <= -0.349065850399) region = 3;		/* If between -100 and -20 */
   else if (lon <= 1.3962634016) region = 4;		/* If between -20 and 80 */
   else region = 5;					/* If between 80 and 180 */
   }

   delta_lon = adjust_lon(lon - lon_center[region]);
   theta = lat;
   constant = PI * sin(lat);

/* Iterate using the Newton-Raphson method to find theta
  -----------------------------------------------------*/
   for (i=0;;i++)
      {
      delta_theta = -(theta + sin(theta) - constant) / (1.0 + cos(theta));
      theta += delta_theta;
      if (fabs(delta_theta) < EPSLN) break;
      if (i >= 50) 
         {
         p_error("Iteration failed to converge","InterMoll-forward");
	 return(251);
         }
      }
   theta /= 2.0;

   /* If the latitude is 90 deg, force the x coordinate to be
      "0 + false easting" this is done here because of precision problems
      with "cos(theta)"
      ------------------------------------------------------------------*/
   if (PI / 2 - fabs(lat) < EPSLN)
      delta_lon = 0;
   *x = false_easting  + feast[region] + 0.900316316158 * R * delta_lon * cos(theta);
   *y = false_northing + R * 1.4142135623731 * sin(theta);

return(OK);
}
