/*
 File: projbrowse.c

 Linux Compiler options:
      cc -O projbrowse.c -o projbrowse ./LIB/libproj.a -I$HDFINC -L$HDFLIB -lmfhdf -ldf -lz -lm -ljpeg

 Description:
   Project a list of SDS or metadata into a geolocated projection. Various
   output projections are available. The output domain is either the actual
   bounding box of the complete input data set or the whole Earth. If the
   input is not in HDF-EOS format, some input characteristics are requested.
   Also coastlines, rivers and boundaries can be plotted.
   The output is an HDF file.

 Revision History:
   Version 1.0        May 11, 1998, J. Descloitres (MODLAND/LDOPE)
   Version 1.1        May 26, 1998, J. Descloitres (MODLAND/LDOPE)
   Version 1.1.2  January 13, 1999, J. Descloitres (MODLAND/LDOPE)
   Version 1.1.3  February 4, 1999, J. Descloitres (MODLAND/LDOPE)
   Version 1.1.4     March 9, 1999, J. Descloitres (MODLAND/LDOPE)
   Version 2.0        June 4, 2000, J. Descloitres (MODIS/Land/250m)
   Version 2.1    October 18, 2000, J. Descloitres (MODIS/Land/250m)
   Version 2.2       June 28, 2001, J. Descloitres (MODIS/Land/RR)
   Version 2.3   November 10, 2001, J. Descloitres (MODIS/Land/RR)

 Team-unique Header:
   This software was developed by:
   MODIS Land group (MODLAND) for the National Aeronautics and Space Administration, Goddard Space Flight Center.

 References and Credits:
    Jacques Descloitres (MODIS / Land / Rapid Response)
    Science Systems & Applications, Inc. (SSAI)
    NASA / GSFC  Code 922
    Bldg 33 Room G108
    Greenbelt, MD 20771
    Phone: 301 614 5456
    Email: jack@ltpmail.gsfc.nasa.gov
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>		/* Defines FLT_MAX and DBL_MAX (fix for RedHat 7.2) */
#include "mfhdf.h"
#include "../proj_lib/cproj.h"

#include <malloc.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/dir.h>
#include <sys/types.h>

#define SOLZENMAX       7500              /* Maximum solar zenith angle: added 07/13 */
#define	DEFPIXSZ	5.		/* Default pixel size (km) */
#define MINPIXSZ	0.05		/* Min pixel size (km) */
#define MAXPIXSZ	100.		/* Max pixel size (km) */
#define	FILL_INT8	  125		/* Fill values for output SDS */
#define	FILL_UINT8	  253
#define	FILL_INT16	 -200
#define	FILL_UINT16	32750
#define	FILL_INT32	 -200
#define	FILL_UINT32	2147483600
#define	FILL_FLOAT32	-8421.
#define MAX_ATTR_SIZE	65535		/* Max. size of attributes (same as MAX_FIELD_SIZE ?) */
#define MAXMETALENGTH	10000		/* Max length for one metadata field (must be lower than MAX_ATTR_SIZE) */
#define MAXFILELENGTH	200		/* Max length for a file name */
#define MAXLINELENGTH	1000		/* Max length for a line in a file */
#define MAXNAMELENGTH	64		/* Max. length for SDS names because MAX_NC_NAME=256 is apparently too long for SDwritedata() */
#define DOMAINOFFSET	-50
#define OUTLINEOFFSET	-60
#define SATUROFFSET	-80
enum { MERIDIANS, PARALLELS, TILES, NODES, NGRIDS };
#define GRIDTYPES	{ "meridians", "parallels", "MODLAND L3 grid", "nodes" }
#define GRIDOFFSETS	{ -20, -20, -30, -40 }
#define NVECTORS	8		/* Number of vector files */
#define VECTORTYPES	{ "primary coastlines", "secondary coastlines", "intermittent coastlines", \
                          "primary rivers", "secondary rivers and canals", "intermittent rivers",  \
                          "international political boundaries", "national political boundaries" }
#define VECTOROFFSETS	{ -100, -100, -100, -110, -110, -110, -120, -130 }
#define DNFLAGTYPE	{ "Day", "Night", "Both (mixed)", "Both (mixed), Northern hemisphere", "Both (mixed), Southern hemisphere" }
#define	VECTORUNDEF	-8421		/* Undef value in vector files */
#define	PROJUNDEF	-1000		/* Undef value for projection coordinates */
#define RAD2DEG		57.29577951308232087684		/* 180 / PI */
#define DEG2RAD		0.01745329251994329576		/* PI / 180 */
/*
#define EARTHRADIUS	6370997.			/ * Earth radius in meters (from GCTP) * /
#define DEG2KM		111.19487428468117751582
*/
#define EARTHRADIUS	6371007.181			/* Earth radius in meters (from Robert Wolfe) */
#define DEG2KM		111.19505197665232305852	/* EARTHRADIUS * PI / 180 / 1000 = Length (in km) for 1 deg. of latitude */
#define PI		M_PI
#define BUFINC		10000		/* Buffer increment (in bytes) for compressed vector mask */
#define DEFRADIUS	500.		/* Default radius (in km) for site representation */
#define	DEFHEIGHT	35800.		/* Default observation height (in km) for vertical perspective projection */
#define ANCPATHVAR	"ANCPATH"
#define MODISPATHVAR	"MODISPATH"
#define MAXHDFDATATYPE	28
#define DEFMAXMEMORY	1800		/* Default maximum memory to allocate (in Mbytes) */
#define	MAXRANGE	80		/* Maximum range from center (in deg.) for a gnomonic projection */
#define MAXNZONES	100		/* Maximum number of zones for interrupted projections */
#define ZONEGAP		5.		/* Gap between zones for interrupted projections */
#define	MISSING		65534		/* Missing data in L1B */
#define	SATURATED	65533		/* Saturated data in L1B */

/*
#define MAX(x,y)	((x) > (y) ? (x) : (y))
#define MIN(x,y)	((x) < (y) ? (x) : (y))
*/
#define CROSS_PRODUCT(x1, y1, x2, y2, x3, y3)	( ((x2) - (x1)) * ((y3) - (y1)) - ((y2) - (y1)) * ((x3) - (x1)) )



enum { SWATH,			/* Enumeration of supported projections */
       SINUSOIDAL,
       INTEGERSIN,
       INTERSINUS1,
       INTERSINUS2,
       PLATECARREE,
       ROBINSON,
       MOLLWEIDE,
       INTERMOLL,
       GOODE,
       GOODE2,
       PERSPECTIVE,
       ORTHOGRAPHIC,
       GNOMONIC,
       STEREOGRAPHIC,
       LAMBERT,
       HAMMER,
       AITOFF,
       ECKERTIV,
       AZIMEQUI,
       ALBERS,
       INTERSINUS3,
       GEOGRAPHIC,
       NUMBEROFPROJ };

enum { EQUATORIAL, NORTHPOLE, SOUTHPOLE };	/* Gnomonic projections */
enum { GLOBAL, EXTBOX, USERBOX };		/* Output projection domain */
enum { FORWARD };				/* Mapping method */
enum { NONE, NEAREST };				/* Interpolation method */
enum { CENTERS, BOUNDS };
enum { BSQ, PIL, LIL };
enum { CREATE, APPEND, REPLACE };

/* Reduced help message */
#define HELPMSG	"Usage:\n"\
		"projection -help\n"\
		"projection -proj=projection -of=output_file [-DN=1,...,3]\n"\
		"           [-depth] [-append] [-global] [-separate]\n"\
                "           [-pixsz=pixel_size] [-box=startcol,startrow,width,height]\n"\
		"           -sds=SDS_name filenames -meta=metadata_name filenames\n"\
		"           [-sds=SDS_name filenames] [-meta=metadata_name filenames]\n"\
		"           [-ref=file,SDS[,projection,resolution,center_column,center_row]]\n"\
		"           [-vector=1,...,8] [-grid=1,...,4,resolution,stop]\n"
/* Full help message */
#define FULLHELP	"\n"\
		"NAME\n"\
		"     projection - project data sets into a global projection\n"\
		"\n"\
		"SYNOPSIS\n"\
		"     projection -help\n"\
		"     projection -proj=projection -of=output_file [-DN=1,...,3]\n"\
		"                [-depth] [-append] [-global] [-separate]\n"\
                "                [-pixsz=pixel_size] [-box=startcol,startrow,width,height]\n"\
		"                -sds=SDS_name filenames -meta=metadata_name filenames\n"\
		"                [-sds=SDS_name filenames] [-meta=metadata_name filenames]\n"\
		"                [-ref=file,SDS[,projection,pixel_size,center_column,center_row]]\n"\
		"                [-vector=1,...,8] [-grid=1,...,4,step,stop]\n"\
		"\n"\
		"     echo -proj=projection -of=output_file [-DN=1,...,3]\n"\
		"          [-depth] [-append] [-global] [-separate]\n"\
                "          [-pixsz=pixel_size] [-box=startcol,startrow,width,height]\n"\
		"          -sds=SDS_name filenames -meta=metadata_name filenames\n"\
		"          [-sds=SDS_name filenames] [-meta=metadata_name filenames]\n"\
		"          [-ref=file,SDS[,projection,pixel_size,center_column,center_row]]\n"\
		"          [-vector=1,...,8] [-grid=1,...,4,step,stop] | projection\n"\
		" or  projection < parameter_file\n"\
		"\n"\
		"DESCRIPTION\n"\
		"     Project a list of SDS or metadata into a geolocated projection. Various\n"\
		"     output projections are available. The output domain is either the actual\n"\
		"     bounding box of the complete input data set or the full globe. If the\n"\
		"     input is not in HDF-EOS format, some input characteristics are requested.\n"\
		"     Also coastlines, rivers and boundaries can be plotted.\n"\
		"     The output is an HDF file.\n"\
		"\n"\
		"OPTIONS\n"\
		"     -help                 Display this help message\n"\
		"     -proj=no              Specify the output projection (see below for details)\n"\
		"     -of=filename          Specify the output file\n"\
		"     -DN=1,...,3           Restrict the process to products that carry the specified values\n"\
		"                           of the Day/Night flag\n"\
		"                           Default is 1,2,3 (no restriction)\n"\
		"                           1 = Day\n"\
		"                           2 = Night\n"\
		"                           3 = Both (mixed)\n"\
		"     -depth                Allow recursive search for files from the specified paths\n"\
		"                           Default is recursive search\n"\
		"     -append               Append the specified output file\n"\
		"                           Default is overwrite access mode\n"\
		"     -global               Extend the output domain to the whole globe\n"\
		"                           Default is extended box (actual limits of the specified products)\n"\
		"     -separate             Creates a separate SDS and vectors for all overlays (grids, etc.).\n"\
		"                           Default is overlays burned into the images.\n"\
		"     -pixsz=pixel_size     Set the output pixel size (in km)\n"\
		"                           Default is 5 km\n"\
		"     -box=startcol,startrow,width,height\n"\
		"                           Restrict the output projection to a specific subwindow\n"\
		"                           The output subwindow is specified with the coordinates (in pixels)\n"\
                "                           of the upper left corner and the dimensions\n"\
		"     -sds=SDS[.no]         Specify the SDS to project (either SDS name or SDS index)\n"\
		"                           Apply to all following files, until -sds or -meta option\n"\
		"                           is used again to specify a new SDS name or metadata name\n"\
		"                           For a 3D SDS, a .no suffix can be used to specified the\n"\
		"                           plane number to project (default is 1)\n"\
		"     -meta=metadata_name   Specify the metadata to project\n"\
		"                           If the specified metadata is numeric, the specified granules or\n"\
		"                           are projected and shaded with the corresponding metadata value\n"\
		"                           No geolocation file associated to the products is needed with\n"\
		"                           that option (the granule edges are approximated by the great\n"\
		"                           circles)\n"\
		"                           If the specified metadata is a string, the metadata content is\n"\
		"                           reprensented by an integer, and a look-up table (LUT) is stored\n"\
		"                           in the file attributes\n"\
		"                           A partial name can be used (the first matching metadata\n"\
		"                           name will be considered)\n"\
		"                           Apply to all following files, until -sds or -meta option\n"\
		"                           is used again to specify a new SDS name or metadata name\n"\
		"     -ref=file,SDS[,projection,pixel_size,center_column,center_row]\n"\
		"                           Specify a reference product to project in the same output\n"\
		"                           projection as the specified HDF-EOS products\n"\
		"                           Some standard products (e.g., DAO) are recognized\n"\
		"                           For all others, the projection and the pixel size must be\n"\
		"                           specified\n"\
		"                           Also the projection center can be specified (default is the\n"\
		"                           center of the specified SDS)\n"\
		"     -vector=1,...,8       Add any combination of geographic vectors\n"\
		"                           1 = primary coastlines\n"\
		"                           2 = secondary coastlines\n"\
		"                           3 = intermittent coastlines\n"\
		"                           4 = primary rivers\n"\
		"                           5 = secondary rivers and canals\n"\
		"                           6 = intermittent rivers\n"\
		"                           7 = international political boundaries\n"\
		"                           8 = national political boundaries (USA - Canada)\n"\
		"     -grid=1,...,4,step,stop\n"\
		"                           Add any combination of grids\n"\
		"                           For a graticule, a step (in deg.) and a grid\n"\
		"                           stop (maximum latitude) can be specified\n"\
		"                           1 = meridians\n"\
		"                           2 = parallels\n"\
		"                           3 = MODLAND L3 grid\n"\
		"                           4 = graticule nodes\n"\
		"\n"\
		"ENVIRONMENT VARIABLES\n"\
		"     ANCPATH      Path for ancillary data files (coastlines, etc.)\n"\
		"     MODISPATH    Path for MODIS products (coastlines, etc.)\n"\
		"     If -depth is used, search for files is recursive from $ANCPATH and $MODISPATH\n"\
		"     Default is current directory (./)\n"\
		"\n"
#define HELPTAIL	"\n"\
		"IN PROGRESS\n"\
		"     An aggregation scheme will be implemented, in addition to subsampling.\n"\
		"     Overlapping products will be handled by a compositing scheme.\n"\
		"     The command line syntax will be more flexible.\n"\
		"\n"\
		"Last update: \n"\
		"Please report problems to Jacques Descloitres (jack@ltpmail.gsfc.nasa.gov)\n"

const char *typename[MAXHDFDATATYPE] = {
  "none", NULL, NULL, "uchar8", "char8", "float32", "float64",				/*  0 to 6  */
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,		/*  7 to 19 */
  "int8", "uint8", "int16", "uint16", "int32", "uint32", "int64", "uint64" };		/* 20 to 27 */

typedef struct {
  int16 int8;
  uint16 uint8;
  int16 int16;
  uint16 uint16;
  int32 int32;
  uint32 uint32;
  float32 float32;
} FILL;

typedef struct {
  char name[MAXFILELENGTH];
  int32 id;
} HDFfile;

typedef struct {
  char *filename, name[MAXNAMELENGTH], **planeID;
  int32 file_id, id, index, num_type, rank, n_attr;
  int32 start[MAX_VAR_DIMS], edges[MAX_VAR_DIMS], dim_sizes[MAX_VAR_DIMS];
  int Nl, Np, Nplanes, maxplanes;
  void *data, *fillvalue;
  unsigned char format;
  double min, max;
  int Nlut;
  char *lut;
} SDS;

typedef struct {
  char *name;
  unsigned char lonsym, latsym, xprojsym, yprojsym, azimuthal, elliptic;
  char *intlon;
  unsigned char Nintlon;
} PROJINFO;

typedef struct {
  unsigned char type;
  double lon_center, lat_center, lon_pole, lat_pole, coslatpole, sinlatpole, range, radius, height, cosphi, sinphi;	/* Order of fields must not be changed because of initialization */
  int nzones;	/* Idem - Do not change order */
  double xcorr;	/* Idem - Do not change order */
  double trunc_lat;
  double stdlat1, stdlat2;
} PROJPARAM;

typedef struct {
  int l1, l2, p1, p2;
} BOUNDINGBOX;

typedef struct {
  HDFfile file;
  SDS sds, meta, lgid;
  unsigned char *senz;
  PROJINFO info;
  double pixsz;
  double row0, col0;
  int row1, row2;
  unsigned char domain, access_mode;
  int Nl_glb, Np_glb;
  PROJPARAM param;
  BOUNDINGBOX bbox, allocbox;
  double lon_target, lat_target;
  int width, height;
} PROJECTION;

typedef struct {
  char name[MAXNAMELENGTH];
  char fullname[MAXNAMELENGTH];
  int32 num_type;
  char string[MAXMETALENGTH];
} METADATA;

typedef struct {
  HDFfile file, geofile;
  SDS sds, lon, lat, sz; /* added sz 07/13 */
  PROJPARAM projparam;
  METADATA meta, lgid;
  int Nl_geo, Np_geo;
  int ih, iv;
  double xpixsz, ypixsz;
  float bounds[4], gringlon[4], gringlat[4];
  char *allmetadata;
  unsigned char mapping, maskMSbit;
  float coloffset, rowoffset, colstep, rowstep;
  int rowmin, rowmax;
  unsigned char adjacent;
} PRODUCT;

typedef struct {
  char *type[NVECTORS];	/* Must be 1st field to allow initialization */
  int offset[NVECTORS];	/* Must be 2nd field to allow initialization */
  unsigned char request, drawline;
  char *path;
} VECTOR;

typedef struct {
  char *type[NGRIDS];	/* Must be 1st field to allow initialization */
  int offset[NGRIDS];	/* Must be 2nd field to allow initialization */
  unsigned char request, separate;
  float step, stop;
  float *points[NGRIDS];
  int bufsize[NGRIDS], Npoints[NGRIDS];
} GRID;

typedef struct {
  char *type[5];	/* Must be 1st field to allow initialization */
  char value[10];
  unsigned char request;
} DAYNIGHTFLAG;

typedef struct {
  unsigned char interpol, depth, gzip, minsenz;
} OPTIONS;


void add_grids(void *data, int32 num_type, PROJECTION proj, GRID grid);
void add_vectors(void *data, int32 num_type, PROJECTION proj, VECTOR vector);
int  aitint(double r, double center_long, double false_east, double false_north);
int  aitfor(double lon, double lat, double *xproj, double *yproj);
int  aitinv(double xproj, double yproj, double *lon, double *lat);
int  alberfor(double lon, double lat, double *xproj, double *yproj);
int  alberforint(double r_maj, double r_min, double lat1, double lat2, double center_lon, double center_lat, double false_east, double false_north);
int  alberinv(double xproj, double yproj, double *lon, double *lat);
int  alberinvint(double r_maj, double r_min, double lat1, double lat2, double center_lon, double center_lat, double false_east, double false_north);
int  azimfor(double lon, double lat, double *xproj, double *yproj);
int  azimforint(double r, double center_lon, double center_lat, double false_east, double false_north);
int  aziminv(double xproj, double yproj, double *lon, double *lat);
int  aziminvint(double r, double center_lon, double center_lat, double false_east, double false_north);
int  alloc_new_array(void **data, int size, int32 num_type, void *fillvalue);
char *capitalize(char *string);
int  check_continuity(double lon1, double lat1, double lon2, double lat2, double xproj1, double xproj2, PROJPARAM *projparam, double pixsz);
int  check_geolocation_consistency(PRODUCT granule);
int  check_memory(unsigned long size);
int  cmp_projections(PROJPARAM projparam1, PROJPARAM projparam2);
void compute_grids(PROJECTION proj, GRID *grid);
int  convert_data(void **data, int size, int32 type1, int32 type2, unsigned char verbose);
int  convert_datatype(SDS *sds, int32 new_num_type);
int  convert_metadata_string(SDS *projmeta, METADATA *metadata);
void copy_value(void *data2, void *data1, unsigned long idx2, unsigned long idx1, int32 num_type);
int  create_2D_SDS(SDS *sds, unsigned char access_mode);
int  create_and_write_1D_SDS(void *data, int32 Nitems, char *sds_name, int32 num_type, int32 file_id);
void draw_cross(void *data, int irow, int jcol, int row1, int row2, int Np, int32 num_type, int halfsize, int offset);
void draw_lonlatline(PROJECTION *proj, double lon1, double lat1, double lon2, double lat2, GRID *grid, int gridno);
void draw_xyline(void *data, int32 num_type, PROJECTION *proj, double xproj1, double yproj1, double xproj2, double yproj2, int offset);
int  eckivfor(double lon, double lat, double *xproj, double *yproj);
int  eckivforint(double r, double center_long, double false_east, double false_north);
int  eckivinv(double xproj, double yproj, double *lon, double *lat);
int  eckivinvint(double r, double center_long, double false_east, double false_north);
void fill_domain(PROJECTION *proj);
void fill_sym_line(PROJECTION *proj, int irow, int jcolmin, int jcolmax, int Np, double row0, double col0, double pixsz, double xmax);
float fudge(int iscan, int georowsperscan, PRODUCT *input);
int  getattribute(int32 id, char *name, void *value);
int  get_fudgefactor(PRODUCT *input, float *fudgefactor1, float *fudgefactor2);
void get_georows(double fractrow, PRODUCT *input, int iscan, int *georow1, int *georow2, double *t);
void get_xybounds(PROJECTION *proj, double *xmin, double *xmax, double *ymin, double *ymax);
int  get_estimbox(PROJECTION proj, PRODUCT input, BOUNDINGBOX *estimHbox);
void get_file_from_inputpointer(PRODUCT input, char *shortname, char *filename);
int  get_geolocation(PRODUCT *input);
int  get_gnomonic_Gring(PRODUCT input, double *xarr, double *yarr, double *lon0, int *gnomoproj);
int  get_lonlat(double xproj, double yproj, double *lon, double *lat, PROJPARAM *projparam);
int  get_metadata(char *metadata, char *object, void *value, int32 *num_type, char *fullname);
void get_min_max(void *data, int size, int32 num_type, double *min, double *max);
void get_new_type(void *data, int size, int32 num_type, double *min, double *max, int32 *new_num_type);
char *get_next_string(char *arg, char *string);
void get_projinfo(PROJINFO *projinfo);
int  get_row_bounds(int irow, PRODUCT *input, double *latrowarr, double *lonrowarr, double *xrowarr, double *yrowarr, PROJPARAM *projparam);
int  get_row_pixel_centers(int irow, PRODUCT *input, double *lat, double *lon);
char *get_SDS_ID(char *arg, SDS *sds);
int  get_SDS_info(SDS *sds);
void get_step_and_offset(int n1, int n2, float *offset, float *step, unsigned char projtype, unsigned char crosstrack);
int  get_tileID(char *allmetadata, PRODUCT *tile);
void get_virtual_row(double *lon1, double *lat1, double *lon2, double *lat2, double t, int ncols, double *lon3, double *lat3);
int  get_xyproj(double lon, double lat, double *xproj, double *yproj, PROJPARAM *projparam);
int  gnomfor(double lon, double lat, double *xproj, double *yproj);
int  gnomforint(double r, double center_lon, double center_lat, double false_east, double false_north);
int  gnominv(double xproj, double yproj, double *lon, double *lat);
int  gnominvint(double r, double center_lon, double center_lat, double false_east, double false_north);
int  gvnspfor(double lon, double lat, double *xproj, double *yproj);
int  gvnspforint(double r, double h, double center_long, double center_lat, double false_east, double false_north);
int  gvnspinv(double xproj, double yproj, double *lon, double *lat);
int  gvnspinvint(double r, double h, double center_long, double center_lat, double false_east, double false_north);
int  goodfor(double lon, double lat, double *xproj, double *yproj);
int  goodfor2(double lon, double lat, double *xproj, double *yproj);
int  goodforint(double r, double false_east, double false_north);
int  goodforint2(double r, double false_east, double false_north);
int  goodinv(double xproj, double yproj, double *lon, double *lat);
int  goodinv2(double xproj, double yproj, double *lon, double *lat);
int  goodinvint(double r, double false_east, double false_north);
int  goodinvint2(double r, double false_east, double false_north);
int  hamfor(double lon, double lat, double *xproj, double *yproj);
int  hamforint(double r, double center_long, double false_east, double false_north);
int  haminv(double xproj, double yproj, double *lon, double *lat);
int  haminvint(double r, double center_long, double false_east, double false_north);
int  imolwfor(double lon, double lat, double *xproj, double *yproj);
int  imolwforint(double r, double false_east, double false_north);
int  imolwinv(double xproj, double yproj, double *lon, double *lat);
int  imolwinvint(double r, double false_east, double false_north);
int  init_forward_projection(unsigned char projtype, double height, double stdlat1, double stdlat2);
void init_projection(PROJECTION *proj);
int  inside_Gring(double lon, double lat, double *xarr, double *yarr, double lon0, int gnomoproj);
int  intersinfor1(double lon, double lat, double *x, double *y);
int  intersinfor2(double lon, double lat, double *x, double *y);
int  intersininv1(double x, double y, double *lon, double *lat);
int  intersininv2(double x, double y, double *lon, double *lat);
int  intersinint1(double r, double false_east, double false_north);
int  intersinint2(double r, double false_east, double false_north);
int  isinusfor(double lon, double lat, double *x, double *y);
int  isinusinv(double x, double y, double *lon, double *lat, int aggfactor);
int  isinusint(double r, double center_long, double false_east, double false_north, int nzones);
int  lamazfor(double lon, double lat, double *xproj, double *yproj);
int  lamazforint(double r_maj, double center_lon, double center_lat, double false_east, double false_north);
int  lamazinv(double xproj, double yproj, double *lon, double *lat);
int  lamazinvint(double r, double center_lon, double center_lat, double false_east, double false_north);
void make_fill_SDS(int32 file_id, char *filename, PROJECTION proj, GRID grid, VECTOR vector);
int  mask_production_date(char *filename);
int  molwfor(double lon, double lat, double *xproj, double *yproj);
int  molwforint(double r, double center_long, double false_east, double false_north);
int  molwinv(double xproj, double yproj, double *lon, double *lat);
int  molwinvint(double r, double center_long, double false_east, double false_north);
void nearest_pixel_interpol(PROJECTION *proj, PRODUCT input, BOUNDINGBOX *bbox, BOUNDINGBOX *estimbox, int margin);
int  open_file(HDFfile *file);
int  orthfor(double lon, double lat, double *xproj, double *yproj);
int  orthforint(double r_maj, double center_lon, double center_lat, double false_east, double false_north);
int  orthinv(double xproj, double yproj, double *lon, double *lat);
int  orthinvint(double r_maj, double center_lon, double center_lat, double false_east, double false_north);
void outline_domain(void *data, int32 num_type, PROJECTION proj);
void project_pixel_centers(PROJECTION *proj, PRODUCT input, BOUNDINGBOX *bbox, BOUNDINGBOX *estimbox, int margin);
int  read_and_convert_line(PRODUCT *input, int firstline, int nlines, PROJECTION *proj);
int  read_metadata(int32 sd_id_in, char **allmetadata);
int  read_SDS(SDS *sds, int firstline, int nlines);
void resample_row_geolocation(unsigned char output, double *oldlon, double *oldlat, PRODUCT *input, double *newlon, double *newlat);
int  resize(PROJECTION *proj, BOUNDINGBOX *bbox);
int  resize_data(void **data, int newNl, int newNp, int curNl, int curNp, int Hoffset, int Voffset, int32 num_type, unsigned char subset, void *fillvalue);
int  resize_SDS(SDS *sds, int Nl_sub, int Np_sub, int irow1, int jcol1, PROJECTION proj, unsigned char subset);
int  robfor(double lon, double lat, double *xproj, double *yproj);
int  robforint(double r, double center_long, double false_east, double false_north);
int  robinv(double xproj, double yproj, double *lon, double *lat);
int  robinvint(double r, double center_long, double false_east, double false_north);
int  search_file(char *filename, char *path);
void setcolor(uint8 *palette, int index, unsigned char red, unsigned char green, unsigned char blue);
void setfillvalue(void *data, int idx0, int size, int32 num_type, int offset);
void set_meta_value(void *data, int idx, int32 data_num_type, void *meta, int32 meta_num_type);
void set_proj_dimensions(PROJECTION *proj, int *Nrows, int *Ncols);
int  sterfor(double lon, double lat, double *xproj, double *yproj);
int  sterforint(double r_maj, double center_lon, double center_lat, double false_east, double false_north);
int  sterinv(double xproj, double yproj, double *lon, double *lat);
int  sterinvint(double r_maj, double center_lon, double center_lat, double false_east, double false_north);
int  string2lutindex(SDS *projmeta, char *string);
int  substitute(char *filename, char *prefix1, char *prefix2);
char testfillvalue(void *data, int idx, int32 num_type, int offset, void *fillvalue);
void update_box(BOUNDINGBOX *bbox, int irow, int jcol);
void update_grid(GRID *grid, int gridno, double xproj1, double yproj1, double xproj2, double yproj2);
int  update_meta_field(SDS *projmeta, PROJECTION *proj, PRODUCT input, METADATA metadata);
int  update_output_pixel(int irow, int jcol, PROJECTION *proj, void *data, PRODUCT *input, int inputidx, int inputrow, int inputcol, BOUNDINGBOX *bbox, BOUNDINGBOX *estimbox, int margin);
void update_projection(PROJECTION *proj, PRODUCT input);
void write_file_attributes(char *cmd_line, int32 file_id);
void write_grid_SDS(int32 file_id, GRID grid, PROJECTION proj);
int  write_SDS(SDS sds, int firstline, int nlines, unsigned char gzip);
void write_SDS_attributes(SDS sds, PROJECTION proj, GRID grid, VECTOR vector);

extern int errno;

PROJINFO projinfo[NUMBEROFPROJ];
char *MODISPATH;
char *VECTORPATH;
unsigned char MOD03=FALSE;
unsigned char FUDGE=FALSE;
unsigned char FILLBURN=FALSE;
unsigned char FILLMISSINGDATA=FALSE;
unsigned char SNOW=FALSE;
unsigned char SNOW_C6=FALSE; /* added SS 05/09 */
unsigned char FIRE=FALSE;
unsigned char COUNT=FALSE;
unsigned char MOD09CRSQA=FALSE;
unsigned char MOD14CRS=FALSE;
unsigned char WATER=FALSE; /* added SS 02/18/2020 for the new MODWATER flood water ratio product */
unsigned char LGID=FALSE;
unsigned char CLEAN=FALSE;
unsigned char NODATA=FALSE;
unsigned char L1B=FALSE;
int  MAGNIFY=FALSE;
int  MAXMEMORY=DEFMAXMEMORY;
float FUDGEFACTOR1=-1, FUDGEFACTOR2=-1;
PROJPARAM isinusparam;
OPTIONS options;
FILL fill;

/* Nested Integerized Sinusoidal (prior to day ???) */
const PROJPARAM isinusparam1 = {INTEGERSIN, 0., 0., 0., 90., 0., 1., 0.,  0., 0., 1., 0., 21600, 1.};
/* Unnested Integerized Sinusoidal (from day ???) */
const PROJPARAM isinusparam2 = {INTEGERSIN, 0., 0., 0., 90., 0., 1., 0.,  0., 0., 1., 0., 86400, 1.};
PROJPARAM aitoffparam = {AITOFF,     0., 0., 0., 90., 0., 1., 0.,  0., 0., 1., 0., 1, 1.};

const PROJPARAM geogparam = {GEOGRAPHIC,     0., 0., 0., 90., 0., 1., 0.,  0., 0., 1., 0., 1, 1.};

int main(int argc, char *argv[])
{
char  string[MAXMETALENGTH], argstring[MAXMETALENGTH], sdsname[MAXNAMELENGTH];
char  *ptr, *arg, *cmd_line, opt[20], inputpath[MAXFILELENGTH];
int k, kmin, kmax, iarg, projtype, access_mode;
float phi;
PROJECTION proj;
PRODUCT input;
VECTOR vector = { VECTORTYPES,
                  VECTOROFFSETS };
GRID grid = { GRIDTYPES,
              GRIDOFFSETS };
DAYNIGHTFLAG DNflag = { DNFLAGTYPE };
METADATA lgidcopy;
int tileid; /* 02/14/2020 */

   get_projinfo(projinfo);

   if ( argc > 1 && (strcmp(argv[1], "-help") == 0 || strcmp(argv[1], "-h") == 0) ) {		/* Help option specified */
     fprintf(stderr, "%s", FULLHELP);
     fprintf(stderr, "\nSUPPORTED PROJECTIONS\n");
     for (k=1; k<NUMBEROFPROJ; k++) fprintf(stderr, "     %2d. %s\n", k, projinfo[k].name);
     fprintf(stderr, "%s", HELPTAIL);
     exit(-1);
   }

   printf("\n");
   if ((VECTORPATH = getenv(ANCPATHVAR)) != NULL) {				/* Path for ancillary data */
       printf("Path for ancillary data: %s\n", VECTORPATH);
       while ( (ptr = strchr(VECTORPATH, 58)) != NULL ) *ptr = 32;		/* Replace colons by spaces */
   } else {
       printf("Found no path for ancillary data ($%s variable). Current directory is used instead.\n", ANCPATHVAR);
       VECTORPATH = strdup(".");
   }
   if ((MODISPATH = getenv("MODISPATH")) != NULL)	{		/* Path for MODIS files */
       printf("Path for MODIS products: %s\n", MODISPATH);
       while ( (ptr = strchr(MODISPATH, 58)) != NULL ) *ptr = 32;	/* Replace colons by spaces */
   } else {
       printf("Found no path for MODIS products ($MODISPATH variable). Current directory is used instead.\n");
       MODISPATH = strdup(".");
   }
   printf("\n");

   iarg = 1;		/* Argument counter - Used only in case of command-line arguments */
   arg = argstring;	/* Variable arg is just a pointer - Must point on allocated memory in case of scanf */
   cmd_line = (char *) malloc((strlen(argv[0]) + 1 + 1) * sizeof(char));
   sprintf(cmd_line, "\n%s", argv[0]);

   DNflag.request = CHAR_MAX;
   vector.request = 0;
   grid.request = 0;
   grid.separate = FALSE;
   for (k=0; k<NGRIDS; k++) {
     grid.Npoints[k] = 0;
     grid.bufsize[k] = 0;
     grid.points[k] = NULL;
   }
   proj.param.type = 0;
   proj.file.name[0] = '\0';
   vector.path = VECTORPATH;
   vector.drawline = FALSE;
   inputpath[0] = '\0';
   input.file.name[0] = '\0';
   input.sds.name[0] = '\0';
   input.sz.name[0] = '\0'; /* added 07/13 */
   input.sds.filename = input.file.name;
   input.sz.filename = input.file.name; /* added 07/13 */
   input.sds.Nplanes = 0;
   input.sds.planeID = NULL;
   input.allmetadata = NULL;
   input.maskMSbit = FALSE;
   input.mapping = FORWARD;
   input.rowmin = 0;
   input.rowmax = USHRT_MAX;
   input.meta.name[0] = '\0';
   input.adjacent = FALSE;
   strcpy(input.lgid.name, "LOCALGRANULEID");
   strcpy(input.lgid.fullname, "LOCALGRANULEID");
   proj.access_mode = CREATE;
   proj.domain = EXTBOX;
   proj.param.lon_center = proj.param.lat_center = proj.param.range = PROJUNDEF;
   proj.param.lon_pole = proj.param.lat_pole = PROJUNDEF;
   proj.param.height = DEFHEIGHT;
   proj.param.nzones = 0;
   proj.param.cosphi = 1;
   proj.param.sinphi = 0;
   proj.param.stdlat1 = PROJUNDEF;
   proj.param.stdlat2 = PROJUNDEF;
   proj.param.xcorr = 1;
   proj.lon_target = proj.lat_target = PROJUNDEF;
   strcpy(proj.lgid.name, "LOCALGRANULEID");
   isinusparam = isinusparam2;		/* Default is unnested IS projection */
   options.interpol = NEAREST;
   options.depth = FALSE;
   options.gzip = FALSE;
   options.minsenz = FALSE;
   fill.int8 = FILL_INT8;
   fill.uint8 = FILL_UINT8;
   fill.int16 = FILL_INT16;
   fill.uint16 = FILL_UINT16;
   fill.int32 = FILL_INT32;
   fill.uint32 = FILL_UINT32;
   fill.float32 = FILL_FLOAT32;


   while ( (argc > 1  && iarg < argc) ||		/* If there is at least one argument, then parse all the command-line arguments */
           (argc == 1 && scanf("%s", arg) != EOF) ) {	/* Else parse stdin instead */

     if (argc > 1) arg = argv[iarg++];			/* If command-line arguments, points on iarg-th argument and increment iarg */

     cmd_line = (char *) realloc(cmd_line, (strlen(cmd_line) + strlen(arg) + 4) * sizeof(char));
     strcat(cmd_line, " ");
     if (! strchr(arg, 32))		/* Argument does not contains space characters */
       strcat(cmd_line, arg);
     else {
       if ( ! (ptr = strchr(arg, 61)) )
         sprintf(cmd_line, "\"%s\"", arg);
       else {
         strcat(cmd_line, arg);
         sprintf(cmd_line + strlen(cmd_line) - strlen(arg) + (ptr - arg) + 1, "\"%s\"", ptr + 1);
       }
     }

     strcpy(opt, "-depth");
     if ( strcmp(arg, opt) == 0 ) {
       options.depth = TRUE;
       printf("Recursive file search activated.\n\n");
       continue;
     }

     strcpy(opt, "-fast");
     if ( strcmp(arg, opt) == 0 ) {
       options.interpol = NONE;
       printf("Fast projection mode requested.\n\n");
       continue;
     }

     strcpy(opt, "-gzip");
     if ( strcmp(arg, opt) == 0 ) {
       options.gzip = TRUE;
       printf("Gzip compression requested for output projection.\n\n");
       continue;
     }

     strcpy(opt, "-clean");		/* In case of overlap we may prefer valid value to fill value */
     if ( strcmp(arg, opt) == 0 ) {
       CLEAN = TRUE;
       printf("Output projection requested to be cleaned from fill value.\n\n");
       continue;
     }

     strcpy(opt, "-nesting");
     if ( strcmp(arg, opt) == 0 ) {
       isinusparam = isinusparam1;
       printf("Nesting requested for Integerized Sinusoidal projection.\n\n");
       continue;
     }

     strcpy(opt, "-nogap");
     if ( strcmp(arg, opt) == 0 ) {
       FUDGE = TRUE;
       printf("Interscan gap will be filled by extrapolation.\n\n");
       continue;
     }

     strcpy(opt, "-MOD03");
     if ( strcmp(arg, opt) == 0 ) {
       MOD03 = TRUE;
       printf("Use of MOD03 file requested for geolocation calculations.\n\n");
       continue;
     }

     strcpy(opt, "-snow");
     if ( strcmp(arg, opt) == 0 ) {
       SNOW = TRUE;
       printf("Snow-type compositing scheme requested.\n\n");
       continue;
     }
     
     /* added SS 05/09 for C6 snow fraction */
     strcpy(opt, "-snow_c6");
     if ( strcmp(arg, opt) == 0 ) {
       SNOW_C6 = TRUE;
       printf("Snow-C6-type compositing scheme requested.\n\n");
       continue;
     }
     
     /* Added 07/13 to handle solar correction for L1B data */
     strcpy(opt, "-l1b");
     if ( strcmp(arg, opt) == 0 ) {
       L1B = TRUE;
       printf("L1B-type compositing scheme requested.\n\n");
       continue;
     }

     strcpy(opt, "-mod09crsqa");
     if ( strcmp(arg, opt) == 0 ) {
       MOD09CRSQA = TRUE;
       printf("MOD09CRS-QA-type compositing scheme requested.\n\n");
       continue;
     }

     strcpy(opt, "-mod14crs");
     if ( strcmp(arg, opt) == 0 ) {
       MOD14CRS = TRUE;
       printf("MOD14CRS-type compositing scheme requested.\n\n");
       continue;
     }

     strcpy(opt, "-fire");
     if ( strcmp(arg, opt) == 0 ) {
       FIRE = TRUE;
       printf("Fire-type compositing scheme requested.\n\n");
       continue;
     }
       
     /* Added for MODWATER by SS-02/18/2020 */
     strcpy(opt, "-water");
     if ( strcmp(arg, opt) == 0 ) {
       WATER = TRUE;
       printf("modwater WaterRatio-type compositing scheme requested.\n\n");
       continue;
     }
     
     strcpy(opt, "-count");
     if ( strcmp(arg, opt) == 0 ) {
       COUNT = TRUE;
       printf("Count-type compositing scheme requested.\n\n");
       continue;
     }

     strcpy(opt, "-minsenz");
     if ( strcmp(arg, opt) == 0 ) {
       options.minsenz = TRUE;
       printf("Minimum view zenith angle compositing criterion requested.\n\n");
       continue;
     }

     strcpy(opt, "-bandaid");
     if ( strcmp(arg, opt) == 0 ) {
       FILLMISSINGDATA = TRUE;
       printf("Input data will be corrected to fill missing packets (by duplication).\n\n");
       continue;
     }

     strcpy(opt, "-fillburn");
     if ( strcmp(arg, opt) == 0 ) {
       FILLBURN = TRUE;
       printf("Coastlines requested on top of fill value only.\n\n");
       continue;
     }

     strcpy(opt, "-separate");
     if ( strcmp(arg, opt) == 0 ) {
       grid.separate = TRUE;
       printf("Separation of grid and data requested.\n\n");
       continue;
     }

     strcpy(opt, "-nodata");
     if ( strcmp(arg, opt) == 0 ) {
       NODATA = TRUE;
       printf("Output requested for empty projection\n\n");
       continue;
     }

     strcpy(opt, "-append");
     if ( strcmp(arg, opt) == 0 ) {
       proj.access_mode = APPEND;
       printf("Output file access mode: Append\n\n");
       continue;
     }

     strcpy(opt, "-replace");
     if ( strcmp(arg, opt) == 0 ) {
       proj.access_mode = REPLACE;
       printf("Output file access mode: Replace\n\n");
       continue;
     }

     strcpy(opt, "-global");
     if ( strcmp(arg, opt) == 0 ) {
       proj.domain = GLOBAL;
       printf("Output domain selected: Global\n\n");
       continue;
     }

     strcpy(opt, "-adjacent");
     if ( strcmp(arg, opt) == 0 ) {
       input.adjacent = TRUE;
       printf("Next input granule adjacent to previous one\n\n");
       continue;
     }

     strcpy(opt, "-15bit");		/* In L1B most significant bit is set if data is faulty */
     if ( strcmp(arg, opt) == 0 ) {
       input.maskMSbit = TRUE;
       printf("Most significant bit of 16-bit data will be masked.\n\n");
       continue;
     }

     strcpy(opt, "-line");		/* Force to draw vectors as lines (default is dots at vector limits) */
     if ( strcmp(arg, opt) == 0 ) {
       vector.drawline = TRUE;
       printf("Line requested between vector limits.\n\n");
       continue;
     }

     strcpy(opt, "-lgid");		/* Create an SDS representing local granule ID */
     if ( strcmp(arg, opt) == 0 ) {
       LGID = TRUE;
       printf("Local Granule ID SDS requested.\n\n");
       continue;
     }

     strcpy(opt, "-sds=");					/* New SDS name specifier */
     if ( strstr(arg, opt) == arg ) {
       get_SDS_ID(arg + strlen(opt), &input.sds);	/* Parse argument, starting just after the = sign */
       input.meta.name[0] = '\0';
       continue;					/* Skip to next argument */
     }

     /* Added 07/13 to handle optional solar zenith dataset */
     strcpy(opt, "-sz=");					/* SolZen SDS name specifier */
     if ( strstr(arg, opt) == arg ) {
       get_SDS_ID(arg + strlen(opt), &input.sz);	/* Parse argument, starting just after the = sign */
       continue;					/* Skip to next argument */
     }
     
     strcpy(opt, "-meta=");					/* New metadata specifier */
     if ( strstr(arg, opt) == arg ) {
       strcpy(input.meta.name, arg + strlen(opt));	/* Copy argument, starting just after the = sign */
       printf("Requested metadata: %s.\n\n", input.meta.name);
       input.sds.name[0] = '\0';
       continue;					/* Skip to next argument */
     }

     strcpy(opt, "-path=");				/* Input path specifier */
     if ( strstr(arg, opt) == arg ) {
       strcpy(inputpath, arg + strlen(opt));		/* Copy argument, starting just after the = sign */
       printf("Requested input path: %s.\n\n", inputpath);
       continue;
     }

     strcpy(opt, "-of=");					/* Output file specifier */
     if ( strstr(arg, opt) == arg ) {
       strcpy(proj.file.name, arg + strlen(opt));	/* Copy argument, starting just after the = sign */
       printf("Requested output file: %s.\n\n", proj.file.name);
       continue;					/* Skip to next argument */
     }

     strcpy(opt, "-DN=");
     if ( strstr(arg, opt) == arg ) {
       arg += strlen(opt);
       while ( ptr = strchr(arg, 44) ) *ptr = 32;	/* Replace commas by spaces (if any) */
       if (! *(ptr = arg)) {
         printf("Unable to identify a Day/Night flag specifier. Request ignored.\n\n");
         continue;
       }
       while ( sscanf(ptr, "%d", &kmin) == 1 ) {	/* Get next value in argument string */
         if ( sscanf(ptr, "%d-%d", &kmin,&kmax) != 2 ) kmax = kmin;
         kmin--;
         kmax--;
         for (k=kmin; k<=kmax; k++) {
           if (k >= 0 && k <= 4) {			/* Flag no. must be between 0 and 2 */
                  if (DNflag.request == CHAR_MAX) DNflag.request = 0;	/* Request was initialized to "all" */
                  DNflag.request |= 1 << k;				/* DNflag.request = 2**k */
                  printf("Day/Night flag request: %s.\n", DNflag.type[k]);
           } else printf("%s%d is not a valid Day/Night flag request. Ignored.\n", opt, k + 1);
           if ( ptr = strchr(ptr, 32) ) ptr++;		/* Repoints ptr just after next space character */
           else ptr = strchr(arg, '\0');		/* If no more field, then points on end of string */
         }
       }
       printf("\n");
       continue;					/* Skip to next argument */
     }

     strcpy(opt, "-map=");					/* Obsolete vector overlay specifier */
     if ( strstr(arg, opt) == arg ) {
       printf("-map is an obsolete vector overlay specifier. Use -vector instead.\n");
       exit(-1);
     }

     strcpy(opt, "-vector=");				/* Vector overlay specifier */
     if ( strstr(arg, opt) == arg ) {
       arg += strlen(opt);
       while ( ptr = strchr(arg, 44) ) *ptr = 32;	/* Replace commas by spaces (if any) */
       ptr = arg;
       while ( sscanf(ptr, "%d", &kmin) == 1 ) {	/* Get next value in argument string */
         if ( sscanf(ptr, "%d-%d", &kmin,&kmax) != 2 ) kmax = kmin;
         kmin--;
         kmax--;
         for (k=kmin; k<=kmax; k++) {
           if (k >= 0 && k <= NVECTORS - 1) {			/* Vector no. must be between 0 and NVECTORS-1 */
                  vector.request |= 1 << k;				
                  printf("Vector overlay request: %s.\n", vector.type[k]);
           } else printf("%s%d is not a valid vector overlay request. Ignored.\n", opt, k + 1);
           if ( ptr = strchr(ptr, 32) ) ptr++;		/* Repoints ptr just after next space character */
           else ptr = strchr(arg, '\0');		/* If no more field, then points on end of string */
         }
       }
       printf("\n");
       continue;					/* Skip to next argument */
     }

     strcpy(opt, "-grid=");					/* Grid request specifier */
     if ( strstr(arg, opt) == arg ) {
       arg += strlen(opt);
       if (! (ptr = get_next_string(arg, string))) {
         printf("Unable to identify a valid grid specifier. Request ignored.\n\n");
         continue;
       }
       grid.step = -1;
       grid.stop = 90;
       while (ptr) {
         if (! strchr(string, 46)) {
           if (sscanf(string, "%d", &k) == 1) {
             k--;
             if (k >= 0  &&  k < NGRIDS) {
                 grid.request |= 1 << k;
                 printf("Grid request: %s.\n", grid.type[k]);
             } else {
                 printf("%s%d is not a valid grid request. Ignored.\n", opt, k + 1);
                 printf("Use %d. if you want to specify a step or a grid stop.\n", k + 1);
             }
           }
         } else if ( grid.step < 0 &&
                     sscanf(string, "%g", &grid.step) == 1 ) {
             if (grid.step >= 1.)	/* Minimum step 1 deg. */
               printf("Grid step requested: %g degrees.\n", grid.step);
             else {
               printf("Grid step (%g deg.) is out of valid range.\n", grid.step);
               grid.step = -1;
             }
         } else if (sscanf(string, "%g", &grid.stop) == 1) {
             if (grid.stop >= grid.step  &&  grid.stop <= 90)
               printf("Grid stop requested: latitude %g degrees.\n", grid.stop);
             else {
               printf("Grid stop (latitude %g deg.) is out of valid range.\n", grid.stop);
               grid.stop = 90;
             }
         }
         if (! *ptr) break;
         ptr = get_next_string(ptr + 1, string);
       }
       if (grid.step < 0) {
         grid.step = 10.;
         printf("No valid grid step identified. %g degrees is used.\n", grid.step);
       }
       printf("\n");
       continue;					/* Skip to next argument */
     }

     strcpy(opt, "-res=");					/* Obsolete resolution specifier */
     if ( strstr(arg, opt) == arg ) {
       printf("-res is an obsolete resolution specifier. Use -pixsz instead.\n");
       exit(-1);
     }

     strcpy(opt, "-pixsz=");					/* Output pixel size specifier */
     if ( strstr(arg, opt) == arg ) {
       arg += strlen(opt);
       if (input.file.name[0])
         printf("Output pixel size must be specified before first input file. Ignored.\n\n");
       else
         if (sscanf(arg, "%lg", &proj.pixsz) == 1) {		/* Get argument, starting just after the = sign */
           if (! strstr(arg, "deg")) proj.pixsz /= DEG2KM;
           printf("%g km (%g deg.) requested for output projection.\n",
                  proj.pixsz * DEG2KM, proj.pixsz);
           if ( proj.pixsz * DEG2KM < MINPIXSZ  ||
                proj.pixsz * DEG2KM > MAXPIXSZ ) {
             printf("Requested resolution is out of bounds. Ignored.\n");
             proj.pixsz = 0;
           }
         }
       printf("\n");
       continue;
     }

     strcpy(opt, "-proj=");					/* Output projection specifier */
     if ( strstr(arg, opt) == arg ) {
       projtype = 0;
       if (input.file.name[0])
         printf("Output projection must be specified before first file. Ignored.\n\n");
       else
         if (sscanf(arg + strlen(opt), "%d", &projtype) == 1 &&	/* Get argument, starting just after the = sign */
             projtype > 0 && projtype < NUMBEROFPROJ )		/* projtype is in valid range */
           printf("Output projection selected: %s\n", projinfo[proj.param.type = projtype].name);
         else {
           printf("Invalid output projection specifier");
           if (projtype != 0) printf(" \"%d\"", projtype);
           printf(".\n");
         }
       printf("\n");
       continue;
     }

     strcpy(opt, "-box=");				/* Subwindow for output projection */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%d,%d,%d,%d",
                   &proj.bbox.p1, &proj.bbox.l1, &proj.width, &proj.height) == 4 ) {
         proj.bbox.p2 = proj.bbox.p1 + proj.width - 1;
         proj.bbox.l2 = proj.bbox.l1 + proj.height - 1;
         printf( "Output subwindow selected: %dx%d (column %d, row %d) to (column %d, row %d)\n",
                 proj.width, proj.height, proj.bbox.p1, proj.bbox.l1, proj.bbox.p2, proj.bbox.l2 );
         if ( proj.bbox.l1 >= 0  &&
              proj.bbox.p1 >= 0  &&
              proj.width > 0  &&
              proj.height > 0 )
           proj.domain = USERBOX;
         else
           printf("Bad specification for output subwindow. Request ignored.\n");
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-targetbox=");				/* Center and dimensions of output subwindow */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg,%lg,%d,%d", &proj.lon_target, &proj.lat_target, &proj.width, &proj.height) != 4 )
         printf("Unable to parse valid target coordinates. Request ignored.\n");
       else {
         printf( "Output subwindow selected: %dx%d\n", proj.width, proj.height);
         if (fabs(proj.lon_target) <= 180)
           printf("Target longitude selected:  %g deg.\n", proj.lon_target);
         else {
           printf("Invalid target longitude (%g deg.). Request ignored.\n", proj.lon_target);
           proj.lon_target = proj.lat_target = PROJUNDEF;
         }
         if (fabs(proj.lat_target) <= 90)
           printf("Target latitude selected: %g deg.\n", proj.lat_target);
         else {
           printf("Invalid target latitude (%g deg.). Request ignored.\n", proj.lat_target);
           proj.lon_target = proj.lat_target = PROJUNDEF;
         }
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-rows=");				/* First and last input rows to use */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%d,%d", &input.rowmin, &input.rowmax) == 2 ) {
         printf( "Input rows selected: %d-%d\n", input.rowmin, input.rowmax);
         if ( input.rowmin > input.rowmax) {
           input.rowmin = 0;
           input.rowmax = USHRT_MAX;
           printf("Bad specification for input rows to consider. Request ignored.\n");
         }
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-pole=");				/* North pole location for the output projection */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg,%lg", &proj.param.lon_pole, &proj.param.lat_pole) != 0 ) {
         if (proj.param.lon_pole != PROJUNDEF) {
           if (fabs(proj.param.lon_pole) <= 180)
             printf("North pole longitude selected for output projection: %g deg.\n", proj.param.lon_pole);
           else {
             printf("Invalid North pole longitude (%g deg.). Request ignored.\n", proj.param.lon_pole);
             proj.param.lon_pole = PROJUNDEF;
           }
         }
         if (proj.param.lat_pole != PROJUNDEF) {
           if ( proj.param.lat_pole >= 0  &&
                proj.param.lat_pole <= 90 )
             printf("North pole latitude selected for output projection: %g deg.\n", proj.param.lat_pole);
           else {
             printf("Invalid North pole latitude (%g deg.). Request ignored.\n", proj.param.lat_pole);
             proj.param.lat_pole = PROJUNDEF;
           }
         }
       } else printf("Unable to parse a valid North pole. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-rotation=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%g", &phi) != 1 )
         printf("Cannot parse a rotation angle.\n");
       else {
         printf("Rotation of output projection: %g deg.\n", phi);
         proj.param.cosphi = cos(phi * DEG2RAD);
         proj.param.sinphi = sin(phi * DEG2RAD);
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-std=");				/* Standard parallels for Albers projection */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg,%lg", &proj.param.stdlat1, &proj.param.stdlat2) != 2 )
         if ( sscanf(arg + strlen(opt), "%lg", &proj.param.stdlat1) != 1 )
           printf("Unable to parse standard parallels. Request ignored.\n");
         else {
           printf("Standard parallel: %g deg.\n", proj.param.stdlat1);
           proj.param.stdlat1 *= DEG2RAD;
         }
       else {
         printf("Standard parallels: %g deg. and %g deg.\n", proj.param.stdlat1, proj.param.stdlat2);
         proj.param.stdlat1 *= DEG2RAD;
         proj.param.stdlat2 *= DEG2RAD;
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-center=");				/* Center of output projection */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg,%lg", &proj.param.lon_center, &proj.param.lat_center) != 0 ) {
         if (proj.param.lon_center != PROJUNDEF) {
           if (fabs(proj.param.lon_center) <= 180)
             printf("Center longitude selected for output projection: %g deg.\n", proj.param.lon_center);
           else {
             printf("Invalid center longitude (%g deg.). Request ignored.\n", proj.param.lon_center);
             proj.param.lon_center = PROJUNDEF;
           }
         }
         if (proj.param.lat_center != PROJUNDEF) {
           if (fabs(proj.param.lat_center) <= 90)
             printf("Center latitude selected for output projection: %g deg.\n", proj.param.lat_center);
           else {
             printf("Invalid center latitude (%g deg.). Request ignored.\n", proj.param.lat_center);
             proj.param.lat_center = PROJUNDEF;
           }
         }
       } else printf("Unable to parse a valid projection center. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-nz=");				/* Number of zones for output projection (if applicable) */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%d", &proj.param.nzones) == 1 ) {
         if ( proj.param.nzones > 0  &&  proj.param.nzones < MAXNZONES )
           printf("Selected number of zones for projection center: %d\n", proj.param.nzones);
         else {
           printf("Invalid number of zones (%d). Request ignored.\n", proj.param.nzones);
           proj.param.nzones = 0;
         }
       } else printf("Unable to parse a valid number of zones. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-xcorr=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg", &proj.param.xcorr) == 1 )
         printf("Stretch factor: %lg\n", proj.param.xcorr);
       else {
         printf("Unable to parse a valid stretch factor. Request ignored.\n");
         proj.param.xcorr = 1;
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-range=");				/* Projection range from center (if applicable) */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg", &proj.param.range) == 1 ) {
         if ( proj.param.range > 0  &&  proj.param.range <= 90 )
           printf("Selected range from projection center: %g deg.\n", proj.param.range);
         else {
           printf("Invalid range from center (%g deg.). Request ignored.\n", proj.param.range);
           proj.param.range = PROJUNDEF;
         }
       } else printf("Unable to parse a valid range from projection center. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-height=");			/* Observation height (if applicable) */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%lg", &proj.param.height) == 1 ) {
         if (proj.param.height > 0)
           printf("Observation height selected for output projection: %g km.\n", proj.param.height);
         else {
           printf("Requested observation height (%g km) must be greater than 0. Default (%g km) is used.\n",
                  proj.param.height, DEFHEIGHT);
           proj.param.height = DEFHEIGHT;
         }
       } else printf("Unable to parse a valid observation height. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-maxmem=");			/* Maximum memory to allocate */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%d", &MAXMEMORY) == 1 ) {
         if (MAXMEMORY > 0)
           printf("Requested maxmimum memory to allocate: %d Mbytes\n", MAXMEMORY);
         else {
           printf("Maxmimum memory to allocate must be greater than 0. Default (%d Mbytes) is used.\n", DEFMAXMEMORY);
           MAXMEMORY = DEFMAXMEMORY;
         }
       } else printf("Unable to parse maxmimum memory to allocate. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-fillint8=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%hd", &fill.int8) == 1 )
         printf("Requested fill value for INT8 data type: %d\n", fill.int8);
       else
         printf("Unable to parse fill value for INT8 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-filluint8=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%hu", &fill.uint8) == 1 )
         printf("Requested fill value for UINT8 data type: %d\n", fill.uint8);
       else
         printf("Unable to parse fill value for UINT8 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-fillint16=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%hd", &fill.int16) == 1 )
         printf("Requested fill value for INT16 data type: %d\n", fill.int16);
       else
         printf("Unable to parse fill value for INT16 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-filluint16=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%hu", &fill.uint16) == 1 )
         printf("Requested fill value for UINT16 data type: %d\n", fill.uint16);
       else
         printf("Unable to parse fill value for UINT16 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-fillint32=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%d", &fill.int32) == 1 )
         printf("Requested fill value for INT32 data type: %d\n", fill.int32);
       else
         printf("Unable to parse fill value for INT32 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-filluint32=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%u", &fill.uint32) == 1 )
         printf("Requested fill value for UINT32 data type: %d\n", fill.uint32);
       else
         printf("Unable to parse fill value for UINT32 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-fillfloat32=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%g", &fill.float32) == 1 )
         printf("Requested fill value for FLOAT32 data type: %g\n", fill.float32);
       else
         printf("Unable to parse fill value for FLOAT32 data type. Request ignored.\n");
       printf("\n");
       continue;
     }

     strcpy(opt, "-fudge=");			/* Fudge factor to fill interscan gap */
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%g,%g", &FUDGEFACTOR1, &FUDGEFACTOR2) == 2 ) {
         if (FUDGEFACTOR1 >= 0  &&  FUDGEFACTOR1 < 1) {
             printf("Requested fudge factor to fill gap between 1st and 2nd scan: %g\n", FUDGEFACTOR1);
             FUDGE = TRUE;
         } else {
             printf("Fudge factor must be in range [0,1]. Request ignored.\n");
             FUDGEFACTOR1 = 0;
         }
         if (FUDGEFACTOR2 >= 0  &&  FUDGEFACTOR2 < 1) {
             printf("Requested fudge factor to fill gap between 2nd and 3rd scan: %g\n", FUDGEFACTOR2);
             FUDGE = TRUE;
         } else {
             printf("Fudge factor must be in range [0,1]. Request ignored.\n");
             FUDGEFACTOR2 = 0;
         }
       }
       printf("\n");
       continue;
     }

     strcpy(opt, "-magnify=");
     if ( strstr(arg, opt) == arg ) {
       if ( sscanf(arg + strlen(opt), "%d", &MAGNIFY) == 1  &&
            MAGNIFY > 0 )
         printf("Magnification request for fire pixels: %d pixels around actual location\n", MAGNIFY);
       else {
         printf("Unable to parse a valid magnification factor. Request ignored.\n");
         MAGNIFY = 0;
       }
       printf("\n");
       continue;
     }

     if (! input.file.name[0])			/* No file has been processed yet */
       init_projection(&proj);			/* Initialize projection */

     if (inputpath[0])				/* All valid options parsed - Current argument is a file name */
       sprintf(input.file.name, "%s/%s", inputpath, arg);
     else
       strcpy(input.file.name, arg);

     if (! input.sds.name[0]  &&  ! input.meta.name[0]) {
       printf("No SDS name or metadata name specified for file %s. File ignored.\n\n", input.file.name);
       input.file.name[0] = '\0';
       continue;			/* Skip to next argument */
     }
     
      /* Added 07/13 */
     if (L1B && ! input.sz.name[0]){
      printf("No solar zenith sds name specified for file %s. File ignored.\n\n", input.file.name);
      input.file.name[0] = '\0';
      continue;
     }
 
     if (open_file(&input.file) == -1) continue;		/* Skip to next input file */

     if (read_metadata(input.file.id, &input.allmetadata) == -1) {
       printf("  Unable to read ECS metadata in file %s.\n", input.file.name);
/*
       printf("  File ignored.\n");
       SDendaccess(input.sds.id);
       SDend(input.file.id);
       continue;			/ * Skip to next input file * /
*/
     }

     DNflag.value[0] = '\0';
     if ( get_metadata(input.allmetadata, " DAYNIGHTFLAG\n", DNflag.value, NULL, NULL) == -1  &&
          getattribute(input.file.id, "DAYNIGHTFLAG", DNflag.value) == -1 )
       printf("  Unable to retrieve Day/Night flag in file %s.\n", input.file.name);
     else {
       printf("  Day/Night flag: %s.", DNflag.value);
       if ( strcasecmp(DNflag.value, "day")   == 0  &&  ! (DNflag.request & 1)  ||
            strcasecmp(DNflag.value, "night") == 0  &&  ! (DNflag.request & 2)  ||
            strcasecmp(DNflag.value, "both")  == 0  &&  ! (DNflag.request >> 2) ) {
              printf(" File ignored.\n\n");
              SDendaccess(input.sds.id);
	      if (L1B) SDendaccess(input.sz.id); /* added 07/13 */
              SDend(input.file.id);
              continue;			/* Skip to next input file */
       } else printf("\n");
     }

     if ( get_metadata(input.allmetadata, " LOCALGRANULEID\n", input.lgid.string, &input.lgid.num_type, NULL) == -1 ) {
         printf("  Unable to retrieve Local ID in file %s.\n", input.file.name);
         input.lgid.string[0] = '\0';
     } else {
         if (ptr = strrchr(input.lgid.string, 47)) 		/* If filename contains "/", */
           strcpy(input.lgid.string, ptr + 1);			/* keep basename only */
         printf("  Local ID %s\n", input.lgid.string);
     }

     /* bounds correspond to edge (not center) of bounding pixels */
     /* bounds[0] / bounds[1] are the North / South bounding latitude of the tile */
     /* the latitude for the center of 1st line is bounds[0] - 0.5 * height of one line in the tile */
     if ( get_metadata(input.allmetadata, "NORTHBOUNDINGCOORDINATE\n", &input.bounds[0], NULL, NULL) == -1  &&
          getattribute(input.file.id, "NORTHBOUNDINGCOORDINATE", &input.bounds[0]) == -1  ||
          get_metadata(input.allmetadata, "SOUTHBOUNDINGCOORDINATE\n", &input.bounds[1], NULL, NULL) == -1  &&
          getattribute(input.file.id, "SOUTHBOUNDINGCOORDINATE", &input.bounds[1]) == -1  ||
          get_metadata(input.allmetadata, "EASTBOUNDINGCOORDINATE\n",  &input.bounds[2], NULL, NULL) == -1  &&
          getattribute(input.file.id, "EASTBOUNDINGCOORDINATE", &input.bounds[2]) == -1  ||
          get_metadata(input.allmetadata, "WESTBOUNDINGCOORDINATE\n",  &input.bounds[3], NULL, NULL) == -1  &&
          getattribute(input.file.id, "WESTBOUNDINGCOORDINATE", &input.bounds[3]) == -1  ) {
         printf("  Unable to read bounding coordinates in file %s.\n", input.file.name);
         input.bounds[0] = input.bounds[1] = input.bounds[2] = input.bounds[3] = PROJUNDEF;
     } else
         printf("  Bounds: N %g  S %g  E %g  W %g\n", input.bounds[0], input.bounds[1], input.bounds[2], input.bounds[3]);

     if ( get_metadata(input.allmetadata, "GRINGPOINTLATITUDE\n", input.meta.string, NULL, NULL) == -1  &&
          getattribute(input.file.id, "GRINGPOINTLATITUDE.1", input.meta.string) == -1  ||
          sscanf(input.meta.string, "%g %g %g %g", &input.gringlat[0], &input.gringlat[1], &input.gringlat[2], &input.gringlat[3]) != 4  ||
          get_metadata(input.allmetadata, "GRINGPOINTLONGITUDE\n", input.meta.string, NULL, NULL) == -1  &&
          getattribute(input.file.id, "GRINGPOINTLONGITUDE.1", input.meta.string) == -1  ||
          sscanf(input.meta.string, "%g %g %g %g", &input.gringlon[0], &input.gringlon[1], &input.gringlon[2], &input.gringlon[3]) != 4 ) {
         printf("  Unable to read G ring coordinates in file %s.\n", input.file.name);
         input.gringlat[0] = input.gringlat[1] = input.gringlat[2] = input.gringlat[3] = PROJUNDEF;
         input.gringlon[0] = input.gringlon[1] = input.gringlon[2] = input.gringlon[3] = PROJUNDEF;
     } else
         printf("  G ring: (%g,%g) (%g,%g) (%g,%g) (%g,%g).\n",
                input.gringlon[0], input.gringlat[0], input.gringlon[1], input.gringlat[1],
                input.gringlon[2], input.gringlat[2], input.gringlon[3], input.gringlat[3]);

     if ( input.allmetadata  &&
          ( strstr(input.allmetadata, "TILEID") != NULL  ||
            strstr(input.allmetadata, "TileID") != NULL  ||
            strstr(input.allmetadata, "TileId") != NULL  ||
            strstr(input.allmetadata, "TILENUMBER") != NULL ) )
     {
              if ( ( get_metadata(input.allmetadata, "TILEID", &tileid, NULL, NULL) != -1  ||
              get_metadata(input.allmetadata, "TileID", &tileid, NULL, NULL) != -1  ||
              get_metadata(input.allmetadata, "TileId", &tileid, NULL, NULL) != -1 ) &&
              ((int)(tileid / 1e7) == 5) || ((int)(tileid / 1e7) == 6))
              {
                     if (((int)(tileid / 1e7) == 5))
                     {
                            input.projparam = isinusparam;				/* Tile product */
                            isinusint(1.0, 0., 0., 0., isinusparam.nzones);
                     }
                     else /* for LLL */
                     {
                            input.projparam = geogparam;
                     }
              }
              else
                     input.projparam.type = SWATH;    
     } else
         input.projparam.type = SWATH;				/* Granule product */

/* Depending on the season the night scans of the terminator granules may overlap a day granule - Those can be skipped on request */
     if ( input.projparam.type == SWATH  &&
          DNflag.value[0]  &&				/* Day/Night flag was successfully read */
          strcasecmp(DNflag.value, "both") == 0  &&	/* Terminator granule */
          ! (DNflag.request & 4) ) {			/* Not all terminators requested */
       if ( DNflag.request & 8  &&
            input.gringlat[0] < 0) {			/* Southern hemisphere */
         printf("  File is a terminator granule in Southern hemisphere. File ignored.\n\n");
         SDendaccess(input.sds.id);
	 if (L1B) SDendaccess(input.sz.id); /* added 07/13 */
         SDend(input.file.id);
         continue;			/* Skip to next input file */
       }
       if ( (DNflag.request & 16)  &&
            input.gringlat[0] > 0) {			/* Northern hemisphere */
         printf("  File is a terminator granule in Northern hemisphere. File ignored.\n\n");
         SDendaccess(input.sds.id);
	 if (L1B) SDendaccess(input.sz.id); /* added 07/13 */
         SDend(input.file.id);
         continue;			/* Skip to next input file */
       }
     }

     lgidcopy = input.lgid;

     if (LGID) {
       input.lgid = lgidcopy;
       convert_metadata_string(&proj.lgid, &input.lgid);		/* input.lgid.string and input.lgid.num_type are modified */
       proj.lgid.min = proj.lgid.max = 0;
       proj.lgid.num_type = DFNT_UINT8;
     }

     if (input.sds.name[0]) {		/* We are here for a SDS */

         input.sds.file_id = input.file.id;
	 input.sz.file_id = input.file.id;
         if ( get_SDS_info(&input.sds) == -1  ||
              get_geolocation(&input) == -1 ) {
           SDendaccess(input.sds.id);
	   if (L1B) SDendaccess(input.sz.id); /* added 07/13 */
           SDend(input.file.id);
           continue;		/* Skip to next input file */
         }
	 
	 /* Added solar zenith sds info for L1B option only */
	 if (L1B)
	 {
	    if ( get_SDS_info(&input.sz) == -1)
	    {
	      printf(" No sds info for solar zenith sds %s file %s\n. File ignored", input.sz.name, input.file.name);
	      SDendaccess(input.sds.id);
	      SDendaccess(input.sz.id); /* added 07/13 */
	      SDend(input.file.id);
	      continue;		/* Skip to next input file */
	    }
	 }

         if (input.sds.rank < 3)
           sprintf(proj.sds.name, "%s (%g km)", input.sds.name, proj.pixsz * DEG2KM);
         else
           sprintf(proj.sds.name, "%s.%s (%g km)", input.sds.name, input.sds.planeID[0], proj.pixsz * DEG2KM);

         update_projection(&proj, input);

         SDendaccess(input.sds.id);
	 if (L1B) SDendaccess(input.sz.id); /* added 07/13 */
         SDend(input.file.id);
         if (input.projparam.type == SWATH) {
           SDendaccess(input.lat.id);
           SDendaccess(input.lon.id);
         }
         input.adjacent = FALSE;
         printf("\n");

     }

     if (input.meta.name[0]) {		/* We are here for a metadata */

         if ( get_metadata(input.allmetadata, input.meta.name, input.meta.string, &input.meta.num_type, input.meta.fullname) == -1  &&
              getattribute(input.file.id, input.meta.name, input.meta.string) == -1  ||
              sprintf(proj.meta.name, "%s (%g km)", input.meta.fullname, proj.pixsz * DEG2KM) == -1  ||
              update_meta_field(&proj.meta, &proj, input, input.meta) == -1 ) {
           printf("\n");
           SDend(input.file.id);
           continue;			/* Skip to next input file */
         }
         printf("\n");
     }

     SDend(input.file.id);

   }	/* Go to next input file */


if (! proj.sds.data  &&  ! proj.meta.data) {		/* No valid input file was processed */
     fprintf(stderr, "No valid input file was processed. Empty projection.\n");
     if (! NODATA) {
       fprintf(stderr, "%s", HELPMSG);
       if ( grid.request == 0  ||  grid.separate == FALSE )	/* No separate grid requested */
       exit(-1);
     }
     init_projection(&proj);
     if (proj.domain == USERBOX) {
         proj.row0 = proj.Nl_glb / 2. + 0.5 - 1 - proj.bbox.l1;
         proj.col0 = proj.Np_glb / 2. + 0.5 - 1 - proj.bbox.p1;
     } else {
         proj.row0 = proj.Nl_glb / 2. + 0.5 - 1;
         proj.col0 = proj.Np_glb / 2. + 0.5 - 1;
     }
     compute_grids(proj, &grid);
     if (NODATA) {
       proj.sds.num_type = DFNT_UINT8;
       sprintf(proj.sds.name, "%s (%g km)", input.sds.name, proj.pixsz * DEG2KM);
       resize(&proj, NULL);
       proj.row1 = 0;			/* No restriction on domain */
       proj.row2 = proj.sds.Nl - 1;
       outline_domain(proj.sds.data, proj.sds.num_type, proj);
       add_grids(proj.sds.data, proj.sds.num_type, proj, grid);
       add_vectors(proj.sds.data, proj.sds.num_type, proj, vector);
       printf("\n");
     }

} else {

   if ( proj.domain == EXTBOX  &&
        ( proj.bbox.l1 != proj.allocbox.l1  ||  proj.bbox.p1 != proj.allocbox.p1  ||
          proj.bbox.l2 != proj.allocbox.l2  ||  proj.bbox.p2 != proj.allocbox.p2 ) ) {
     printf("Adjusting the final size of the output projection...\n");
     resize(&proj, NULL);
     printf("\n");
   }
   printf("Output subwindow in global projection: col=%d row=%d width=%d height=%d\n\n",
          proj.bbox.p1, proj.bbox.l1, proj.bbox.p2 - proj.bbox.p1 + 1, proj.bbox.l2 - proj.bbox.l1 + 1);
   printf("Option to use to match this output subwindow: -box=%d,%d,%d,%d\n\n",
          proj.bbox.p1, proj.bbox.l1, proj.bbox.p2 - proj.bbox.p1 + 1, proj.bbox.l2 - proj.bbox.l1 + 1);

   fill_domain(&proj);
   compute_grids(proj, &grid);

   if (proj.sds.data) {
     proj.row1 = 0;			/* No restriction on domain */
     proj.row2 = proj.sds.Nl - 1;
     outline_domain(proj.sds.data, proj.sds.num_type, proj);
     add_grids(proj.sds.data, proj.sds.num_type, proj, grid);
     add_vectors(proj.sds.data, proj.sds.num_type, proj, vector);
     printf("\n");
   }

   if (proj.meta.data) {
     proj.row1 = 0;			/* No restriction on domain */
     proj.row2 = proj.meta.Nl - 1;
     outline_domain(proj.meta.data, proj.meta.num_type, proj);
     add_grids(proj.meta.data, proj.meta.num_type, proj, grid);
     add_vectors(proj.meta.data, proj.meta.num_type, proj, vector);
     printf("\n");
   }
}

   if (! proj.file.name[0]) {
     fprintf(stderr, "No output file specified.\n");
     fprintf(stderr, "%s", HELPMSG);
     exit(-1);
   }

   switch (proj.access_mode) {
     case CREATE:
       printf("Opening %s (Overwrite mode)...\n", proj.file.name);
       access_mode = DFACC_CREATE;
       break;
     case APPEND:
       printf("Opening %s (Append mode)...\n",    proj.file.name);
       access_mode = DFACC_WRITE;
       break;
     case REPLACE:
       printf("Opening %s (Replace mode)...\n",   proj.file.name);
       access_mode = DFACC_WRITE;
       break;
   }
   if ( (proj.file.id = SDstart(proj.file.name, access_mode)) == -1  &&
        (proj.file.id = SDstart(proj.file.name, DFACC_CREATE)) == -1 ) {	/* Try to create the file if append mode */
     fprintf(stderr, "Cannot open file %s.\n", proj.file.name);
     exit(-1);
   }
   printf("\n");

   if (proj.lgid.data) {
     sprintf(proj.lgid.name, "%s (%g km)", proj.lgid.name, proj.pixsz * DEG2KM);
     proj.lgid.file_id = proj.file.id;
     if ( create_2D_SDS(&proj.lgid, proj.access_mode) == -1  ||
          write_SDS(proj.lgid, 0, proj.lgid.Nl, options.gzip) == -1 )
       exit(-1);
     write_SDS_attributes(proj.lgid, proj, grid, vector);
     free(proj.lgid.data);
     free(proj.lgid.lut);
     SDendaccess(proj.lgid.id);
   }

   if (proj.sds.data) {

     k = 1;
     strcpy(sdsname, proj.sds.name);
     if (proj.access_mode != REPLACE)
       while ( SDnametoindex(proj.file.id, proj.sds.name) != -1 )
         sprintf(proj.sds.name, "%s - %d", sdsname, k++);
     proj.sds.file_id = proj.file.id;
     if ( create_2D_SDS(&proj.sds, proj.access_mode) == -1 ) exit(-1);
     if ( write_SDS(proj.sds, 0, proj.sds.Nl, options.gzip) == -1 ) exit(-1);
     write_SDS_attributes(proj.sds, proj, grid, vector);
     SDendaccess(proj.sds.id);
     free(proj.sds.data);

   }

   if (proj.meta.data) {

     k = 1;
     strcpy(sdsname, proj.meta.name);
     if (proj.access_mode != REPLACE)
       while ( SDnametoindex(proj.file.id, proj.meta.name) != -1 )
         sprintf(proj.meta.name, "%s - %d", sdsname, k++);
     proj.meta.file_id = proj.file.id;
     if ( create_2D_SDS(&proj.meta, proj.access_mode) == -1 ) exit(-1);
     if ( write_SDS(proj.meta, 0, proj.meta.Nl, options.gzip) == -1 ) exit(-1);
     write_SDS_attributes(proj.meta, proj, grid, vector);
     free(proj.meta.data);
     if (proj.meta.Nlut > 0) free(proj.meta.lut);
     SDendaccess(proj.meta.id);

   }

   printf("\n");

   if (grid.separate) {
     make_fill_SDS(proj.file.id, proj.file.name, proj, grid, vector);
     write_grid_SDS(proj.file.id, grid, proj);
   }

   write_file_attributes(cmd_line, proj.file.id);

   SDend(proj.file.id);

   return 0;

}




void write_SDS_attributes(SDS sds, PROJECTION proj, GRID grid, VECTOR vector)
{
char name[MAXNAMELENGTH];
double pixsz, offset, value;
double xproj, yproj, lon, lat;
int k, pixoffset;

   strcpy(name, "Projection");
   if ( SDsetattr(sds.id, name, DFNT_CHAR8, (int)strlen(projinfo[proj.param.type].name) + 1, projinfo[proj.param.type].name) == -1 )
     printf("Unable to write projection name in SDS attributes.\n");

   pixsz = proj.pixsz * DEG2KM;
   if ( ! (pixsz = proj.pixsz * DEG2KM)  ||
        SDsetattr(sds.id, "Pixel Size (km)", DFNT_FLOAT64, 1, &pixsz) == -1  ||
        ! (pixsz = proj.pixsz)  ||
        SDsetattr(sds.id, "Pixel Size (deg)", DFNT_FLOAT64, 1, &pixsz) == -1 )
     printf("Unable to write pixel size in SDS attributes.\n");

   if (SDsetattr(sds.id, "Upper Left Corner Horizontal Offset (pixels)", DFNT_INT32, 1, &proj.bbox.p1) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   if (SDsetattr(sds.id, "Upper Left Corner Vertical Offset (pixels)", DFNT_INT32, 1, &proj.bbox.l1) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   offset = proj.bbox.p1 * proj.pixsz * DEG2KM;
   if (SDsetattr(sds.id, "Upper Left Corner Horizontal Offset (km)", DFNT_FLOAT64, 1, &offset) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   offset = proj.bbox.l1 * proj.pixsz * DEG2KM;
   if (SDsetattr(sds.id, "Upper Left Corner Vertical Offset (km)", DFNT_FLOAT64, 1, &offset) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   pixoffset = proj.bbox.p1 - (proj.Np_glb / 2. + 0.5 - 1);
   if (SDsetattr(sds.id, "Upper Left Corner Easting (pixels)", DFNT_INT32, 1, &pixoffset) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   pixoffset = proj.Nl_glb / 2. + 0.5 - 1 - proj.bbox.l1;
   if (SDsetattr(sds.id, "Upper Left Corner Northing (pixels)", DFNT_INT32, 1, &pixoffset) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   offset = (proj.bbox.p1 - (proj.Np_glb / 2. + 0.5 - 1)) * proj.pixsz * DEG2KM;
   if (SDsetattr(sds.id, "Upper Left Corner Easting (km)", DFNT_FLOAT64, 1, &offset) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");
   offset = (proj.Nl_glb / 2. + 0.5 - 1 - proj.bbox.l1) * proj.pixsz * DEG2KM;
   if (SDsetattr(sds.id, "Upper Left Corner Northing (km)", DFNT_FLOAT64, 1, &offset) == -1)
     printf("Unable to write upper left corner in SDS attributes.\n");

   xproj = (proj.bbox.p1 - (proj.Np_glb / 2. + 0.5 - 1) - 0.5) * proj.pixsz;
   yproj = ((proj.Nl_glb / 2. + 0.5 - 1) - proj.bbox.l1 + 0.5) * proj.pixsz;
   if ( get_lonlat(xproj, yproj, &lon, &lat, &proj.param) != -1 ) {
     if (SDsetattr(sds.id, "Upper Left Corner Bounding Longitude (deg.)", DFNT_FLOAT64, 1, &lon) == -1)
       printf("Unable to write upper left corner in SDS attributes.\n");
     if (SDsetattr(sds.id, "Upper Left Corner Bounding Latitude (deg.)", DFNT_FLOAT64, 1, &lat) == -1)
       printf("Unable to write upper left corner in SDS attributes.\n");
   }
   xproj = (proj.bbox.p2 - (proj.Np_glb / 2. + 0.5 - 1) + 0.5) * proj.pixsz;
   yproj = ((proj.Nl_glb / 2. + 0.5 - 1) - proj.bbox.l2 - 0.5) * proj.pixsz;
   if ( get_lonlat(xproj, yproj, &lon, &lat, &proj.param) != -1 ) {
     if (SDsetattr(sds.id, "Lower Right Corner Bounding Longitude (deg.)", DFNT_FLOAT64, 1, &lon) == -1)
       printf("Unable to write lower right corner in SDS attributes.\n");
     if (SDsetattr(sds.id, "Lower Right Corner Bounding Latitude (deg.)", DFNT_FLOAT64, 1, &lat) == -1)
       printf("Unable to write lower right corner in SDS attributes.\n");
   }

   setfillvalue(&value, 0, 1, sds.num_type, DOMAINOFFSET);
   if (SDsetattr(sds.id, "Outer Domain Value", sds.num_type, 1, &value) == -1)
     printf("Unable to write outer domain value in SDS attributes.\n");
   setfillvalue(&value, 0, 1, sds.num_type, OUTLINEOFFSET);
   if (SDsetattr(sds.id, "Outline Value", sds.num_type, 1, &value) == -1)
     printf("Unable to write outline value in SDS attributes.\n");
   if (grid.Npoints[MERIDIANS] > 0) {
     setfillvalue(&value, 0, 1, sds.num_type, grid.offset[MERIDIANS]);
     if (SDsetattr(sds.id, "Meridian Value", sds.num_type, 1, &value) == -1)
       printf("Unable to write meridian value in SDS attributes.\n");
   }
   if (grid.Npoints[PARALLELS] > 0) {
     setfillvalue(&value, 0, 1, sds.num_type, grid.offset[PARALLELS]);
     if (SDsetattr(sds.id, "Parallel Value", sds.num_type, 1, &value) == -1)
       printf("Unable to write parallel value in SDS attributes.\n");
   }
   if (grid.Npoints[TILES] > 0) {
     setfillvalue(&value, 0, 1, sds.num_type, grid.offset[TILES]);
     if (SDsetattr(sds.id, "L3 Grid Value", sds.num_type, 1, &value) == -1)
       printf("Unable to write L3 grid value in SDS attributes.\n");
   }
   if (grid.Npoints[NODES] > 0) {
     setfillvalue(&value, 0, 1, sds.num_type, grid.offset[NODES]);
     if (SDsetattr(sds.id, "Node Value", sds.num_type, 1, &value) == -1)
       printf("Unable to write Node value in SDS attributes.\n");
   }

   for (k=0; k<NVECTORS; k++) {
     if ( ! (vector.request & 1 << k) ) continue;		/* Skip to next vector layer */
     setfillvalue(&value, 0, 1, sds.num_type, vector.offset[k]);
     if (SDsetattr(sds.id, capitalize(vector.type[k]), sds.num_type, 1, &value) == -1)
       printf("Unable to write \"%s\" value in SDS attributes.\n", vector.type[k]);
   }

   if ( sds.lut  &&
        SDsetattr(sds.id, "LUT", DFNT_CHAR8, (int)strlen(sds.lut) + 1, sds.lut) == -1 )
     printf("Unable to write attribute LUT for SDS %s.\n", sds.name);

}




int getattribute(int32 id, char *name, void *value)
{
int32 num_type, count, attr_index;

  if ( (attr_index = SDfindattr(id, name)) == -1  ||
       SDattrinfo(id, attr_index, name, &num_type, &count) == -1 ) {
    fprintf(stderr, "Cannot find attribute \"%s\" or get info\n", name);
    return -1;
  }
  if ((value = realloc(value, count * DFKNTsize(num_type))) == NULL  ||
       SDreadattr(id, attr_index, value) == -1 ) {
    fprintf(stderr, "Cannot read attribute \"%s\"\n", name);
    return -1;
  }

  return 0;

}




char *capitalize(char *string)
{
char *ptr, *newstring;

  newstring = strdup(string);
  ptr = newstring - 1;
  while ( ptr == newstring - 1  ||
          (ptr = strchr(ptr, 32)) ) {
    ptr++;
    if ( *ptr >= 65  &&
         *ptr <= 122 )
      *ptr -= 32;
  }
  return newstring;

}




void write_file_attributes(char *cmd_line, int32 file_id)
{
char name[MAXNAMELENGTH];
int k;

   k = 1;
   while ( sprintf(name, "Command Line #%d", k) != -1  &&
           SDfindattr(file_id, name) != -1 )
     k++;

   if ( strlen(cmd_line) >= MAX_ATTR_SIZE )
     printf("Command line was truncated to max. attribute size (%d) in file attributes.\n", MAX_ATTR_SIZE);

   if ( SDsetattr(file_id, name, DFNT_CHAR8, MIN((int)strlen(cmd_line) + 1, MAX_ATTR_SIZE), cmd_line) == -1 )
         printf("Unable to write command line in file attributes.\n");

}




int read_metadata(int32 sd_id_in, char **allmetadata)
{
int32 attr_index, num_type, count1=0, count2;
char  *ptr, name[MAXNAMELENGTH];

   if ( (attr_index = SDfindattr(sd_id_in, "StructMetadata.0")) != -1 ) {

     SDattrinfo(sd_id_in, attr_index, name, &num_type, &count1);
     if ( check_memory((count1 + 1) * sizeof(char)) == -1 ) return -1;
     *allmetadata = (char *) realloc(*allmetadata, (count1 + 1) * sizeof(char));	/* + 1 is for terminating NULL character */

     if (SDreadattr(sd_id_in, attr_index, *allmetadata) == -1) {
       printf("  Unable to read structure metadata.\n");
       free(*allmetadata);
       return -1;
     }
     *(*allmetadata + count1) = '\0';
     if ( (ptr = strstr(*allmetadata, "END\n")) != NULL) {
      // *(ptr + 4) = '\0';
       count1 = (int32) strlen(*allmetadata);
     }

   } else printf("  Unable to find structure metadata.\n");


   if ( (attr_index = SDfindattr(sd_id_in, "CoreMetadata.0")) == -1 ) {
     printf("  Unable to find core metadata.\n");
     return -1;
   }

   SDattrinfo(sd_id_in, attr_index, name, &num_type, &count2);
   if ( check_memory((count1 + count2 + 1) * sizeof(char)) == -1 ) return -1;
   *allmetadata = (char *) realloc(*allmetadata, (count1 + count2 + 1) * sizeof(char));

   if (SDreadattr(sd_id_in, attr_index, *allmetadata + count1) == -1) {
     printf("  Unable to read core metadata.\n");
     free(*allmetadata);
     return -1;
   }
   *(*allmetadata + count1 + count2) = '\0';
   if ( (ptr = strstr(*allmetadata + count1, "END\n")) != NULL) {
    // *(ptr + 4) = '\0';
     count1 = (int32) strlen(*allmetadata);
   }


   if ( (attr_index = SDfindattr(sd_id_in, "ArchiveMetadata.0")) == -1  &&
        (attr_index = SDfindattr(sd_id_in, "ArchivedMetadata.0")) == -1 ) {
     printf("  Unable to find archive metadata.\n");
     return -1;
   }

   SDattrinfo(sd_id_in, attr_index, name, &num_type, &count2);
   if ( check_memory((count1 + count2 + 1) * sizeof(char)) == -1 ) return -1;
   *allmetadata = (char *) realloc(*allmetadata, (count1 + count2 + 1) * sizeof(char));

   if (SDreadattr(sd_id_in, attr_index, *allmetadata + count1) == -1) {
     printf("  Unable to read archive metadata.\n");
     free(*allmetadata);
     return -1;
   }
   *(*allmetadata + count1 + count2) = '\0';
   if ( (ptr = strstr(*allmetadata + count1, "END\n")) != NULL)
    // *(ptr + 4) = '\0';

  /* printf("whole metadata is %s \n",*allmetadata); */
   return 0;

}




int get_metadata(char *metadata, char *object, void *value, int32 *num_type, char *fullname)
{
char *ptr0, *ptr, string[MAXMETALENGTH];
long len;

   if (! metadata) return -1;

   if ( (ptr0 = strstr(metadata, object)) == NULL ||
        ( sscanf(ptr0 + strlen(object), "=%s", string) != 1 &&	/* Use string instead of value directly (might be too short if numeric) */
          ( (ptr = strstr(ptr0, " VALUE")) == NULL ||
          sscanf(ptr, "%*s = %s", string) != 1 ) ) ) {
      strcpy(string, object);
      if (string[strlen(string) - 1] == 10)		/* If object name is terminated by a new line character,*/
          string[strlen(string) - 1] = '\0';		/* then delete it before printing error message */
      printf("  Unable to find object %s in metadata.\n", string);
      return -1;
   }
   if (fullname != NULL) {
     while ( ptr0 > metadata  &&
             *(ptr0 - 1) >= 65  &&
             *(ptr0 - 1) <= 90 ) ptr0--;
     sscanf(ptr0, "%s", fullname);			/* Get the object fullname (in case name was only partially specified) */
     if (fullname [strlen(fullname) - 1] == 34) 	/* If object fullname is terminated */
       fullname [strlen(fullname) - 1] = '\0';	/* by a quote, then delete it */
   }
   ptr = strstr(ptr, string);			/* ptr points on requested value in metadata string */

   if ( sscanf(ptr, "%g", (float *)string) != 1 ) {	/* Value is not numeric */

       if (num_type) *num_type = DFNT_UCHAR8;
       if (*ptr == 40) {			/* Value starts with "(", then is an array, must read all items */
         ptr++;					/* Skip opening "(" */
         len = strchr(ptr, 41) - ptr;		/* Array length (closing ")" not included) */
         if (len + 1 > MAXMETALENGTH) {
           if (object [strlen(object) - 1] == 10)
             object [strlen(object) - 1] = '\0';
           printf("  Length of object %s in metadata is greater than max length (%d).  Truncated.\n", object, MAXMETALENGTH);
           len = MAXMETALENGTH - 1;
         }
         strncpy(string, ptr, len * sizeof(char));		/* Copy all array but closing ")" */
         string[len] = '\0';				/* Add terminating NULL character */
         ptr = string;
         while (ptr = strchr(ptr, 10)) strcpy(ptr, ptr + 1);	/* Delete newline characters (if any) */
         ptr = string;
         while (ptr = strchr(ptr, 34)) strcpy(ptr, ptr + 1);	/* Delete quotes (") if any */
         ptr = string;
         while (ptr = strchr(ptr, 44)) *ptr = 32;		/* Replace commas by spaces (if any) */
         strcpy(value, string);					/* Copy result string */
       }
       else if (*ptr == 34) {				/* Value starts with a quote ("), then should end with a quote */
         ptr++;						/* Skip opening quote */
         len = strchr(ptr, 34) - ptr;			/* String length (closing quote and NULL character not included) */
         strncpy(string, ptr, len * sizeof(char));	/* Copy as is, except closing quote */
         string[len] = '\0';			/* Terminate string */
         if (! strpbrk(string, ":")) {			/* Make sure not to parse a time string */
           if (strpbrk(string, "."))			/* The string might actually be a float */
             if (sscanf(string, "%g", (float *)value) == 1) {
                 if (num_type) *num_type = DFNT_FLOAT32;	/* Float value was successfully parsed */
             } else
                 strcpy(value, string);			/* Just copy string into result string as is */
           else						/* The string might actually be an integer */
             if (sscanf(string, "%d", (int *)value) == 1) {	/* Last attempt to parse a number */
                 if (num_type) *num_type = DFNT_INT32;	/* Integer value was successfully parsed */
             } else
                 strcpy(value, string);			/* Just copy string into result string as is */
         } else strcpy(value, string);			/* Just copy string into result string as is */
       }
       else strcpy(value, string);			/* Just copy string into result string as is */

   } else {						/* Value is numeric */

       if (strpbrk(ptr, ".")) {				/* Value is a float */
           sscanf(ptr, "%g", (float *)value);
           if (num_type) *num_type = DFNT_FLOAT32;
       } else {						/* Value is an integer */
           sscanf(ptr, "%d", (int *)value);
           if (num_type) *num_type = DFNT_INT32;
       }

   }

   return 0;

}




void get_file_from_inputpointer(PRODUCT input, char *shortname, char *filename)
{
char *ptr, string[MAXMETALENGTH], *start, attr_name[MAXNAMELENGTH];
int32 attr_index;

         strcpy(attr_name, "MOD03 input file");
         attr_index = SDfindattr(input.file.id, attr_name);

         if (SDreadattr(input.file.id, attr_index, string) != -1) {

             if ( ptr = strrchr(string, 47) ) strcpy(string, ptr + 1);

         } else if ( get_metadata(input.allmetadata, " INPUTPOINTER\n", string, NULL, NULL) == -1 ||
              (ptr = strstr(string, shortname)) == NULL ||
              sscanf(ptr, "%s", string) != 1) {

             sprintf(string, "%s.%s", shortname, input.file.name);
             start = string + strlen(shortname) + 1;
	    printf ("start is %s\n",start);
             if ( ptr = strrchr(start, 47) ) strcpy(start, ptr + 1);
             if ( ptr = strchr(start, 46) ) strcpy(start, ptr + 1);
             ptr = strdup(shortname);
             *(ptr + strlen(shortname)) = '\0';
             printf("  Unable to find %s input file in metadata. %s is assumed.\n", ptr, string);

         }
	printf("ptr is %s\n",ptr);
	printf("string is %s\n",string);
         strcpy(filename, string);

}




int resize(PROJECTION *proj, BOUNDINGBOX *bbox)
{
int Nl_sub, Np_sub;
int irow1, irow2, jcol1, jcol2;
unsigned char subset=FALSE;
unsigned char senzfillvalue=UCHAR_MAX;

     if (! bbox || proj->domain == USERBOX) {
         irow1 = proj->bbox.l1;			/* Subset subwindow to actual size */
         irow2 = proj->bbox.l2;
         jcol1 = proj->bbox.p1;
         jcol2 = proj->bbox.p2;
         subset = TRUE;
     } else if (proj->allocbox.l2 == -1) {	/* No array allocated yet */
         irow1 = bbox->l1;			/* The extreme lines and columns of the new subwindow */
         irow2 = bbox->l2;			/* are those specified by the input bounding box. */
         jcol1 = bbox->p1;			/* This case must be handled separately, because */
         jcol2 = bbox->p2;			/* allocbox.l1 and allocbox.p1 are initialized to 0. */
         printf("  No projection space allocated yet\n");
     } else {
         printf("  New tile or granule doesn't fit in the current projection.\n");
         irow1 = MIN(bbox->l1, proj->allocbox.l1);	/* Compute extreme lines and columns of the */
         irow2 = MAX(bbox->l2, proj->allocbox.l2);	/* subwindow to account for both the current */
         jcol1 = MIN(bbox->p1, proj->allocbox.p1);	/* projection and the new tile or granule */
         jcol2 = MAX(bbox->p2, proj->allocbox.p2);
     }

     if (irow1 < 0) irow1 = 0;			/* Sometimes the requested bounding box needs */
     if (jcol1 < 0) jcol1 = 0;			/* to be adjusted when a safety margin was used */
     if (irow2 > proj->Nl_glb - 1) irow2 = proj->Nl_glb - 1;
     if (jcol2 > proj->Np_glb - 1) jcol2 = proj->Np_glb - 1;

     Nl_sub = irow2 - irow1 + 1;			/* Dimensions of new subwindow */
     Np_sub = jcol2 - jcol1 + 1;

     if ( options.minsenz  &&		/* Resize proj->senz first because proj->sds.Nl and proj->sds.Np are not updated yet */
          resize_data((void **)(&proj->senz), Nl_sub, Np_sub, proj->sds.Nl, proj->sds.Np, proj->allocbox.p1 - jcol1, proj->allocbox.l1 - irow1, DFNT_UINT8, subset, &senzfillvalue) == -1 )
       return -1;

     if ( proj->sds.num_type != DFNT_NONE  &&
          resize_SDS(&proj->sds, Nl_sub, Np_sub, irow1, jcol1, *proj, subset) == -1 )
       return -1;

     if ( proj->meta.num_type != DFNT_NONE  &&
          resize_SDS(&proj->meta, Nl_sub, Np_sub, irow1, jcol1, *proj, subset) == -1 )
       return -1;

     if ( proj->lgid.num_type != DFNT_NONE  &&		/* Only if LGID is TRUE */
          resize_SDS(&proj->lgid, Nl_sub, Np_sub, irow1, jcol1, *proj, subset) == -1 )
       return -1;

     proj->allocbox.l1 = irow1;		/* Update limits and size of the new subwindow */
     proj->allocbox.l2 = irow2;
     proj->allocbox.p1 = jcol1;
     proj->allocbox.p2 = jcol2;
     proj->row0 = proj->Nl_glb / 2. + 0.5 - 1 - proj->allocbox.l1;
     proj->col0 = proj->Np_glb / 2. + 0.5 - 1 - proj->allocbox.p1;

     return 0;

}




int resize_SDS(SDS *sds, int Nl_sub, int Np_sub, int irow1, int jcol1, PROJECTION proj, unsigned char subset)
{
int totalsize;
int Hoffset=0, Voffset=0;

  printf("  Resizing SDS \"%s\" (%dx%d -> %dx%d)...\n", sds->name, sds->Np, sds->Nl, Np_sub, Nl_sub);

  if (sds->data) {				/* Data already exits */
    Hoffset = proj.allocbox.p1 - jcol1;		/* Offset of old subwindow in new subwindow */
    Voffset = proj.allocbox.l1 - irow1;
  }

  if (resize_data(&sds->data, Nl_sub, Np_sub, sds->Nl, sds->Np, Hoffset, Voffset, sds->num_type, subset, NULL) == -1)
    return -1;

  sds->Nl = Nl_sub;
  sds->Np = Np_sub;
  totalsize = sds->Nl * sds->Np * DFKNTsize(sds->num_type);
  printf("  Now using %.1f Mbytes of memory for data.\n", totalsize / 1e6);	/* A resize occurred due to dimensions */

  return 0;

}




int convert_data(void **data, int size, int32 type1, int32 type2, unsigned char verbose)
{
int k;
void *newdata;
int8    *xint8;
uint8   *xuint8;
int16   *xint16;
uint16  *xuint16;
int32   *xint32;
uint32  *xuint32;
float32 *xfloat32;

  if (type1 == type2) return 0;		/* Same type - No conversion needed */

  else {

    if (verbose) printf("  New type (%s) doesn't fit in the current one (%s).\n"
                        "  Converting current projection", typename[type2], typename[type1]);

    switch (type1) {
      case DFNT_INT8:   xint8   = *data;  break;
      case DFNT_UINT8:  xuint8  = *data;  break;
      case DFNT_INT16:  xint16  = *data;  break;
      case DFNT_UINT16: xuint16 = *data;  break;
      case DFNT_INT32:  xint32  = *data;  break;
      case DFNT_UINT32: xuint32 = *data;  break;
      case DFNT_NONE:   break;
      default:
         if (verbose) printf(".\n");
         printf("  Don't know how to convert data type %s. File ignored.\n\n", typename[type1]);
         return -1;
    }

    if ( check_memory(size * DFKNTsize(type2)) == -1  ||
         (newdata = (void *) malloc(size * DFKNTsize(type2))) == NULL ) {
      if (verbose) printf(".\n");
      printf("  Sorry, can't allocate %.1f Mbytes of memory. File ignored.\n\n", size * DFKNTsize(type2) / 1e6);
      return -1;
    }
    
    switch (type2) {
      case DFNT_INT8:    xint8    = newdata;  break;
      case DFNT_UINT8:   xuint8   = newdata;  break;
      case DFNT_INT16:   xint16   = newdata;  break;
      case DFNT_UINT16:  xuint16  = newdata;  break;
      case DFNT_INT32:   xint32   = newdata;  break;
      case DFNT_UINT32:  xuint32  = newdata;  break;
      case DFNT_FLOAT32: xfloat32 = newdata;  break;
      default:
         free(newdata);
         if (verbose) printf(".\n");
         printf("  Don't know how to convert data into type %s. File ignored.\n\n", typename[type2]);
         return -1;
    }

    switch (type2) {
      /* Convert existing data. Data types are different -> do not use memcpy */
      /* Test fill value in the initial data type first */

      case DFNT_INT8:
         switch (type1) {
           case DFNT_UINT8:  for (k=0; k<size; k++) if (xuint8[k] != fill.uint8) xint8[k] = xuint8[k];
                                                                            else xint8[k] = fill.int8;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xint8[k] = fill.int8;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }
         break;

      case DFNT_UINT8:
         switch (type1) {
           case DFNT_INT8:   for (k=0; k<size; k++) if (xint8[k] != fill.int8) xuint8[k] = xint8[k];
                                                                          else xuint8[k] = fill.uint8;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xuint8[k] = fill.uint8;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }
         break;

      case DFNT_INT16:
         switch (type1) {
           case DFNT_INT8:   for (k=0; k<size; k++) if (xint8[k] != fill.int8) xint16[k] = xint8[k];
                                                                          else xint16[k] = fill.int16;
                             break;
           case DFNT_UINT8:  for (k=0; k<size; k++) if (xuint8[k] != fill.uint8) xint16[k] = xuint8[k];
                                                                            else xint16[k] = fill.int16;
                             break;
           case DFNT_UINT16: for (k=0; k<size; k++) if (xuint16[k] != fill.uint16) xint16[k] = xuint16[k];
                                                                              else xint16[k] = fill.int16;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xint16[k] = fill.int16;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }
         break;

      case DFNT_UINT16:
         switch (type1) {
           case DFNT_INT8:   for (k=0; k<size; k++) if (xint8[k] != fill.int8) xuint16[k] = xint8[k];
                                                                          else xuint16[k] = fill.uint16;
                             break;
           case DFNT_UINT8:  for (k=0; k<size; k++) if (xuint8[k] != fill.uint8) xuint16[k] = xuint8[k];
                                                                            else xuint16[k] = fill.uint16;
                             break;
           case DFNT_INT16:  for (k=0; k<size; k++) if (xint16[k] != fill.int16) xuint16[k] = xint16[k];
                                                                            else xuint16[k] = fill.uint16;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xuint16[k] = fill.uint16;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }
         break;

      case DFNT_INT32:
         switch (type1) {
           case DFNT_INT8:   for (k=0; k<size; k++) if (xint8[k] != fill.int8) xint32[k] = xint8[k];
                                                                          else xint32[k] = fill.int32;
                             break;
           case DFNT_UINT8:  for (k=0; k<size; k++) if (xuint8[k] != fill.uint8) xint32[k] = xuint8[k];
                                                                            else xint32[k] = fill.int32;
                             break;
           case DFNT_INT16:  for (k=0; k<size; k++) if (xint16[k] != fill.int16) xint32[k] = xint16[k];
                                                                            else xint32[k] = fill.int32;
                             break;
           case DFNT_UINT16: for (k=0; k<size; k++) if (xuint16[k] != fill.uint16) xint32[k] = xuint16[k];
                                                                              else xint32[k] = fill.int32;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xint32[k] = fill.int32;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }
         break;

      case DFNT_UINT32:
         switch (type1) {
           case DFNT_INT8:   for (k=0; k<size; k++) if (xint8[k] != fill.int8) xuint32[k] = xint8[k];
                                                                          else xuint32[k] = fill.uint32;
                             break;
           case DFNT_UINT8:  for (k=0; k<size; k++) if (xuint8[k] != fill.uint8) xuint32[k] = xuint8[k];
                                                                            else xuint32[k] = fill.uint32;
                             break;
           case DFNT_INT16:  for (k=0; k<size; k++) if (xint16[k] != fill.int16) xuint32[k] = xint16[k];
                                                                            else xuint32[k] = fill.uint32;
                             break;
           case DFNT_UINT16: for (k=0; k<size; k++) if (xuint16[k] != fill.uint16) xuint32[k] = xuint16[k];
                                                                              else xuint32[k] = fill.uint32;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xuint32[k] = fill.uint32;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }
         break;

      case DFNT_FLOAT32:
         switch (type1) {
           case DFNT_INT8:   for (k=0; k<size; k++) if (xint8[k] != fill.int8) xfloat32[k] = xint8[k];
                                                                          else xfloat32[k] = fill.float32;
                             break;
           case DFNT_UINT8:  for (k=0; k<size; k++) if (xuint8[k] != fill.uint8) xfloat32[k] = xuint8[k];
                                                                            else xfloat32[k] = fill.float32;
                             break;
           case DFNT_INT16:  for (k=0; k<size; k++) if (xint16[k] != fill.int16) xfloat32[k] = xint16[k];
                                                                            else xfloat32[k] = fill.float32;
                             break;
           case DFNT_UINT16: for (k=0; k<size; k++) if (xuint16[k] != fill.uint16) xfloat32[k] = xuint16[k];
                                                                              else xfloat32[k] = fill.float32;
                             break;
           case DFNT_INT32:  for (k=0; k<size; k++) if (xint32[k] != fill.int32) xfloat32[k] = xint32[k];
                                                                            else xfloat32[k] = fill.float32;
                             break;
           case DFNT_UINT32: for (k=0; k<size; k++) if (xuint32[k] != fill.uint32) xfloat32[k] = xuint32[k];
                                                                              else xfloat32[k] = fill.float32;
                             break;
           case DFNT_NONE:   for (k=0; k<size; k++) xfloat32[k] = fill.float32;  break;
           default:          if (verbose)  printf(".\n");
                             printf("  Don't know how to convert data type %s into %s. File ignored.\n\n", typename[type1], typename[type2]);
                             return -1;
         }

    }

    if (*data) free(*data);
    *data = newdata;
    if (verbose) printf(" into %s.\n", typename[type2]);

  }
  return 0;

}




void setfillvalue(void *data, int idx0, int size, int32 num_type, int offset)
{
int idx;

    if (num_type == DFNT_NONE) return;

    for (idx=idx0; idx<idx0+size; idx++)
    switch (num_type) {
      case DFNT_INT8:    ((int8    *) data) [idx] = fill.int8    + offset;  break;
      case DFNT_UINT8:
        if ( ( ! MOD09CRSQA  ||					/* Make sure grid or coastline does not overwrite fire pixel */
               ! FIRE  ||
               ((uint8   *) data) [idx] == fill.uint8  ||
               ((uint8   *) data) [idx] < 128 ) &&
             ( ! MOD14CRS ||
               ((uint8   *) data) [idx] == fill.uint8  ||
               ((uint8   *) data) [idx] == 0  ||
               ((uint8   *) data) [idx] > 25 ) )
          ((uint8   *) data) [idx] = fill.uint8   + offset;
        break;
      case DFNT_INT16:   ((int16   *) data) [idx] = fill.int16   + offset;  break;
      case DFNT_UINT16:  ((uint16  *) data) [idx] = fill.uint16  + offset;  break;
      case DFNT_INT32:   ((int32   *) data) [idx] = fill.int32   + offset;  break;
      case DFNT_UINT32:  ((uint32  *) data) [idx] = fill.uint32  + offset;  break;
      case DFNT_FLOAT32: ((float32 *) data) [idx] = fill.float32 + offset;  break;
      default: printf("  Don't know what fill value to use for this data type (%s).\n", typename[num_type]);
               break;
    }
}




int alloc_new_array(void **data, int size, int32 num_type, void *fillvalue)
{
int k, sizeoftype;

printf("arena %.1f Mbytes\n", (double) mallinfo().arena / 1e6);
     if ( check_memory(size * (sizeoftype = DFKNTsize(num_type))) == -1  ||
          (*data = (void *) malloc(size * (sizeoftype = DFKNTsize(num_type)))) == NULL ) {	/* Allocate a new array */
       printf("  Sorry, can't allocate %.1f Mbytes of memory.\n", size * sizeoftype / 1e6);
       return -1;
     }
printf("arena %.1f Mbytes", (double) mallinfo().arena / 1e6);
printf(" (%.1f Mbytes allocated)\n", size * sizeoftype/1e6);

     if (! fillvalue)
       switch (num_type) {			/* Initialize array with generic fill value */
         case DFNT_INT8:    for (k=0; k<size; k++) ((int8    *) *data) [k] = fill.int8;     break;
         case DFNT_UINT8:   for (k=0; k<size; k++) ((uint8   *) *data) [k] = fill.uint8;    break;
         case DFNT_INT16:   for (k=0; k<size; k++) ((int16   *) *data) [k] = fill.int16;    break;
         case DFNT_UINT16:  for (k=0; k<size; k++) ((uint16  *) *data) [k] = fill.uint16;   break;
         case DFNT_INT32:   for (k=0; k<size; k++) ((int32   *) *data) [k] = fill.int32;    break;
         case DFNT_UINT32:  for (k=0; k<size; k++) ((uint32  *) *data) [k] = fill.uint32;   break;
         case DFNT_FLOAT32: for (k=0; k<size; k++) ((float32 *) *data) [k] = fill.float32;  break;
         default: printf("  Don't know how to initialize this data type (%s). File ignored.\n", typename[num_type]);
                  return -1;
       }
     else
       switch (num_type) {			/* Initialize array with specific fill value */
         case DFNT_INT8:    for (k=0; k<size; k++) ((int8    *) *data) [k] = *(int8    *)fillvalue;  break;
         case DFNT_UINT8:   for (k=0; k<size; k++) ((uint8   *) *data) [k] = *(uint8   *)fillvalue;  break;
         case DFNT_INT16:   for (k=0; k<size; k++) ((int16   *) *data) [k] = *(int16   *)fillvalue;  break;
         case DFNT_UINT16:  for (k=0; k<size; k++) ((uint16  *) *data) [k] = *(uint16  *)fillvalue;  break;
         case DFNT_INT32:   for (k=0; k<size; k++) ((int32   *) *data) [k] = *(int32   *)fillvalue;  break;
         case DFNT_UINT32:  for (k=0; k<size; k++) ((uint32  *) *data) [k] = *(uint32  *)fillvalue;  break;
         case DFNT_FLOAT32: for (k=0; k<size; k++) ((float32 *) *data) [k] = *(float32 *)fillvalue;  break;
         default: printf("  Don't know how to initialize this data type (%s). File ignored.\n", typename[num_type]);
                  return -1;
       }

     return 0;

}




void get_new_type(void *data, int size, int32 num_type, double *min, double *max, int32 *new_num_type)
{

   get_min_max(data, size, num_type, min, max);
   if (num_type == DFNT_FLOAT32)  *new_num_type = DFNT_FLOAT32;
   else if (*min >= SCHAR_MIN && *max <= SCHAR_MAX) *new_num_type = DFNT_INT8;
   else if (*min >= 0         && *max <= UCHAR_MAX) *new_num_type = DFNT_UINT8;
   else if (*min >= SHRT_MIN  && *max <= SHRT_MAX)  *new_num_type = DFNT_INT16;
   else if (*min >= 0         && *max <= USHRT_MAX) *new_num_type = DFNT_UINT16;
   else if (*min >= INT_MIN   && *max <= INT_MAX)   *new_num_type = DFNT_INT32;
   else if (*min >= 0         && *max <= UINT_MAX)  *new_num_type = DFNT_UINT32;
   else *new_num_type = DFNT_NONE;

}




void get_min_max(void *data, int size, int32 num_type, double *min, double *max)
{
int k;
int8    *xint8,    min_int8,    max_int8;
uint8   *xuint8,   min_uint8,   max_uint8;
int16   *xint16,   min_int16,   max_int16;
uint16  *xuint16,  min_uint16,  max_uint16;
int32   *xint32,   min_int32,   max_int32;
uint32  *xuint32,  min_uint32,  max_uint32;
float32 *xfloat32, min_float32, max_float32;

   switch (num_type) {

     case DFNT_INT8:    xint8 = data;
                        min_int8 = max_int8 = xint8[0];
                        for(k=0; k<size; k++) {
                          if (xint8[k] > max_int8) max_int8 = xint8[k];
                          if (xint8[k] < min_int8) min_int8 = xint8[k];
                        }
                        if (min_int8 < *min) *min = min_int8;
                        if (max_int8 > *max) *max = max_int8;
                        break;
     case DFNT_UINT8:   xuint8 = data;
                        min_uint8 = max_uint8 = xuint8[0];
                        for(k=0; k<size; k++) {
                          if (xuint8[k] > max_uint8) max_uint8 = xuint8[k];
                          if (xuint8[k] < min_uint8) min_uint8 = xuint8[k];
                        }
                        if (min_uint8 < *min) *min = min_uint8;
                        if (max_uint8 > *max) *max = max_uint8;
                        break;
     case DFNT_INT16:   xint16 = data;
                        min_int16 = max_int16 = xint16[0];
                        for(k=0; k<size; k++) {
                          if (xint16[k] > max_int16) max_int16 = xint16[k];
                          if (xint16[k] < min_int16) min_int16 = xint16[k];
                        }
                        if (min_int16 < *min) *min = min_int16;
                        if (max_int16 > *max) *max = max_int16;
                        break;
     case DFNT_UINT16:  xuint16 = data;
                        min_uint16 = max_uint16 = xuint16[0];
                        for(k=0; k<size; k++) {
                          if (xuint16[k] > max_uint16) max_uint16 = xuint16[k];
                          if (xuint16[k] < min_uint16) min_uint16 = xuint16[k];
                        }
			//printf("min is %f and max is %f min_uint16 is %d and max_uint16 is %d\n",*min, *max, min_uint16, max_uint16);
                        if (min_uint16 < *min) *min = min_uint16;
                        if (max_uint16 > *max) *max = max_uint16;
                        break;
     case DFNT_INT32:   xint32 = data;
                        min_int32 = max_int32 = xint32[0];
                        for(k=0; k<size; k++) {
                          if (xint32[k] > max_int32) max_int32 = xint32[k];
                          if (xint32[k] < min_int32) min_int32 = xint32[k];
                        }
                        if (min_int32 < *min) *min = min_int32;
                        if (max_int32 > *max) *max = max_int32;
                        break;
     case DFNT_UINT32:  xuint32 = data;
                        min_uint32 = max_uint32 = xuint32[0];
                        for(k=0; k<size; k++) {
                          if (xuint32[k] > max_uint32) max_uint32 = xuint32[k];
                          if (xuint32[k] < min_uint32) min_uint32 = xuint32[k];
                        }
                        if (min_uint32 < *min) *min = min_uint32;
                        if (max_uint32 > *max) *max = max_uint32;
                        break;
     case DFNT_FLOAT32: xfloat32 = data;
                        min_float32 = max_float32 = xfloat32[0];
                        for(k=0; k<size; k++) {
                          if (xfloat32[k] > max_float32) max_float32 = xfloat32[k];
                          if (xfloat32[k] < min_float32) min_float32 = xfloat32[k];
                        }
                        if (min_float32 < *min) *min = min_float32;
                        if (max_float32 > *max) *max = max_float32;
                        break;
     default: printf("  Unable to compute min and max for this data type.\n");
              break;

   }

}




int search_file(char *filename, char *pathlist)
{
char fullname[MAXFILELENGTH]="\0", path[MAXFILELENGTH];
char *ptr;
int j, k, nitems, status;
#ifdef DEC
struct dirent **content;
#else
struct direct **content;
#endif
struct stat statbuf;

   ptr = pathlist;
   while ( sscanf(ptr, "%s", fullname) == 1 ) {
     strcat(fullname, "/");
     strcat(fullname, filename);
     status = stat(fullname, &statbuf);
     if (status == 0  ||  errno != ENOENT) {			/* ENOENT = No such file or directory */
       strcpy(filename, fullname);
       return 0;
     }
     ptr = strchr(ptr, 32);			/* Points on the next space character */
     if (! ptr++) break;
   }

   if (options.depth) {
     ptr = pathlist;
     while ( sscanf(ptr, "%s", path) == 1 ) {
       if ( (nitems = scandir(path, &content, NULL , NULL)) == -1 ) continue;
       for (k=0; k<nitems; k++) {
         sprintf(fullname, "%s/%s", path, content[k]->d_name);
         if ( stat(fullname, &statbuf) != -1  &&
              (statbuf.st_mode & S_IFMT) == S_IFDIR  &&		/* Keep directories only */
              strcmp(content[k]->d_name, ".") != 0  &&		/* Skip .  */
              strcmp(content[k]->d_name, "..") != 0  &&		/* Skip .. */
              search_file(filename, fullname) != -1 )
           break;				/* Just exit the loop - Filename is correct */
       }					/* Don't return before freeing the structure */
       for (j=0; j<nitems; j++) free(content[j]);
       free(content);
       if (k < nitems) return 0;
       ptr = strchr(ptr, 32);			/* Points on the next space character */
       if (! ptr++) break;
     }
   }

   if ( strcmp(filename + strlen(filename) - 5, "*.hdf") == 0 ) {
     ptr = pathlist;
     while ( sscanf(ptr, "%s", path) == 1 ) {
       if ( (nitems = scandir(path, &content, NULL , NULL)) != -1 ) {
         for (k=0; k<nitems; k++) {
           sprintf(fullname, "%s/%s", path, content[k]->d_name);
           if ( stat(fullname, &statbuf) != -1  &&
                (statbuf.st_mode & S_IFMT) == S_IFREG ) {		/* Keep files only */
             if ( strncmp(content[k]->d_name, filename, strlen(filename) - 5) == 0  &&
                  strcmp(content[k]->d_name + strlen(content[k]->d_name) - 4, ".hdf") == 0 ) {
               strcpy(filename, fullname);	/* Just copy the fullname and exit the loop */
               break;				/* Don't return before freeing the structure */
             }
           }
         }
         for (k=0; k<nitems; k++) free(content[k]);
         free(content);
         if ( strcmp(filename, fullname) == 0 ) return 0;
       }
       ptr = strchr(ptr, 32);			/* Points on the next space character */
       if (! ptr++) break;
     }
   }
   return -1;
}




int mask_production_date(char *filename)
{
char *ptr;

   if ( (ptr = strstr(filename, ".250m.hdf")) != NULL  ||
        (ptr = strstr(filename, ".pseudo250m.hdf")) != NULL  ||
        (ptr = strstr(filename, ".500m.hdf")) != NULL  ||
        (ptr = strstr(filename, ".1km.hdf")) != NULL  ||
        (ptr = strstr(filename, ".2km.hdf")) != NULL ) {
     strcpy(ptr, ".hdf");
     return 0;
   }
   if ( (ptr = strstr(filename, ".hdf")) != NULL  &&
        ptr - 18 > filename  &&
        *(ptr - 18) == 46 ) {
     sprintf(ptr - 18, ".*%s", ptr);
     return 0;
   }
   return -1;
}




int substitute(char *filename, char *prefix1, char *prefix2)
{
char *ptr, *string;

  if ( (ptr = strstr(filename, prefix1)) != NULL ) {
    string = strdup(filename);
    *(string + (ptr - filename)) = '\0';
    strcat(string, prefix2);
    strcat(string, filename + strlen(prefix1));
    strcpy(filename, string);
    return 0;
  }
  return -1;
}




int read_and_convert_line(PRODUCT *input, int firstline, int nlines, PROJECTION *proj)
{
int32 new_num_type;
int idx;
int irow, jcol, rowsperkm;
int lastgoodjcol, nextgoodjcol;

  if ( read_SDS(&input->sds, firstline, nlines) == -1 ) return -1;
  if (L1B && read_SDS(&input->sz, firstline, nlines) == -1 ) return -1; /* Added 07/13 */

  if (FILLMISSINGDATA) {		/* Quick fix to eliminate dropped packets due to bit flips */
    rowsperkm = (int)(10 * input->sds.Np / 1354. + 0.5) / 10;	/* Number of rows per km */
    for (irow=0; irow<nlines; irow++) {
      lastgoodjcol = 0;
      for (jcol=0; jcol<input->sds.Np; jcol++) {
        idx = irow * input->sds.Np + jcol;
        nextgoodjcol = MIN(lastgoodjcol + rowsperkm + 1, input->sds.Np - 1);
        switch (input->sds.num_type) {
          case DFNT_UINT16:
            if (((uint16 *) input->sds.data)[idx] == MISSING) {
              if ( jcol - lastgoodjcol - 1 < rowsperkm / 2  ||
                   ((uint16 *) input->sds.data)[nextgoodjcol] == MISSING )
                ((uint16 *) input->sds.data)[idx] = ((uint16 *) input->sds.data)[lastgoodjcol];
              else
                ((uint16 *) input->sds.data)[idx] = ((uint16 *) input->sds.data)[nextgoodjcol];
            } else lastgoodjcol = jcol;
            break;
          case DFNT_INT16:
            if ( ((int16 *) input->sds.data)[idx] == MISSING - 32768  ||	/* If handled properly MISSING becomes 32766 in int16 */
                 ((int16 *) input->sds.data)[idx] == MISSING - 65536 ) {	/* Fill values not handled properly in MOD09 - MISSING becomes -2 */
              if ( jcol - lastgoodjcol - 1 < rowsperkm / 2  ||
                   ((int16 *) input->sds.data)[nextgoodjcol] == MISSING - 32768  ||
                   ((int16 *) input->sds.data)[nextgoodjcol] == MISSING - 65536 )
                ((int16 *) input->sds.data)[idx] = ((int16 *) input->sds.data)[lastgoodjcol];
              else
                ((int16 *) input->sds.data)[idx] = ((int16 *) input->sds.data)[nextgoodjcol];
            } else lastgoodjcol = jcol;
            break;
          default: break;
        }
      }
    }
  }

  if (input->maskMSbit)
    for (idx=0; idx<input->sds.Np * nlines; idx++)
      switch (input->sds.num_type) {
        case DFNT_UINT16:
          ((uint16 *) input->sds.data)[idx] &= 32767;
          if (input->sds.fillvalue)
            *(uint16 *) input->sds.fillvalue &= 32767;
          break;
        case DFNT_INT16:
          ((int16  *) input->sds.data)[idx] &= 32767;
          if (input->sds.fillvalue)
            *(int16 *) input->sds.fillvalue &= 32767;
          break;
        default: break;
      }

  get_new_type(input->sds.data, input->sds.Np, input->sds.num_type, &proj->sds.min, &proj->sds.max, &new_num_type);

  if (proj->sds.num_type == DFNT_NONE)		/* No data exists yet */
    new_num_type = input->sds.num_type;		/* Use the input data type */

  if (convert_datatype(&proj->sds, new_num_type) == -1)
    return -1;

  return 0;

}




int convert_datatype(SDS *sds, int32 new_num_type)
{
int totalsize;

  /* printf("sds numtype is %d and new type is %d\n",sds->num_type,new_num_type);*/
  if (sds->num_type == DFNT_FLOAT32)	/* Current projection data type is float */
    new_num_type = DFNT_FLOAT32;	/* Force new type to float, regardless of min and max */

  if ( new_num_type != DFNT_NONE  &&		/* Valid new type was identified */
       new_num_type != sds->num_type  &&	/* New type is different from current type */
       DFKNTsize(new_num_type) > DFKNTsize(sds->num_type) ) {		/* New type need more space */
    printf("  Data range [%g,%g] of SDS \"%s\" may not fit in current data type (%s). Conversion needed.\n",
           sds->min, sds->max, sds->name, typename[sds->num_type]);
    if ( sds->data  &&			/* Convert current array (if already exists) */
         convert_data(&sds->data, sds->Nl * sds->Np, sds->num_type, new_num_type, TRUE) == -1 )
      return -1;
    printf("  SDS \"%s\" converted into %s.\n", sds->name, typename[new_num_type]);
    if ( new_num_type != DFNT_NONE  &&
         DFKNTsize(sds->num_type) != DFKNTsize(new_num_type)) {	/* Size has changed because of conversion */
      totalsize = sds->Nl * sds->Np * DFKNTsize(new_num_type);
      printf("  Using now %.1f Mbytes of memory.\n", totalsize / 1e6);
    }
    sds->num_type = new_num_type;	/* Update the current data type */

  }

  return 0;
}





int get_tileID(char *allmetadata, PRODUCT *tile)
{
int step, Nl_globalgrid, Np_globalgrid;
int tileid;

       if ( get_metadata(allmetadata, "TILEID", &tileid, NULL, NULL) != -1  ||
            get_metadata(allmetadata, "TileID", &tileid, NULL, NULL) != -1  ||
            get_metadata(allmetadata, "TileId", &tileid, NULL, NULL) != -1 ) {		/* Version 2.1 */

           printf("  Found tile ID: %d\n", tileid);
           if (((int)(tileid / 1e7) != 5) && (((int)(tileid / 1e7) != 6))) {
             printf("  Input projection not supported. File ignored.\n\n");
             return -1;
           }
           tileid -= 1e7 * (int)(tileid / 1e7);
           printf("  Tile ID: %d\n", tileid);
           if ( (step = tileid / 1e6) != 1 && step != 2 && step != 4 ) {
             printf("  Unexpected tile size. File ignored.\n\n");
             return -1;
           }
           tileid -= 1e6 * step;
           printf("  Tile ID: %d\n", tileid);
           tile->ih = tileid / 1e3;
           tile->iv = tileid - 1e3 * tile->ih;
           if ( tile->ih < 0 || tile->ih > step * 36 - 1 ||
                tile->iv < 0 || tile->iv > step * 18 - 1 ) {
             printf("  Unable to find suitable tile coordinates. File ignored.\n\n");
             return -1;
           }
           tile->xpixsz = tile->ypixsz = 10. / step / tile->sds.Nl;
           printf("  Presumed resolution: %g km (%g deg.)\n", tile->xpixsz * DEG2KM, tile->xpixsz);

       } else if ( get_metadata(allmetadata, "\"HORIZONTALTILENUMBER\"\n", &tile->ih, NULL, NULL) != -1 &&	/* Version 2.0 */
                   get_metadata(allmetadata, "\"VERTICALTILENUMBER\"\n",   &tile->iv, NULL, NULL) != -1 ) {

           printf("  Found HORIZONTALTILENUMBER=%d and VERTICALTILENUMBER=%d.\n", tile->ih, tile->iv);
           if ( get_metadata(allmetadata, " GLOBALGRIDROWS\n",    &Nl_globalgrid, NULL, NULL) != -1 &&	/* In one-resolution products */
                get_metadata(allmetadata, " GLOBALGRIDCOLUMNS\n", &Np_globalgrid, NULL, NULL) != -1 )
             printf("  Found GLOBALGRIDROWS=%d and GLOBALGRIDCOLUMNS=%d.\n", Nl_globalgrid, Np_globalgrid);
           else {
             Nl_globalgrid = 18 * tile->sds.Nl;
             Np_globalgrid = 36 * tile->sds.Np;
             printf("  Unable to find suitable tile dimensions. Full tile size is assumed.\n");
           }
           if (2 * tile->sds.Np / tile->sds.Nl != Np_globalgrid / Nl_globalgrid) {
             printf("  Horizontal and vertical resolution in tile are different. File ignored.\n\n");
             return -1;
           }
           tile->xpixsz = tile->ypixsz = 180. / Nl_globalgrid;
           printf("  Presumed resolution: %g km (%g deg.)\n", tile->xpixsz * DEG2KM, tile->xpixsz);

       } else {

           printf("  Unable to find position of the tile in the global grid. File ignored.\n\n");
           return -1;

       }

       return 0;
}




void update_box(BOUNDINGBOX *box, int irow, int jcol)
{
        if (irow < box->l1) box->l1 = irow;
        if (irow > box->l2) box->l2 = irow;
        if (jcol < box->p1) box->p1 = jcol;
        if (jcol > box->p2) box->p2 = jcol;
}




int get_estimbox(PROJECTION proj, PRODUCT input, BOUNDINGBOX *estimHbox)
{
int k, irow, jcol, Nbad=0;
double lon, lat;
double xproj, yproj;

  if (proj.domain == USERBOX) {
    if (estimHbox) {
      estimHbox->l1 = proj.bbox.l1;
      estimHbox->p1 = proj.bbox.p1;
      estimHbox->l2 = proj.bbox.l2;
      estimHbox->p2 = proj.bbox.p2;
    }
    return 0;
  }

  for (k=0; k<4; k++) {				/* Estimate the bounding box from the Gring */

    lon = input.gringlon[k];
    lat = input.gringlat[k];
    if ( fabs(lon) >= 180  ||  fabs(lat) >= 90 ) continue;

    if ( get_xyproj(lon, lat, &xproj, &yproj, &proj.param) == -1 ) {	/* We need all 4 points */
      Nbad++;
      if (estimHbox) {
        estimHbox->l1 = 0;			/* Return the whole domain */
        estimHbox->l2 = proj.Nl_glb - 1;
        estimHbox->p1 = 0;
        estimHbox->p2 = proj.Np_glb - 1;
      }
      if (k < 3) continue;		/* Don't return right away - Must know if all 4 are missing */
      else if (Nbad == 4) return -1;	/* Skip to next Gring point instead (unless it's the last one) */
      else return 0;
    }

    if (estimHbox) {
      irow = - yproj / proj.pixsz + proj.row0 + proj.allocbox.l1 + 0.5;
      jcol =   xproj / proj.pixsz + proj.col0 + proj.allocbox.p1 + 0.5;
      update_box(estimHbox, irow, jcol);
    }

    if ( fabs(lat) > 65  &&		/* The first or last scans exceed the right or left bound of the granule */
         ( proj.param.type == SINUSOIDAL ||
           proj.param.type == INTEGERSIN )  ||
         fabs(lat) > 60 &&
         ( proj.param.type == HAMMER )  ||
         fabs(lat) > 80 &&
         ( proj.param.type == MOLLWEIDE ||
           proj.param.type == ROBINSON ||
           proj.param.type == ECKERTIV )  ||
         fabs(lat) < 30 &&
         ( proj.param.type == MOLLWEIDE ||
           proj.param.type == ROBINSON ||
           proj.param.type == HAMMER ||
           proj.param.type == ECKERTIV ) ) {
      if (estimHbox) {
        estimHbox->p1 = 0;
        estimHbox->p2 = proj.Np_glb - 1;
      }
    }

    if ( fabs(lat) > 85  &&
         proj.param.type == ORTHOGRAPHIC  &&
         estimHbox ) {
      if (lat > 0) estimHbox->l1 = 0;
             else  estimHbox->l2 = proj.Nl_glb - 1;
    }


  }

  if (input.bounds[0] != PROJUNDEF) {		/* Use bounds too if they have been read successfully */

    lat = input.bounds[0];			/* Compute top line from North bounding coordinate */
    if ( proj.param.type == HAMMER  &&
         lat >= 0 )
      lon = 180.;				/* In North Hammer top line is at lon=+/-180 */
    else
      lon = 0;					/* All other cases: top line is at lon=0 */
    if ( get_xyproj(lon, lat, &xproj, &yproj, &proj.param) != -1 ) {
      if (estimHbox) {
        irow = - yproj / proj.pixsz + proj.row0 + proj.allocbox.l1 + 0.5;
        if (irow < estimHbox->l1) estimHbox->l1 = irow;
      }
    }

    lat = input.bounds[1];			/* Compute bottom line from South bounding coordinate */
    if ( proj.param.type == HAMMER
         &&  lat <= 0 )
      lon = 180.;				/* In South Hammer bottom line is at lon=+/-180 */
    else
      lon = 0;					/* All other cases: bottom line is at lon=0 */
    if ( get_xyproj(lon, lat, &xproj, &yproj, &proj.param) != -1 ) {
      if (estimHbox) {
        irow = - yproj / proj.pixsz + proj.row0 + proj.allocbox.l1 + 0.5;
        if (irow > estimHbox->l2) estimHbox->l2 = irow;
      }
    }

    if ( input.bounds[2] == 180  &&  input.bounds[3] == -180  ||	/* Polar granule */
         input.bounds[2] < input.bounds[3] ) {			/* Granule across the dateline */
      if (estimHbox) {
        estimHbox->p1 = 0;
        estimHbox->p2 = proj.Np_glb - 1;
      }
    }

  }

  return 0; 

}




void update_projection(PROJECTION *proj, PRODUCT input)
{
int margin, bufsize;
BOUNDINGBOX bbox, estimbox;
/* Tolerance must be very small or there will be an overlap between neighbor pixels */

  get_step_and_offset(input.sds.Nl, input.Nl_geo, &input.rowoffset, &input.rowstep, input.projparam.type, FALSE);
  get_step_and_offset(input.sds.Np, input.Np_geo, &input.coloffset, &input.colstep, input.projparam.type, TRUE);
  margin = 10. / proj->pixsz;
  if (proj->domain == USERBOX) margin = 0;

  bufsize = input.sds.maxplanes * input.sds.Np;
  if ( check_memory(bufsize * DFKNTsize(input.sds.num_type)) == -1 ) return;
  input.sds.data = (void *) malloc(bufsize * DFKNTsize(input.sds.num_type));
  
  /* Added 07/13: Allocate space for solzen sds data */
  if (L1B){
    bufsize = input.sz.maxplanes * input.sz.Np;
    if ( check_memory(bufsize * DFKNTsize(input.sz.num_type)) == -1 ) return;
    input.sz.data = (void *) malloc(bufsize * DFKNTsize(input.sz.num_type));
  }

  estimbox.l1 = bbox.l1 = proj->Nl_glb;
  estimbox.p1 = bbox.p1 = proj->Np_glb;
  estimbox.l2 = bbox.l2 = -1;
  estimbox.p2 = bbox.p2 = -1;

  proj->row0 = proj->Nl_glb / 2. + 0.5 - 1 - proj->allocbox.l1;
  proj->col0 = proj->Np_glb / 2. + 0.5 - 1 - proj->allocbox.p1;

  if ( proj->domain == GLOBAL ) {
      estimbox.l1 = 0;                        /* Consider the whole domain */
      estimbox.p1 = 0;
      estimbox.l2 = proj->Nl_glb - 1;
      estimbox.p2 = proj->Np_glb - 1;
  } else
      get_estimbox(*proj, input, &estimbox);

  printf("  Reading input SDS \"%s\"", input.sds.name);
  if (input.sds.rank == 3) {
    if (input.sds.Nplanes == 1)
      printf(" (plane %s)", input.sds.planeID[0]);
    else
      printf(" (planes %s to %s)", input.sds.planeID[0], input.sds.planeID[input.sds.Nplanes - 1]);
  }
  printf("...\n");

  if (options.interpol == NONE)		/* Forward mapping - Pixel center only - No interpolation */

      project_pixel_centers(proj, input, &bbox, &estimbox, margin);

  else {		/* Forward mapping - Nearest pixel */

      if ( SNOW  ||
	   SNOW_C6 ||  /* added SS 05/09 for C6 snow fraction */
           FIRE  ||   /* Added for MODWATER by SS-02/18/2020 */
           MOD09CRSQA  ||
           MOD14CRS ||
           WATER)
        project_pixel_centers(proj, input, &bbox, &estimbox, margin);
      nearest_pixel_interpol(proj, input, &bbox, &estimbox, margin);
  }

  if (bbox.l2 == -1 || bbox.p2 == -1)
    printf("  Warning: none of the input pixels was projected successfully.\n");

  free(input.sds.data);
  if (L1B) free(input.sz.data); /*added 07/13 */
  
  if (proj->domain != USERBOX) {
    if (bbox.l1 < proj->bbox.l1) proj->bbox.l1 = bbox.l1;
    if (bbox.l2 > proj->bbox.l2) proj->bbox.l2 = bbox.l2;
    if (bbox.p1 < proj->bbox.p1) proj->bbox.p1 = bbox.p1;
    if (bbox.p2 > proj->bbox.p2) proj->bbox.p2 = bbox.p2;
  }

}




int get_xyproj(double lon, double lat, double *xproj, double *yproj, PROJPARAM *projparam)
{
double newlon;
double dlon, lon_center, xproj_center, xprojtmp;
int kzone;

   if ( lon == PROJUNDEF  ||
        lat == PROJUNDEF ) {
     *xproj = *yproj = PROJUNDEF;
     return -1;
   }

   lon -= projparam->lon_center;
   if (projparam->lat_pole != 90.) {
     newlon = RAD2DEG * atan2(cos(lat * DEG2RAD) * sin(lon * DEG2RAD),
                                    projparam->sinlatpole * cos(lat * DEG2RAD) * cos(lon * DEG2RAD) + projparam->coslatpole * sin(lat * DEG2RAD));
     lat = RAD2DEG * asin(projparam->sinlatpole * sin(lat * DEG2RAD) - projparam->coslatpole * cos(lat * DEG2RAD) * cos(lon * DEG2RAD));
     lon = newlon - projparam->lon_pole;
   }
   if (lon < -180) lon += 360;
   if (lon >  180) lon -= 360;

   switch (projparam->type) {

     case SINUSOIDAL: *xproj = lon * cos(lat * DEG2RAD);
                      *yproj = lat;
                      break;
     case INTEGERSIN: isinusfor(lon, lat, xproj, yproj);
                      break;
     case INTERSINUS1: intersinfor1(lon, lat, xproj, yproj);
                       break;
     case INTERSINUS2: intersinfor2(lon, lat, xproj, yproj);
                       break;
     case INTERSINUS3:
       dlon = (360. / projparam->nzones);
       kzone = (180 + lon) / dlon;
       lon_center = -180 + dlon * (kzone + 0.5);
       xproj_center = -(360. + (projparam->nzones - 1) * ZONEGAP) / 2. + (dlon + ZONEGAP) * kzone + dlon / 2;
       *xproj = xproj_center + (lon - lon_center) * cos(lat * DEG2RAD);
       *yproj = lat;
       break;
     case PLATECARREE: *xproj = lon;
                       *yproj = lat;
                       break;
     case ROBINSON: robfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                    break;
     case MOLLWEIDE: molwfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                     break;
     case INTERMOLL: imolwfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                     break;
     case GOODE: goodfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                 break;
     case GOODE2: goodfor2(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                  break;
     case PERSPECTIVE:  if ( gvnspfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj) != 0 ) return -1;
                        break;
     case ORTHOGRAPHIC:  if ( orthfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj) != 0 ) return -1;
                         break;
     case GNOMONIC: if ( gnomfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj) != 0  ||
                         *xproj * *xproj + *yproj * *yproj > projparam->radius * projparam->radius )
                      return -1;
                    break;
     case STEREOGRAPHIC: if ( sterfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj) != 0  ||
                              *xproj * *xproj + *yproj * *yproj > 13131.2254)	/* 13131.2254 = 4 * DEG2RAD * DEG2RAD */
                           return -1;
                         break;
     case LAMBERT:  if ( lamazfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj) != 0  ||
                         *xproj * *xproj + *yproj * *yproj > projparam->radius * projparam->radius )
                      return -1;
                    break;
     case AZIMEQUI: if ( azimfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj) != 0  ||
                         *xproj * *xproj + *yproj * *yproj > projparam->radius * projparam->radius )
                      return -1;
                    break;
     case HAMMER:   hamfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                    break;
     case ALBERS:   alberfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                    break;
     case AITOFF:   aitfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                    break;
     case ECKERTIV: eckivfor(lon * DEG2RAD, lat * DEG2RAD, xproj, yproj);
                    break;
     case GEOGRAPHIC:  *xproj = lon;
                       *yproj = lat;
                       break;        
     default: fprintf(stderr, "Forward projection %s not implemented.\n",  projinfo[projparam->type].name);
              exit(-1);
              break;

   }

   if (projparam->xcorr != 1)
     *xproj *= projparam->xcorr;

   if (projparam->sinphi != 0  ||  projparam->cosphi != 1) {
     xprojtmp = *xproj * projparam->cosphi - *yproj * projparam->sinphi;
     *yproj =   *xproj * projparam->sinphi + *yproj * projparam->cosphi;
     *xproj = xprojtmp;
   }

   return 0;

}




int get_lonlat(double xproj, double yproj, double *lon, double *lat, PROJPARAM *projparam)
{
double newlon;
double dlon, lon_center, xproj_center, xprojtmp;
int kzone;

   if (projparam->sinphi != 0  ||  projparam->cosphi != 1) {
     xprojtmp = xproj * projparam->cosphi + yproj * projparam->sinphi;
     yproj  = - xproj * projparam->sinphi + yproj * projparam->cosphi;
     xproj  = xprojtmp;
   }

   if (projparam->xcorr != 1)
     xproj /= projparam->xcorr;

   switch (projparam->type) {
       
    case GEOGRAPHIC: *lat = yproj;
                     *lon = xproj; 
                     if (fabs(*lat) > 90) return -1;
                     if (fabs(*lon) > 180) return -1;
                     break;

     case SINUSOIDAL: *lat = yproj;
                      if (fabs(*lat) >= 90) return -1;
                      *lon = xproj / cos(*lat * DEG2RAD);
                      if (fabs(*lon) > 180) return -1;
                      break;
     case INTEGERSIN: if ( isinusinv(xproj, yproj, lon, lat, isinusparam.nzones / projparam->nzones) != 0 ) return -1;
                      break;
     case INTERSINUS1: if ( intersininv1(xproj, yproj, lon, lat) != 0 ) return -1;
                       break;
     case INTERSINUS2: if ( intersininv2(xproj, yproj, lon, lat) != 0 ) return -1;
                       break;
     case INTERSINUS3:
       *lat = yproj;
       if (fabs(*lat) >= 90) return -1;
       dlon = (360. / projparam->nzones);
/*
       kzone = (180 + xproj) / (dlon + ZONEGAP);
       kzone = ((360. + (projparam->nzones - 1) * ZONEGAP) / 2. + xproj) / (dlon + ZONEGAP);
printf("kzone %d\n",kzone);
*/
       kzone = ((360. + (projparam->nzones - 1) * ZONEGAP) / 2. + xproj) / (dlon + ZONEGAP);
       lon_center = -180 + dlon * (kzone + 0.5);
       xproj_center = -(360. + (projparam->nzones - 1) * ZONEGAP) / 2. + (dlon + ZONEGAP) * kzone + dlon / 2;
       *lon = lon_center + (xproj - xproj_center) / cos(*lat * DEG2RAD);
       if (fabs(*lon - lon_center) > dlon / 2) return -1;
       break;
     case PLATECARREE: *lat = yproj;
                       *lon = xproj;
                       if (fabs(*lat) > 90) return -1;
/*
                       if (fabs(*lat) > 90  ||  fabs(*lon) > 180) return -1;
*/
                       break;
     case ROBINSON: if ( robinv(xproj, yproj, lon, lat) != 0 ) return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case MOLLWEIDE: if ( molwinv(xproj, yproj, lon, lat) != 0 ) return -1;
                     *lat *= RAD2DEG;
                     *lon *= RAD2DEG;
                     break;
     case INTERMOLL: if ( imolwinv(xproj, yproj, lon, lat) != 0 ) return -1;
                     *lat *= RAD2DEG;
                     *lon *= RAD2DEG;
                     break;
     case GOODE: if ( goodinv(xproj, yproj, lon, lat) != 0 ) return -1;
                 *lat *= RAD2DEG;
                 *lon *= RAD2DEG;
                 break;
     case GOODE2: if ( goodinv2(xproj, yproj, lon, lat) != 0 ) return -1;
                  *lat *= RAD2DEG;
                  *lon *= RAD2DEG;
                  break;
     case PERSPECTIVE:  if ( gvnspinv(xproj, yproj, lon, lat) != 0 ) return -1;
                        *lat *= RAD2DEG;
                        *lon *= RAD2DEG;
                        break;
     case ORTHOGRAPHIC:  if ( orthinv(xproj, yproj, lon, lat) != 0 ) return -1;
                         *lat *= RAD2DEG;
                         *lon *= RAD2DEG;
                         break;
     case GNOMONIC: if ( xproj * xproj + yproj * yproj > projparam->radius * projparam->radius  ||
                         gnominv(xproj, yproj, lon, lat) != 0 )
                      return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case STEREOGRAPHIC: if ( xproj * xproj + yproj * yproj > 13131.2254  ||	/* 13131.2254 = 4 * DEG2RAD * DEG2RAD */
                              sterinv(xproj, yproj, lon, lat) != 0 )
                           return -1;
                         *lat *= RAD2DEG;
                         *lon *= RAD2DEG;
                         break;
     case LAMBERT:  if ( xproj * xproj + yproj * yproj > projparam->radius * projparam->radius  ||
                         lamazinv(xproj, yproj, lon, lat) != 0 )
                      return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case AZIMEQUI: if ( xproj * xproj + yproj * yproj > projparam->radius * projparam->radius  ||
                         aziminv(xproj, yproj, lon, lat) != 0 )
                      return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case HAMMER:   if ( haminv(xproj, yproj, lon, lat) != 0 ) return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case ALBERS:   if ( alberinv(xproj, yproj, lon, lat) != 0 ) return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case AITOFF:   if ( aitinv(xproj, yproj, lon, lat) != 0 ) return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     case ECKERTIV: if ( eckivinv(xproj, yproj, lon, lat) != 0 ) return -1;
                    *lat *= RAD2DEG;
                    *lon *= RAD2DEG;
                    break;
     default: fprintf(stderr, "Inverse projection %s not implemented.\n",  projinfo[projparam->type].name);
              exit(-1);
              break;

   }

   if (projparam->lat_pole != 90.) {
     *lon += projparam->lon_pole;
     newlon = 180 + RAD2DEG * atan2(- cos(*lat * DEG2RAD) * sin(*lon * DEG2RAD),
                                    - projparam->sinlatpole * cos(*lat * DEG2RAD) * cos(*lon * DEG2RAD) + projparam->coslatpole * sin(*lat * DEG2RAD));
     *lat = RAD2DEG * asin(projparam->sinlatpole * sin(*lat * DEG2RAD) + projparam->coslatpole * cos(*lat * DEG2RAD) * cos(*lon * DEG2RAD));
     *lon = newlon;
   }
   *lon += projparam->lon_center;
   if (*lon < -180) *lon += 360.;
   if (*lon >  180) *lon -= 360.;

   return 0;

}




void fill_domain(PROJECTION *proj)
{
double xproj, yproj, lon, lat;
int irow, jcol, Nrows, Ncols;

   printf("Outlining output projection:\n");
   printf("  Outlining the valid domain...\n");

   if ( proj->param.type == PLATECARREE ||
        proj->param.type == GEOGRAPHIC) return;		/* No fill area in PLATECARREE */

   set_proj_dimensions(proj, &Nrows, &Ncols);

   if ( proj->info.Nintlon == 0  &&			/* Non-interrupted projections */
        projinfo[proj->param.type].xprojsym  &&		/* with horizontal and vertical symmetry */
        projinfo[proj->param.type].yprojsym ) {

       for (irow=0; irow<Nrows; irow++)
         fill_sym_line(proj, irow, 0, Ncols - 1, Ncols, proj->row0, proj->col0, proj->pixsz, 0.);

   } else if (proj->param.type == INTERSINUS1) {

       for (irow=0; irow<Nrows; irow++) {
         fill_sym_line(proj, irow,
                       0, (int)(proj->col0 - proj->Np_glb * 40 / 360.),
                       Ncols, proj->row0, proj->col0 - proj->Np_glb * 110 / 360.,
                       proj->pixsz, 70.);
         fill_sym_line(proj, irow,
                       (int)(proj->col0 - proj->Np_glb * 40 / 360.) + 1, (int)(proj->col0 + proj->Np_glb * 40 / 360.),
                       Ncols, proj->row0, proj->col0,
                       proj->pixsz, 40.);
         fill_sym_line(proj, irow,
                       (int)(proj->col0 + proj->Np_glb * 40 / 360.) + 1, Ncols - 1,
                       Ncols, proj->row0, proj->col0 + proj->Np_glb * 110 / 360.,
                       proj->pixsz, 70.);
       }

   } else if (proj->param.type == INTEGERSIN) {

       for (irow=0; irow<Nrows; irow++) {
proj->row0 = proj->Nl_glb / 2. - proj->allocbox.l1;
proj->col0 = proj->Np_glb / 2. - proj->allocbox.p1;
         lat = (proj->row0 - irow - 0.5) * proj->pixsz;
         lon = -179.999999;
         if ( get_xyproj(lon, lat, &xproj, &yproj, &proj->param) == -1 )
           continue;
         jcol = proj->col0 + xproj / proj->pixsz;
         jcol--;
         if (proj->sds.data)
           setfillvalue(proj->sds.data,  irow * Ncols, jcol + 1, proj->sds.num_type,  DOMAINOFFSET);
         if (proj->meta.data)
           setfillvalue(proj->meta.data, irow * Ncols, jcol + 1, proj->meta.num_type, DOMAINOFFSET);
         lon = 179.999999;
         if ( get_xyproj(lon, lat, &xproj, &yproj, &proj->param) == -1 )
           continue;
         jcol = proj->col0 + xproj / proj->pixsz;
         jcol++;
         if (proj->sds.data)
           setfillvalue(proj->sds.data,  irow * Ncols + jcol, Ncols - jcol, proj->sds.num_type,  DOMAINOFFSET);
         if (proj->meta.data)
           setfillvalue(proj->meta.data, irow * Ncols + jcol, Ncols - jcol, proj->meta.num_type, DOMAINOFFSET);
       }

   } else {

       printf("  No particular symmetry was defined. Scanning the whole domain...\n");
       for (irow=0; irow<Nrows; irow++) {
         yproj = (proj->row0 - irow) * proj->pixsz;
         for (jcol=0; jcol<Ncols; jcol++)
           if ( get_lonlat((jcol - proj->col0) * proj->pixsz, yproj, &lon, &lat, &proj->param) == -1 ) {
             if (proj->sds.data)
               setfillvalue(proj->sds.data,  irow * Ncols + jcol, 1, proj->sds.num_type,  DOMAINOFFSET);
             if (proj->meta.data)
               setfillvalue(proj->meta.data, irow * Ncols + jcol, 1, proj->meta.num_type, DOMAINOFFSET);
           }
       }

   }

}




void fill_sym_line(PROJECTION *proj, int irow, int jcolmin, int jcolmax, int Np, double row0, double col0, double pixsz, double xmax)
{
int jcol1, jcol2;
double xproj, yproj;
const double pr[20] = { -0.062,  0.0,    0.062,  0.124,  0.186,  0.248,  0.31,   0.372,  0.434,  0.4958,
                         0.5571, 0.6176, 0.6769, 0.7346, 0.7903, 0.8435, 0.8936, 0.9394, 0.9761, 1.0 };
const double xlr[20] = { 0.9986, 1.0,    0.9986, 0.9954, 0.99,   0.9822, 0.973,  0.96,   0.9427, 0.9216,
                         0.8962, 0.8679, 0.835,  0.7986, 0.7597, 0.7186, 0.6732, 0.6213, 0.5722, 0.5322 };
double p2, a, b, c;
int ip1;

  /* Note: Although *proj is not modified in this routine, it may be significantly */
  /* faster to pass the address of this structure instead of the whole big structure */

         if (jcolmax > Np - 1) jcolmax = Np - 1;
         if (jcolmin < 0) jcolmin = 0;
         if (jcolmax < jcolmin) return;

         yproj = (row0 - irow) * pixsz;
         switch (proj->param.type) {
           case SINUSOIDAL:    xproj = 180  * cos(yproj * DEG2RAD);
                               break;
           case INTERSINUS1:   xproj = xmax * cos(yproj * DEG2RAD);
                               break;
           case MOLLWEIDE:     yproj *= DEG2RAD;
                               xproj = 162.05693690827909251987 * sqrt(1 - yproj * yproj / 2);
                               break;		/* 162.05693690827909251987 = 180 * 2 * sqrt(2.) / PI */
           case ROBINSON:      ip1 = (int)fabs(yproj / 5.) - 1;
                               a = (pr[ip1 + 3] - 2.0 * pr[ip1 + 2] + pr[ip1 + 1]) / 2.0;
                               b = (pr[ip1 + 3] - pr[ip1 + 1]) / 2.0;
                               c = pr[ip1 + 2] - fabs(yproj) / 90;
                               if (a != 0) p2 = (- b + sqrt(b * b - 4 * a * c)) / 2 / a;
                                      else p2 = - c / b;
                               a = (xlr[ip1 + 3] - 2.0 * xlr[ip1 + 2] + xlr[ip1 + 1]) / 2.0;
                               b = (xlr[ip1 + 3] - xlr[ip1 + 1]) / 2.0;
                               c = xlr[ip1 + 2];
                               xproj = 0.9858 * 180 * (a * p2 * p2 + b * p2 + c);
                               break;
           case HAMMER:        xproj = sqrt(26262.4508 - 4 * yproj * yproj);	/* 26262.4508 = 8 * DEG2RAD * DEG2RAD */
                               break;
           case ECKERTIV:      yproj *= DEG2RAD;
                               xproj = 76.00287605683881626729  * (1 + sqrt(1 - yproj * yproj / 1.75960338595377049635));
                               break;		/* 76.00287605683881626729 = 180 * 4 * sqrt(PI / (4 + PI)) / PI */
						/* 1.75960338595377049635  = 4 * PI / (4 + PI) */
           case PERSPECTIVE:   xproj = sqrt(3282.80635001174379478648 / (1. + 2. * DEG2KM * RAD2DEG / proj->param.height) - yproj * yproj);
                               break;		/* 3282.8063500 = RAD2DEG * RAD2DEG */
           case ORTHOGRAPHIC:  xproj = sqrt(3282.80635001174379478648 - yproj * yproj);
                               break;		/* 3282.8063500 = RAD2DEG * RAD2DEG */
           case GNOMONIC:      xproj = sqrt(proj->param.radius * proj->param.radius - yproj * yproj);
                               break;
           case STEREOGRAPHIC: xproj = sqrt(13131.2254 - yproj * yproj);	/* 13131.2254 = 4 * DEG2RAD * DEG2RAD */
                               break;
           case LAMBERT:       xproj = sqrt(proj->param.radius * proj->param.radius - yproj * yproj);
                               break;
           case AZIMEQUI:      xproj = sqrt(proj->param.radius * proj->param.radius - yproj * yproj);
                               break;
           default: printf("  Don't know how to outline the valid domain for that projection.\n");
                    return;
         }
         jcol1 = jcolmin;
         jcol2 = - xproj / pixsz + col0 + 0.5;
         jcol2--;
         if (jcol2 > jcolmax) jcol2 = jcolmax;
         if (jcol2 >= jcol1) {
           if (proj->sds.data)
             setfillvalue(proj->sds.data,  irow * Np + jcol1, jcol2 - jcol1 + 1, proj->sds.num_type,  DOMAINOFFSET);
           if (proj->meta.data)
             setfillvalue(proj->meta.data, irow * Np + jcol1, jcol2 - jcol1 + 1, proj->meta.num_type, DOMAINOFFSET);
         }
         jcol1 = xproj / pixsz + col0 + 0.5;
         jcol1++;
         if (jcol1 < jcolmin) jcol1 = jcolmin;
         jcol2 = jcolmax;
         if (jcol2 >= jcol1) {
           if (proj->sds.data)
             setfillvalue(proj->sds.data,  irow * Np + jcol1, jcol2 - jcol1 + 1, proj->sds.num_type,  DOMAINOFFSET);
           if (proj->meta.data)
             setfillvalue(proj->meta.data, irow * Np + jcol1, jcol2 - jcol1 + 1, proj->meta.num_type, DOMAINOFFSET);
         }
}




void add_vectors(void *data, int32 num_type, PROJECTION proj, VECTOR vector)
{
int j, k, bufsize, irow, jcol, Hvectoridx, idx;
int Nrows, Ncols;
unsigned long nitems;
float *lonlat;
double xproj, yproj, lon, lat;
double lon1, lon2, lat1, lat2, xproj1, yproj1, xproj2, yproj2;
char  filename[MAXFILELENGTH];
FILE *fd;
static unsigned char first_time=1, prev_num_type=DFNT_NONE;
static int *Hvector[NVECTORS];
int Hvectorsize;

   if (vector.request == 0) return;

   set_proj_dimensions(&proj, &Nrows, &Ncols);

   if ( proj.pixsz * DEG2KM < 1.  ||		/* Resolution lower than 1km */
        vector.drawline == TRUE ) {
     if ( check_memory((bufsize = 2 * 50000) * sizeof(float)) == -1 ) return;
     lonlat = (float *) malloc(bufsize * sizeof(float));	/* Buffer size must be even number */
     for (k=0; k<NVECTORS; k++) {
       if ( ! (vector.request & 1 << k) ) continue;		/* Skip to next vector layer */
       sprintf(filename, "vector%d", k+1);
       if ( search_file(filename, vector.path) == -1 ) {
         printf("  Unable to find file containing %s in path specified by $%s.\n", vector.type[k], ANCPATHVAR);
         break;					/* Skip all vector layers */
       }
       if ( (fd = fopen(filename, "r")) == NULL )
         printf("  Can't open ancillary file to add %s.\n", vector.type[k]);
       else {
         printf("  Drawing %s (draw technique)...\n", vector.type[k]);
         xproj1 = yproj1 = VECTORUNDEF;
         lon1 = lat1 = VECTORUNDEF;
         while ( (nitems = fread(lonlat, sizeof(float), bufsize, fd)) >= 2 )	/* Read up to buffer size, instead of one by one */
           for (j=0; j<nitems; j+=2) {						/* Then loop over number of items read */
             lon2 = lonlat[j];
             lat2 = lonlat[j + 1];
             if (lat2 == VECTORUNDEF) {
               xproj1 = yproj1 = VECTORUNDEF;
               lon1 = lat1 = VECTORUNDEF;
               continue;		/* Skip to next line if VECTORUNDEF (i.e. break) in contour */
             }
             if (lon2 >  180) lon2 -= 360;
             if (lon2 < -180) lon2 += 360;
             if (lat2 >  90)  lat2 =  90;
             if (lat2 < -90)  lat2 = -90;
             if  ( get_xyproj(lon2, lat2, &xproj2, &yproj2, &proj.param) != -1  &&
                   xproj1 != VECTORUNDEF  &&  yproj1 != VECTORUNDEF  &&
                   xproj2 != VECTORUNDEF  &&  yproj2 != VECTORUNDEF  &&
                   lat1 != VECTORUNDEF  &&  lon1 != VECTORUNDEF  &&
                   check_continuity(lon1, lat1, lon2, lat2, xproj1, xproj2, &proj.param, proj.pixsz) != -1 )
               draw_xyline(data, num_type, &proj, xproj1, yproj1, xproj2, yproj2, vector.offset[k]);
             xproj1 = xproj2;
             yproj1 = yproj2;
             lon1 = lon2;
             lat1 = lat2;
           }
         fclose(fd);
       }
     }
     free(lonlat);
     return;
   }

   if (first_time  ||  num_type != prev_num_type) {

     if ( check_memory((bufsize = 2 * 50000) * sizeof(float)) == -1 ) return;
     lonlat = (float *) malloc(bufsize * sizeof(float));	/* Buffer size must be even number */
     for (k=0; k<NVECTORS; k++) {
       if ( ! (vector.request & 1 << k) ) continue;		/* Skip to next vector layer */
       sprintf(filename, "vector%d", k+1);
       if ( search_file(filename, vector.path) == -1 ) {
         printf("  Unable to find file containing %s (%s) in path specified by $%s.\n", vector.type[k], filename, ANCPATHVAR);
         break;				/* Skip all vector layers */
       }
       if ( (fd = fopen(filename, "r")) == NULL )
         printf("  Can't open ancillary file to add %s.\n", vector.type[k]);
       else {
         Hvector[k] = NULL;
         Hvectorsize = 0;
         Hvectoridx = 0;
         printf("  Outlining %s (dot technique)...\n", vector.type[k]);
         while ( (nitems = fread(lonlat, sizeof(float), bufsize, fd)) >= 2 )	/* Read up to buffer size, instead of one by one */
         for (j=0; j<nitems; j+=2) {						/* Then loop over number of items read */
           lon = lonlat[j];
           lat = lonlat[j + 1];
           if (lat == VECTORUNDEF) continue;		/* Skip to next line if VECTORUNDEF (i.e. break) in contour */
           if (lon >  180) lon -= 360;
           if (lon < -180) lon += 360;
           if (lat >  90)  lat =  90;
           if (lat < -90)  lat = -90;
           if ( get_xyproj(lon, lat, &xproj, &yproj, &proj.param) == -1 )
             continue;		/* Skip to next point */
           irow = - yproj / proj.pixsz + proj.row0 + 0.5;
           jcol =   xproj / proj.pixsz + proj.col0 + 0.5;
           if (irow >= 0 && irow < Nrows && jcol >= 0 && jcol < Ncols) {
             if (Hvectoridx >= Hvectorsize) {
               if ( check_memory((Hvectorsize += BUFINC) * sizeof(int)) == -1 ) break;
               Hvector[k] = (int *) realloc(Hvector[k], Hvectorsize * sizeof(int));
             }
             Hvector[k][Hvectoridx++] = irow * Ncols + jcol;
           }
         }
         fclose(fd);
         if (Hvectoridx > 0) Hvector[k][0] = Hvectoridx;		/* Hvector[k] is NULL if Hvectoridx = 0 */
       }
     }
     free(lonlat);

     first_time = 0;
     prev_num_type = num_type;

  }

   for (k=0; k<NVECTORS; k++)
     if (Hvector[k])
       for (j=1; j<Hvector[k][0]; j++) {
/*
         if ( Hvector[k][j] >= proj.row1 * Ncols  &&  Hvector[k][j] < proj.row2 * Ncols ) {
           idx = Hvector[k][j] - proj.row1 * Ncols;
*/
         idx = Hvector[k][j] - proj.row1 * Ncols;
         if ( idx >= 0  &&
              idx < (proj.row2 - proj.row1) * Ncols ) {
           if ( ! FILLBURN  ||
                testfillvalue(data, idx, num_type, 0, NULL)  ||
                testfillvalue(data, idx, num_type, DOMAINOFFSET, NULL) )
             setfillvalue(data, idx, 1, num_type, vector.offset[k]);
         }
       }

}




char testfillvalue(void *data, int idx, int32 num_type, int offset, void *fillvalue)
{
   if (! data)
     return TRUE;
   if (! fillvalue)
     switch (num_type) {
       case DFNT_INT8:    if ( ((int8    *) data) [idx] == fill.int8    + offset ) return TRUE;  break;
       case DFNT_UINT8:   if ( ((uint8   *) data) [idx] == fill.uint8   + offset ) return TRUE;  break;
       case DFNT_INT16:   if ( ((int16   *) data) [idx] == fill.int16   + offset ) return TRUE;  break;
       case DFNT_UINT16:  if ( ((uint16  *) data) [idx] == fill.uint16  + offset ) return TRUE;  break;
       case DFNT_INT32:   if ( ((int32   *) data) [idx] == fill.int32   + offset ) return TRUE;  break;
       case DFNT_UINT32:  if ( ((uint32  *) data) [idx] == fill.uint32  + offset ) return TRUE;  break;
       case DFNT_FLOAT32: if ( ((float32 *) data) [idx] == fill.float32 + offset ) return TRUE;  break;
       default: break;
     }
   else
     switch (num_type) {
       case DFNT_INT8:    if ( ((int8    *) data) [idx] == *(int8    *)fillvalue ) return TRUE;  break;
       case DFNT_UINT8:   if ( ((uint8   *) data) [idx] == *(uint8   *)fillvalue ) return TRUE;  break;
       case DFNT_INT16:   if ( ((int16   *) data) [idx] == *(int16   *)fillvalue ) return TRUE;  break;
       case DFNT_UINT16:  if ( ((uint16  *) data) [idx] == *(uint16  *)fillvalue ) return TRUE;  break;
       case DFNT_INT32:   if ( ((int32   *) data) [idx] == *(int32   *)fillvalue ) return TRUE;  break;
       case DFNT_UINT32:  if ( ((uint32  *) data) [idx] == *(uint32  *)fillvalue ) return TRUE;  break;
       case DFNT_FLOAT32: if ( ((float32 *) data) [idx] == *(float32 *)fillvalue ) return TRUE;  break;
       default: break;
     }
   return FALSE;
}




void outline_domain(void *data, int32 num_type, PROJECTION proj)
{
int k, kmax;
double xproj1, yproj1, xproj2, yproj2;
double a, b, angular_step;

   if ( proj.info.elliptic ) {
/*
   if ( proj.param.type == PERSPECTIVE  ||
        proj.param.type == ORTHOGRAPHIC  ||
        proj.param.type == GNOMONIC  ||
        proj.param.type == STEREOGRAPHIC  ||
        proj.param.type == LAMBERT ||
        proj.param.type == AZIMEQUI ||
        proj.param.type == MOLLWEIDE ||
        proj.param.type == HAMMER ) {
*/
       a = proj.Np_glb / 2.;
       b = proj.Nl_glb / 2.;
   } else
       return;

   angular_step = 2. / a;		/* The max. angular step is 2 pixels */
   kmax = (ceil)(2 * PI / angular_step);

   xproj1 = a * proj.pixsz;
   yproj1 = 0;

   for (k=1; k<=kmax; k++) {
     xproj2 = a * cos(k * angular_step) * proj.pixsz;
     yproj2 = b * sin(k * angular_step) * proj.pixsz;
     draw_xyline(data, num_type, &proj, xproj1, yproj1, xproj2, yproj2, OUTLINEOFFSET);
     xproj1 = xproj2;
     yproj1 = yproj2;
   }

}




void compute_grids(PROJECTION proj, GRID *grid)
{
int k, j, ilat, ilon, ilonmin, ilonmax, ilatmin, ilatmax;
double lon1, lon2, lat1, lat2;
double xproj, yproj;
float Hlonres=1., Hlatres, latmin, latmax, intlon;
char *ptr;

  if (grid->request == 0) return;

  /* No need to oversample - pixsz is a good guess. For INTEGERSIN, oversampling */
  /* MUST be avoided anyway, or the vertical lines might be 2 pixels thick */
  Hlatres = proj.pixsz * 2;		/* twice coarser than output resolution */

  if (proj.param.type == INTERSINUS2  ||			/* Interrupted projections */
      proj.param.type == INTERMOLL  ||
      proj.param.type == GOODE2 )
    Hlonres = proj.pixsz;
  if (proj.param.type == PLATECARREE)
    Hlonres = proj.pixsz;

  for (k=0; k<NGRIDS; k++) {

    if (! (grid->request & 1 << k)) continue;

    printf("  Plotting %s...\n", grid->type[k]);

    latmin = -grid->stop;
    latmax =  grid->stop;

    if (k == TILES) {
      isinusint(1.0, 0., 0., 0., isinusparam.nzones);
      grid->step = 10.;		/* After parallels and meridians are plotted grid->step can be modified */
      latmin = -90.;
      latmax =  90.;
      Hlonres = proj.pixsz;
      Hlatres = MAX(1, (int)(proj.pixsz * DEG2KM)) / DEG2KM;	/* The land tile grid must be computed at */
      								/* a resolution that is a multiple of 1km */
      if (proj.param.type == PLATECARREE ||	/* Strong distorsion of the L3 grid -> oversampling needed */
          proj.param.type == ROBINSON ||
          proj.param.type == GOODE ||
          proj.param.type == GOODE2 ||
          proj.param.type == ECKERTIV )
        Hlatres = MAX(1, (int)(proj.pixsz / 5 * DEG2KM)) / DEG2KM;
    }

    if (k == PARALLELS  ||  k == TILES) {		/* In case of TILES, variables *lon* are actually */
      ilatmin = latmin/grid->step;			/* horizontal coordinates, not longitudes */
      ilatmax = latmax/grid->step;
      for (ilat=ilatmin; ilat<=ilatmax; ilat++) {
        lat2 = lat1 = ilat * grid->step;
        for (ilon=-180/Hlonres; ilon<(int)(180/Hlonres); ilon++) {	/* ilon must not reach 180/Hlonres */
          lon1 =  ilon      * Hlonres + 1e-8;				/* because of lon2 -> here use (int)() */
          lon2 = (ilon + 1) * Hlonres - 1e-8;
          draw_lonlatline(&proj, lon1, lat1, lon2, lat2, grid, k);
        }
      }
    }

    if (k == MERIDIANS  ||  k == TILES) {		/* In case of TILES, variables *lon* are actually */
      ilonmin = -180/grid->step;				/* horizontal coordinates, not longitudes */
      ilonmax =  180/grid->step;
      for (ilon=ilonmin; ilon<=ilonmax; ilon++) {	/* Here ilon must reach ilonmax */
        lon1 = lon2 = ilon * grid->step;			/* Longitude at equator */
        ilatmin = floor(latmin/Hlatres);		/* Use floor() because of negative values */
        ilatmax = floor(latmax/Hlatres);		/* Use floor() too, not ceil(), because of lat2 */
        for (ilat=ilatmin; ilat<=ilatmax; ilat++) {
          lat1 = MAX(latmin,  ilat      * Hlatres);
          lat2 = MIN(latmax, (ilat + 1) * Hlatres);
          draw_lonlatline(&proj, lon1, lat1, lon2, lat2, grid, k);
        }
      }
    }

    if (k == MERIDIANS) {
      if (ptr = proj.info.intlon) {
        for (j=0; j<proj.info.Nintlon ; j++) {
          if ( sscanf(ptr, "%g,%g,%g", &intlon, &latmin, &latmax) != 3 ) break;
          for (ilon=-1; ilon<2; ilon+=2) {
            lon1 = lon2 = intlon + ilon * 1e-6;
            ilatmin = floor(latmin/Hlatres);			/* Use floor() because of negative values */
            ilatmax = floor(latmax/Hlatres);			/* Use floor() too, not ceil(), because of lat2 */
            for (ilat=ilatmin; ilat<=ilatmax; ilat++) {
              lat1 = MAX(latmin,  ilat      * Hlatres);
              lat2 = MIN(latmax, (ilat + 1) * Hlatres);
              draw_lonlatline(&proj, lon1, lat1, lon2, lat2, grid, k);
            }
          }
          if ( ptr = strchr(ptr, 32) ) ptr++;	/* Repoints ptr just after next space character */
          else break;				/* If no more field, then stop parsing */
        }
      }
    }

/*
    if (k == TILES) {					/ * Plot the dateline * /
*/
/* Following complex tests are not necessary anymore for most cases since the valid domain is outlined in outline_domain() */
/* But still necessary for non-elliptic projections */
    if ( k == TILES  ||					/* Plot the dateline */
         k == MERIDIANS && grid->stop < 90  &&		/* Up to the pole in case meridians are stopped before */
         proj.param.lat_pole == 90.  &&			/* Unless the globe is tilted */
         ! proj.info.azimuthal ) {			/* Unless the projection is azimuthal */

      for (ilon=-1; ilon<2; ilon+=2) {			/* In case of TILES, variables *lon* are actually */
        if (k == TILES) {				/* horizontal coordinates, not longitudes */
            lon1 = lon2 = ilon * 180;
            ilatmin = floor(latmin/Hlatres);		/* Use floor() because of negative values */
            ilatmax = floor(latmax/Hlatres);		/* Use floor() too, not ceil(), because of lat2 */
        } else {
            lon1 = lon2 = ilon * 180 + proj.param.lon_center;
/*
            if (lon1 < -180) lon1 = lon2 = ilon * 180 + proj.param.lon_center + 360;
            if (lon1 >  180) lon1 = lon2 = ilon * 180 + proj.param.lon_center - 360;
*/
            ilatmin = floor(-90 / Hlatres);
            ilatmax = floor( 90 / Hlatres);
        }
        for (ilat=ilatmin; ilat<=ilatmax; ilat++) {
          lat1 = MAX(-90,  ilat      * Hlatres);
          lat2 = MIN( 90, (ilat + 1) * Hlatres);
          draw_lonlatline(&proj,					/* In that particular case, the dateline */
                          lon1, lat1, lon2, lat2, grid, MERIDIANS);	/* to plot is a meridian, even if k == TILES */
        }
      }
    }

    if (k == NODES) {
      for (ilat=latmin/grid->step; ilat<=latmax/grid->step; ilat++) {
        lat1 = ilat * grid->step;
        for (ilon=-180/grid->step; ilon<=180/grid->step; ilon++) {
          lon1 = ilon * grid->step;
          if ( get_xyproj(lon1, lat1, &xproj, &yproj, &proj.param) == -1 ) continue;
          update_grid(grid, k, MAX(- proj.col0 * proj.pixsz, xproj - 3 * proj.pixsz), yproj,
                               MIN(  proj.col0 * proj.pixsz, xproj + 3 * proj.pixsz), yproj);
          update_grid(grid, k, xproj, MAX(- proj.row0 * proj.pixsz, yproj - 3 * proj.pixsz),
                               xproj, MIN(  proj.row0 * proj.pixsz, yproj + 3 * proj.pixsz));
        }
      }
    }

  }

}




void make_fill_SDS(int32 file_id, char *filename, PROJECTION proj, GRID grid, VECTOR vector)
{
int k, j;
double xproj1, yproj1, xproj2, yproj2;
uint8 palette[3 * 256];
SDS fillsds;

     if (grid.request == 0) return;

     set_proj_dimensions(&proj, &fillsds.Nl, &fillsds.Np);

     strcpy(fillsds.name, "Grid");
     fillsds.file_id = file_id;
     fillsds.num_type = DFNT_UINT8;
     fillsds.data = (char *) malloc(fillsds.Nl * fillsds.Np * sizeof(char));
     setfillvalue(fillsds.data, 0, fillsds.Nl * fillsds.Np, fillsds.num_type, 0);

     for (k=0; k<NGRIDS; k++) {
       if (grid.Npoints[k] > 0) {
         xproj2 = grid.points[k][0];
         yproj2 = grid.points[k][1];
         for (j=1; j<grid.Npoints[k]; j++) {
           xproj1 = xproj2;
           yproj1 = yproj2;
           xproj2 = grid.points[k][2 * j];
           yproj2 = grid.points[k][2 * j + 1];
           if  ( (xproj1 != VECTORUNDEF  ||  yproj1 != VECTORUNDEF)  &&
                 (xproj2 != VECTORUNDEF  ||  yproj2 != VECTORUNDEF) )
             draw_xyline(fillsds.data, fillsds.num_type, &proj, xproj1, yproj1, xproj2, yproj2, grid.offset[k]);
         }
       }
     }

     outline_domain(fillsds.data, fillsds.num_type, proj);
/*
     add_vectors(fillsds.data, fillsds.num_type, proj, vector);
*/

     if ( create_2D_SDS(&fillsds, APPEND) != -1 ) {
       write_SDS(fillsds, 0, fillsds.Nl, options.gzip);
/*
       write_SDS_attributes(fillsds, proj, grid, vector);
*/
     }

     for (k=0; k<256; k++) palette[3 * k] = palette[3 * k + 1] = palette[3 * k + 2] = k;
     setcolor(palette, fill.uint8 + grid.offset[MERIDIANS], 000, 000, 000);
     setcolor(palette, fill.uint8 + grid.offset[PARALLELS], 000, 000, 000);
     setcolor(palette, fill.uint8 + grid.offset[TILES],     255, 000, 000);
     setcolor(palette, fill.uint8 + grid.offset[NODES],     000, 000, 255);
     setcolor(palette, fill.uint8 + OUTLINEOFFSET,          000, 000, 000);
     DFR8setpalette(palette);
     DFR8addimage(filename, fillsds.data, fillsds.Np, fillsds.Nl, COMP_RLE);

     free(fillsds.data);

}




void setcolor(uint8 *palette, int index, unsigned char red, unsigned char green, unsigned char blue)
{
     if (index >= 0  &&  index <= 255) {
       palette[3 * index] = red;
       palette[3 * index + 1] = green;
       palette[3 * index + 2] = blue;
     }
}




void add_grids(void *data, int32 num_type, PROJECTION proj, GRID grid)
{
int k, j;
double xproj1, yproj1, xproj2, yproj2;

     if (grid.separate) return;

     for (k=0; k<NGRIDS; k++)
       if (grid.Npoints[k] > 0) {
         printf("  Adding grids...\n");
         break;
       }
     for (k=0; k<NGRIDS; k++) {
       if (grid.Npoints[k] > 0) {
         xproj2 = grid.points[k][0];
         yproj2 = grid.points[k][1];
         for (j=1; j<grid.Npoints[k]; j++) {
           xproj1 = xproj2;
           yproj1 = yproj2;
           xproj2 = grid.points[k][2 * j];
           yproj2 = grid.points[k][2 * j + 1];
           if  ( (xproj1 != VECTORUNDEF  ||  yproj1 != VECTORUNDEF)  &&
                 (xproj2 != VECTORUNDEF  ||  yproj2 != VECTORUNDEF) )
             draw_xyline(data, num_type, &proj, xproj1, yproj1, xproj2, yproj2, grid.offset[k]);
         }
       }
     }

}




void draw_lonlatline(PROJECTION *proj, double lon1, double lat1, double lon2, double lat2, GRID *grid, int gridno)
{
double xproj1, xproj2, yproj1, yproj2;

  if (gridno == TILES) {	/* In case of TILES, input "lon" and "lat" are actually horizontal and vertical */
    xproj1 = lon1;	/* coordinates in INTEGERSIN projection, not longitude and latitude */
    xproj2 = lon2;
    yproj1 = lat1;
    yproj2 = lat2;
    if ( get_lonlat(xproj1, yproj1, &lon1, &lat1, &isinusparam) == -1  ||	/* Get the actual (lon,lat) */
         get_lonlat(xproj2, yproj2, &lon2, &lat2, &isinusparam) == -1 ) return;	/* and the valid domain */
  }
  if ( ( gridno != TILES  ||			/* In case of TILES in output projection INTEGERSIN, */
         cmp_projections(proj->param, isinusparam) == -1 )  &&	/* do nothing, xproj and yproj were the input */
       ( get_xyproj(lon1, lat1, &xproj1, &yproj1, &proj->param) == -1 ||
         get_xyproj(lon2, lat2, &xproj2, &yproj2, &proj->param) == -1 ) )  return;
  if ( check_continuity(lon1, lat1, lon2, lat2, xproj1, xproj2, &proj->param, proj->pixsz) == -1 ) return;
  update_grid(grid, gridno, xproj1, yproj1, xproj2, yproj2);

}




void draw_xyline(void *data, int32 num_type, PROJECTION *proj, double xproj1, double yproj1, double xproj2, double yproj2, int offset)
{
int Np, irow, jcol, irow1, irow2, jcol1, jcol2;
double xproj, yproj, a, b;

    if (proj->sds.data)
      Np = proj->sds.Np;
    else if (proj->meta.data)
      Np = proj->meta.Np;
    else if (proj->domain == USERBOX)
      Np = proj->bbox.p2 - proj->bbox.p1 + 1;
    else
      Np = proj->Np_glb;
      
    if (yproj1 == yproj2) {

        jcol1 = MIN(xproj1, xproj2) / proj->pixsz + proj->col0 + 0.5;
        jcol2 = MAX(xproj1, xproj2) / proj->pixsz + proj->col0 + 0.5;
        if (jcol1 < 0) jcol1 = 0;
        if (jcol2 > Np - 1) jcol2 = Np - 1;
        if (jcol2 >= jcol1) {
          irow1 = - yproj1 / proj->pixsz + proj->row0 + 0.5;
          if (irow1 >= proj->row1 && irow1 <= proj->row2)
            setfillvalue(data, (irow1 - proj->row1) * Np + jcol1, jcol2 - jcol1 + 1, num_type, offset);
        }

    } else {

        a = (xproj2 - xproj1) / (yproj2 - yproj1);
        b = (yproj2 * xproj1 - yproj1 * xproj2) / (yproj2 - yproj1);

        irow1 = - MAX(yproj1, yproj2) / proj->pixsz + proj->row0 + 0.5;
        irow2 = - MIN(yproj1, yproj2) / proj->pixsz + proj->row0 + 0.5;
        if (irow1 < proj->row1) irow1 = proj->row1;
        if (irow2 > proj->row2) irow2 = proj->row2;
        jcol1 = MIN(xproj1, xproj2) / proj->pixsz + proj->col0 + 0.5;
        jcol2 = MAX(xproj1, xproj2) / proj->pixsz + proj->col0 + 0.5;
        if (jcol1 < 0) jcol1 = 0;
        if (jcol2 > Np - 1) jcol2 = Np - 1;
        for (irow=irow1; irow<=irow2; irow++) {
          yproj = (proj->row0 - irow) * proj->pixsz;
          xproj = a * yproj + b;
          jcol = xproj / proj->pixsz + proj->col0 + 0.5;
          if (jcol >= jcol1 && jcol <= jcol2)
            setfillvalue(data, (irow - proj->row1) * Np + jcol, 1, num_type, offset);
        }
        if (jcol2 > jcol1) {
          for (jcol=jcol1; jcol<=jcol2; jcol++) {
            xproj = (jcol - proj->col0) * proj->pixsz;
            yproj = (xproj - b) / a;
            irow = - yproj / proj->pixsz + proj->row0 + 0.5;
            if (irow >= irow1 && irow <= irow2)
              setfillvalue(data, (irow - proj->row1) * Np + jcol, 1, num_type, offset);
          }
        }

    }

}




void draw_cross(void *data, int irow, int jcol, int row1, int row2, int Np, int32 num_type, int halfsize, int offset)
{
int j, irow1, irow2, jcol1, jcol2;

    irow1 = MAX(row1,   irow - halfsize);
    irow2 = MIN(row2,   irow + halfsize);
    jcol1 = MAX(0,      jcol - halfsize);
    jcol2 = MIN(Np - 1, jcol + halfsize);
    if ( irow >= row1  &&  irow <= row2  &&
         jcol2 >= jcol1 )
        setfillvalue(data, (irow - row1) * Np + jcol1, jcol2 - jcol1 + 1, num_type, offset);
    if ( jcol >= 0  &&  jcol <= Np - 1 )
      for (j=irow1; j<=irow2; j++)
        setfillvalue(data, (j - row1) * Np + jcol, 1, num_type, offset);

}




void update_grid(GRID *grid, int gridno, double xproj1, double yproj1, double xproj2, double yproj2)
{
float epsilon=1e-4;	/* 1e-4 deg. is about 10 m */

  if (2 * (grid->Npoints[gridno] + 3) * sizeof(float64) > grid->bufsize[gridno]) {
    if ( check_memory( (grid->bufsize[gridno] + BUFINC) * sizeof(float64)) == -1 ) return;
    grid->bufsize[gridno] += BUFINC;
    grid->points[gridno] = (float *) realloc(grid->points[gridno], grid->bufsize[gridno] * sizeof(float));
  }
  if ( grid->Npoints[gridno] == 0  ||
       fabs(xproj1 - grid->points[gridno][2 * grid->Npoints[gridno] - 2]) > epsilon  ||		/* No continuity */
       fabs(yproj1 - grid->points[gridno][2 * grid->Npoints[gridno] - 1]) > epsilon ) {
    if (grid->Npoints[gridno] > 0) {
      grid->points[gridno][2 * grid->Npoints[gridno]]     = VECTORUNDEF;		/* Add a break */
      grid->points[gridno][2 * grid->Npoints[gridno] + 1] = VECTORUNDEF;
      grid->Npoints[gridno]++;
    }
    grid->points[gridno][2 * grid->Npoints[gridno]]     = xproj1;		/* Store 1st point */
    grid->points[gridno][2 * grid->Npoints[gridno] + 1] = yproj1;
    grid->Npoints[gridno]++;
  }
  grid->points[gridno][2 * grid->Npoints[gridno]]     = xproj2;			/* Store 2nd point */
  grid->points[gridno][2 * grid->Npoints[gridno] + 1] = yproj2;
  grid->Npoints[gridno]++;

}




char *get_next_string(char *arg, char *string)
{
char *ptr;

       if (*arg == '\0') return NULL;

       if ( (ptr = strchr(arg, 44))  ||		/* Point on next separator (comma) */
            (ptr = strchr(arg, 0)) ) {		/* If none, point on end-of-string */
         strncpy(string, arg, ptr - arg);	/* Copy beginning of arg (up to just before separator) into string */
         string[ptr - arg] = '\0';	/* Terminate string with a null character */
       }

       return ptr;

}




void init_projection(PROJECTION *proj)
{
double xproj, yproj;
int k, irow, jcol;
double pixsz;

       if (proj->param.type == 0) {	/* Non valid projection */
         fprintf(stderr, "No valid output projection was specified.\nSupported projections:\n");
         for (k=1; k<NUMBEROFPROJ; k++) fprintf(stderr, "%2d. %s\n", k, projinfo[k].name);
         fprintf(stderr, "\n%s", HELPMSG);
         exit(-1);
       }

       if (proj->pixsz == 0) {		/* At this stage the resolution must have been specified */
         proj->pixsz = DEFPIXSZ / DEG2KM;
         printf("%g km (%g deg.) will be used for the output projection.\n\n",
                DEG2KM * proj->pixsz, proj->pixsz);
       }

       proj->Nl_glb = ceil(180. / proj->pixsz);	/* Nl_glb and Np_glb are the dimensions */
       proj->Np_glb = ceil(360. / proj->pixsz);	/* for the full globe at the specified resolution  */

       /* The dimensions in some projections are adjusted below, to fit the area actually used. */
       switch (proj->param.type) {

         case GEOGRAPHIC: break;
         case SINUSOIDAL:  break;
         case INTEGERSIN:  pixsz = 180. / isinusparam.nzones * (int)(proj->pixsz * isinusparam.nzones / 180. + 0.5);
                           if (pixsz == 0)
                             pixsz = 180. / isinusparam.nzones / (int)(180. / isinusparam.nzones / proj->pixsz + 0.5);
                           if (proj->pixsz != pixsz) {
                             proj->pixsz = pixsz;
                             printf("Pixel size must be a divisor of %g deg. for %s projection\n", 180. / isinusparam.nzones, projinfo[proj->param.type].name);
                             printf("Pixel size has been adjusted to %g km (%g deg.)\n", DEG2KM * proj->pixsz, proj->pixsz);
                           }
                           proj->Nl_glb = 180. / proj->pixsz + 0.5;
                           proj->Np_glb = 2 * cos(M_PI_2 / isinusparam.nzones) * 180. / proj->pixsz  + 0.5;
                           isinusint(1.0, 0., 0., 0., isinusparam.nzones);
                           break;
         case INTERSINUS1: intersinint1(1.0, 0., 0.);
                           break;
         case INTERSINUS2: intersinint2(1.0, 0., 0.);
                           break;
         case INTERSINUS3:
           if (proj->param.nzones == 0) proj->param.nzones = 9;
           proj->Np_glb = ceil((360. + (proj->param.nzones - 1) * ZONEGAP) / proj->pixsz);
           break;
         case PLATECARREE:
           if (proj->param.stdlat1 != PROJUNDEF) {
             proj->param.xcorr = cos(proj->param.stdlat1);
             printf("Stretch factor: %lg\n", proj->param.xcorr);
           }
           break;

         case ROBINSON:
           robforint(RAD2DEG, 0., 0., 0.);
           robinvint(RAD2DEG, 0., 0., 0.);
           proj->Np_glb = ceil(0.9858 * 360. / proj->pixsz);
           break;

         case MOLLWEIDE:
           molwforint(RAD2DEG, 0., 0., 0.);
           molwinvint(RAD2DEG, 0., 0., 0.);
           proj->Nl_glb = ceil(2 * sqrt(2.) / PI * 180. / proj->pixsz);
           proj->Np_glb = ceil(2 * sqrt(2.) / PI * 360. / proj->pixsz);
           break;

         case INTERMOLL:
           imolwforint(RAD2DEG, 0., 0.);
           imolwinvint(RAD2DEG, 0., 0.);
           proj->Nl_glb = ceil(2 * sqrt(2.) / PI * 180. / proj->pixsz);
           proj->Np_glb = ceil(2 * sqrt(2.) / PI * 360. / proj->pixsz);
           break;

         case GOODE:
           goodforint(RAD2DEG, 0., 0.);
           goodinvint(RAD2DEG, 0., 0.);
           proj->Nl_glb = ceil(2 * (sqrt(2.) - 0.0528035274542) / PI * 180. / proj->pixsz);
           break;

         case GOODE2:
           goodforint2(RAD2DEG, 0., 0.);
           goodinvint2(RAD2DEG, 0., 0.);
           proj->Nl_glb = ceil(2 * (sqrt(2.) - 0.0528035274542) / PI * 180. / proj->pixsz);
           break;

         case PERSPECTIVE:
           if (proj->param.lon_center == PROJUNDEF) proj->param.lon_center = -55;	/* Default center is longitude 55 W */
           if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 20;	/* Default center is latitude  20 N */
           gvnspforint(RAD2DEG, proj->param.height / DEG2KM, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           gvnspinvint(RAD2DEG, proj->param.height / DEG2KM, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           proj->Nl_glb = ceil(2 * RAD2DEG * sqrt(1. / (1. + 2. * DEG2KM * RAD2DEG / proj->param.height)) / proj->pixsz);
           proj->Np_glb = proj->Nl_glb;
           break;

         case ORTHOGRAPHIC:
           if (proj->param.lon_center == PROJUNDEF) proj->param.lon_center = -55;	/* Default center is longitude 55 W */
           if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 20;	/* Default center is latitude  20 N */
           orthforint(RAD2DEG, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           orthinvint(RAD2DEG, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           proj->Nl_glb = ceil(2 * RAD2DEG / proj->pixsz);
           proj->Np_glb = proj->Nl_glb;
           break;

         case GNOMONIC:
           if (proj->param.lon_center == PROJUNDEF) proj->param.lon_center = -55;	/* Default center is longitude 55 W */
           if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 20;	/* Default center is latitude  20 N */
           if (proj->param.range > MAXRANGE) {
             printf("Range from center (%g deg.) is too big for gnomonic projection. Request ignored.\n", proj->param.range);
             proj->param.range = 60;
           }
           if (proj->param.range == PROJUNDEF) proj->param.range = 60;		/* Default range is 60 deg. from center */
           gnomforint(RAD2DEG, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           gnominvint(RAD2DEG, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           proj->param.radius = RAD2DEG * tan(proj->param.range * DEG2RAD);
           proj->Nl_glb = ceil(2 * proj->param.radius / proj->pixsz);
           proj->Np_glb = proj->Nl_glb;
           break;

         case STEREOGRAPHIC:
           if (proj->param.lon_center == PROJUNDEF) proj->param.lon_center = -55;	/* Default center is longitude 55 W */
           if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 20;	/* Default center is latitude  20 N */
           sterforint(RAD2DEG, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           sterinvint(RAD2DEG, 0., proj->param.lat_center * DEG2RAD, 0., 0.);
           proj->Nl_glb = ceil(4 * RAD2DEG / proj->pixsz);
           proj->Np_glb = proj->Nl_glb;
           break;

         case LAMBERT:
           if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 90;	/* Default center is North Pole */
           if ( fabs(proj->param.lat_center) != 90  &&
                proj->param.range != PROJUNDEF  &&
                proj->param.range != 90  ) {
             printf("Range from center is meaningless a non-polar Lambert Azimuthal projection. Request ignored.\n"),
             proj->param.range = 90;
           }
           if (proj->param.range == PROJUNDEF) proj->param.range = 60;		/* Default range is 60 deg. from center */
           lamazforint(RAD2DEG, 0.,  proj->param.lat_center * DEG2RAD, 0., 0.);
           lamazinvint(RAD2DEG, 0.,  proj->param.lat_center * DEG2RAD, 0., 0.);
           proj->param.radius = RAD2DEG * sin(proj->param.range * DEG2RAD) * sqrt(2. / (1 + cos(proj->param.range * DEG2RAD)));
           proj->Nl_glb = ceil(2 * proj->param.radius / proj->pixsz);
           proj->Np_glb = proj->Nl_glb;
           break;

         case AZIMEQUI:
           if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 90;	/* Default center is North Pole */
           if ( fabs(proj->param.lat_center) != 90  &&
                proj->param.range != PROJUNDEF  &&
                proj->param.range != 90  ) {
             printf("Range from center is meaningless a non-polar Azimuthal Equidistant projection. Request ignored.\n"),
             proj->param.range = 90;
           }
           if (proj->param.range == PROJUNDEF) proj->param.range = 60;		/* Default range is 60 deg. from center */
           azimforint(RAD2DEG, 0.,  proj->param.lat_center * DEG2RAD, 0., 0.);
           aziminvint(RAD2DEG, 0.,  proj->param.lat_center * DEG2RAD, 0., 0.);
           proj->param.radius = proj->param.range;
           proj->Nl_glb = ceil(2 * proj->param.radius / proj->pixsz);
           proj->Np_glb = proj->Nl_glb;
           break;

         case HAMMER:
           hamforint(RAD2DEG, 0., 0., 0.);
           haminvint(RAD2DEG, 0., 0., 0.);
           proj->Nl_glb = ceil(2 * RAD2DEG * sqrt(2.) / proj->pixsz);
           proj->Np_glb = ceil(4 * RAD2DEG * sqrt(2.) / proj->pixsz);
           break;

         case ALBERS:
           if (proj->param.stdlat1 == PROJUNDEF) proj->param.stdlat1 = 45.5 * DEG2RAD;
           if (proj->param.stdlat2 == PROJUNDEF) proj->param.stdlat2 = 29.5 * DEG2RAD;
           alberforint(RAD2DEG, RAD2DEG, proj->param.stdlat1, proj->param.stdlat2, 0., 0., 0., 0.);
           alberinvint(RAD2DEG, RAD2DEG, proj->param.stdlat1, proj->param.stdlat2, 0., 0., 0., 0.);
           break;

         case AITOFF:
           aitint(RAD2DEG, 0., 0., 0.);
           break;

         case ECKERTIV:
           eckivforint(RAD2DEG, 0., 0., 0.);
           eckivinvint(RAD2DEG, 0., 0., 0.);
           proj->Nl_glb = ceil(4 * sqrt(PI / (4 + PI)) / PI * 180. / proj->pixsz);
           proj->Np_glb = ceil(4 * sqrt(PI / (4 + PI)) / PI * 360. / proj->pixsz);
           break;

       }

       if (proj->lon_target != PROJUNDEF) {
         if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = proj->lat_target;
         if (proj->param.lon_center == PROJUNDEF) proj->param.lon_center = proj->lon_target;
       }
       if (proj->param.lat_center == PROJUNDEF) proj->param.lat_center = 0;	/* No center latitude has been specified */
       if (proj->param.lon_center == PROJUNDEF) proj->param.lon_center = 0;	/* No center longitude has been specified */
       if (proj->param.lon_pole == PROJUNDEF) proj->param.lon_pole = 0.;
       if (proj->param.lat_pole == PROJUNDEF) proj->param.lat_pole = 90.;
       proj->param.coslatpole = cos(proj->param.lat_pole * DEG2RAD);
       proj->param.sinlatpole = sin(proj->param.lat_pole * DEG2RAD);

       if (proj->param.nzones == 0) proj->param.nzones = 1;

       if (proj->domain == GLOBAL)
         printf("Dimensions of output projection: %dx%d\n\n", proj->Np_glb, proj->Nl_glb);
       else
         printf("Dimensions of global projection at the output resolution: %dx%d\n\n", proj->Np_glb, proj->Nl_glb);

       proj->info = projinfo[proj->param.type];
       if (proj->param.lat_pole != 90) {
         proj->info.latsym = FALSE;
         if (proj->param.lon_pole != 0) proj->info.lonsym = FALSE;
       }
       if (proj->param.lon_center != 0) proj->info.lonsym = FALSE;

       if (proj->lon_target != PROJUNDEF) {
         if ( get_xyproj(proj->lon_target, proj->lat_target, &xproj, &yproj, &proj->param) == -1 ) {
             proj->lat_target = proj->lon_target = PROJUNDEF;
             proj->param.lat_center = proj->param.lon_center = 0;
         } else {
             irow = - yproj / proj->pixsz + 0.5;
             jcol =   xproj / proj->pixsz + 0.5;
             proj->bbox.l2 = (proj->Nl_glb / 2. + 0.5 - 1) - yproj / proj->pixsz + proj->height / 2 + 0.5;
             proj->bbox.l1 = proj->bbox.l2 - proj->height + 1;
             if (proj->bbox.l1 < 0) proj->bbox.l1 = 0;
             proj->bbox.p2 = (proj->Np_glb / 2. + 0.5 - 1) + xproj / proj->pixsz + proj->width / 2 + 0.5;
             proj->bbox.p1 = proj->bbox.p2 - proj->width + 1;
             if (proj->bbox.p1 < 0) proj->bbox.p1 = 0;
             proj->domain = USERBOX;
         }
       }

       if ( proj->bbox.l2 > proj->Nl_glb - 1  ||
            proj->bbox.p2 > proj->Np_glb - 1 ) {
         printf("The requested output subwindow falls outside the global projection\n");
         if (proj->bbox.p2 > proj->Np_glb - 1) {
           proj->bbox.p2 = proj->Np_glb - 1; 
           printf("Width adjusted to %d\n", proj->bbox.p2 - proj->bbox.p1 + 1);
         }
         if (proj->bbox.l2 > proj->Nl_glb - 1) {
           proj->bbox.l2 = proj->Nl_glb - 1; 
           printf("Height adjusted to %d\n", proj->bbox.l2 - proj->bbox.l1 + 1);
         }
         printf("\n");
       }

       proj->allocbox.l1 = 0;		/* Must be initialized to 0, because it is used in */
       proj->allocbox.p1 = 0;		/* get_estimbox() before an array is actually allocated */
       if (proj->domain != USERBOX) {
         proj->bbox.l1 = proj->Nl_glb;
         proj->bbox.p1 = proj->Np_glb;
         proj->bbox.l2 = -1;
         proj->bbox.p2 = -1;
       }
       proj->allocbox.l2 = -1;
       proj->allocbox.p2 = -1;
       proj->sds.Nl = 0;
       proj->sds.Np = 0;
       proj->sds.data = NULL;
       proj->sds.num_type = DFNT_NONE;
       proj->sds.min =  DBL_MAX;
       proj->sds.max = -DBL_MAX;
       proj->sds.lut = NULL;
       proj->meta.Nl = 0;
       proj->meta.Np = 0;
       proj->meta.data = NULL;
       proj->meta.num_type = DFNT_NONE;
       proj->meta.min =  DBL_MAX;
       proj->meta.max = -DBL_MAX;
       proj->meta.lut = strdup("\n");
       proj->meta.Nlut = 0;
       proj->lgid.Nl = 0;
       proj->lgid.Np = 0;
       proj->lgid.data = NULL;
       proj->lgid.num_type = DFNT_NONE;
       proj->lgid.min =  DBL_MAX;
       proj->lgid.max = -DBL_MAX;
       proj->lgid.lut = strdup("\n");
       proj->lgid.Nlut = 0;
       proj->senz = NULL;

}




int  check_geolocation_consistency(PRODUCT granule)
{

    if (granule.lat.rank != 2) {
      printf("  Rank of latitude is not 2. File ignored.\n\n");
      SDendaccess(granule.lat.id);
      SDendaccess(granule.lon.id);
      SDendaccess(granule.sds.id);
      if (L1B) SDendaccess(granule.sz.id); /*added 07/13 */
      SDend(granule.geofile.id);
      SDend(granule.file.id);
      return -1;
    }
    if (granule.lon.rank != 2) {
      printf("  Rank of longitude is not 2. File ignored.\n\n");
      SDendaccess(granule.lat.id);
      SDendaccess(granule.lon.id);
      SDendaccess(granule.sds.id);
      if (L1B) SDendaccess(granule.sz.id); /*added 07/13 */
      SDend(granule.geofile.id);
      SDend(granule.file.id);
      return -1;
    }
    granule.Nl_geo = granule.lat.dim_sizes[0];
    granule.Np_geo = granule.lat.dim_sizes[1];
    if ( (granule.sds.Nl >= granule.Nl_geo  &&  granule.sds.Nl % granule.Nl_geo != 0) ||
         ( granule.sds.Np >= granule.Np_geo  &&
           granule.sds.Np % granule.Np_geo != 0  &&
/* Handle coarse geolocation with 271 columns */
           granule.sds.Np % (granule.Np_geo - 1) >= granule.sds.Np / (granule.Np_geo - 1) ) ) {
      printf("  Dimensions of \"%s\" in file %s (%dx%d) and\n"\
             "                \"%s\" in file %s (%dx%d)\n"\
             "                are inconsistent. File ignored.\n\n",
             granule.lat.name, granule.geofile.name, granule.Np_geo, granule.Nl_geo,
             granule.sds.name, granule.file.name, granule.sds.Np, granule.sds.Nl);
      SDendaccess(granule.lat.id);
      SDendaccess(granule.lon.id);
      SDendaccess(granule.sds.id);
      if (L1B) SDendaccess(granule.sz.id); /*added 07/13 */
      SDend(granule.geofile.id);
      SDend(granule.file.id);
      return -1;
    }
    if ( granule.lon.dim_sizes[0] != granule.Nl_geo  ||
         granule.lon.dim_sizes[1] != granule.Np_geo ) {
      printf("  Dimensions of \"%s\" (%dx%d)\n"\
             "            and \"%s\" (%dx%d) in file %s are not identical. File ignored.\n\n",
             granule.lon.name, granule.lon.dim_sizes[1], granule.lon.dim_sizes[0],
             granule.lat.name, granule.Np_geo, granule.Nl_geo, granule.geofile.name);
      SDendaccess(granule.lat.id);
      SDendaccess(granule.lon.id);
      SDendaccess(granule.sds.id);
      if (L1B) SDendaccess(granule.sz.id); /*added 07/13 */
      SDend(granule.geofile.id);
      SDend(granule.file.id);
      return -1;
    }

    return 0;

}




int  get_geolocation(PRODUCT *input)
{

   if (input->projparam.type == SWATH) {			/* Granule product */

       /* Search Latitude and Longitude in the same file as the product */
       /* Assumption: must be same dimensions as the product */
     if ( ! MOD03  &&		/* Use of MOD03 file forced by user */
          ( (input->lat.index = SDnametoindex(input->file.id, strcpy(input->lat.name, "Latitude"))) != -1  ||
            (input->lat.index = SDnametoindex(input->file.id, strcpy(input->lat.name, "PixelLatitude"))) != -1  ||
            (input->lat.index = SDnametoindex(input->file.id, strcpy(input->lat.name, "Coarse Resolution Latitude"))) != -1 )  &&
          (input->lat.id = SDselect(input->file.id, input->lat.index)) != -1  &&
          SDgetinfo(input->lat.id, input->lat.name, &input->lat.rank, input->lat.dim_sizes, &input->lat.num_type, &input->lat.n_attr) != -1 &&
          input->lat.rank == 2 &&
          (input->Nl_geo = input->lat.dim_sizes[0]) == input->sds.Nl &&
          (input->Np_geo = input->lat.dim_sizes[1]) == input->sds.Np &&
          ( (input->lon.index = SDnametoindex(input->file.id, strcpy(input->lon.name, "Longitude"))) != -1  ||
            (input->lon.index = SDnametoindex(input->file.id, strcpy(input->lon.name, "PixelLongitude"))) != -1  ||
            (input->lon.index = SDnametoindex(input->file.id, strcpy(input->lon.name, "Coarse Resolution Longitude"))) != -1 )  &&
          (input->lon.id = SDselect(input->file.id, input->lon.index)) != -1 &&
          SDgetinfo(input->lon.id, input->lon.name, &input->lon.rank, input->lon.dim_sizes, &input->lon.num_type, &input->lon.n_attr) != -1 &&
          input->lon.rank == 2 &&
          input->lon.dim_sizes[0] == input->sds.Nl  &&
          input->lon.dim_sizes[1] == input->sds.Np ) {

          printf("  Found suitable Latitude and Longitude in file.\n");
          input->lat.file_id  = input->file.id;
          input->lon.file_id  = input->file.id;
          input->lat.filename = input->file.name;
          input->lon.filename = input->file.name;
          input->lat.edges[1] = input->Np_geo;		/* Same initializations as get_SDS_info() */
          input->lon.edges[1] = input->Np_geo;
          input->lat.start[1] = 0;
          input->lon.start[1] = 0;

     } else {		/* No suitable lat and lon in the HDF file - Search geolocation file */

         if (! MOD03)
           printf("  No suitable Latitude and Longitude found in file.\n");
         printf("  Searching geolocation file...\n");

         get_file_from_inputpointer(*input, "MYD03", input->geofile.name);
/* Quick fix to use MYD03 for Aqua granules */
         if ( strstr(input->file.name, "crefl2")  ||
              strstr(input->file.name, "ndvi2")  ||
              strstr(input->file.name, "fire2") )
           substitute(input->geofile.name, "MOD03", "MYD03");

         if ( search_file(input->geofile.name, MODISPATH) != -1 )
           printf("  Found geolocation file: %s\n", input->geofile.name);
         else {
           printf("  Unable to find %s file in path specified by $MODISPATH.\n", input->geofile.name);
           if ( mask_production_date(input->geofile.name) != -1 ) {
               printf("  Searching %s instead...\n", input->geofile.name);
               if ( search_file(input->geofile.name, MODISPATH) != -1  ||
                    ( mask_production_date(input->geofile.name) != -1  &&
                      search_file(input->geofile.name, MODISPATH) != -1 ) )
                 printf("  Found geolocation file: %s\n", input->geofile.name);
               else {
                 printf("  Unable to find geolocation file. Searching a coarse geolocation instead...\n");
                 if ( (input->lat.index = SDnametoindex(input->file.id, strcpy(input->lat.name, "Latitude"))) != -1  &&
                      (input->lat.id = SDselect(input->file.id, input->lat.index)) != -1 &&
                      SDgetinfo(input->lat.id, input->lat.name, &input->lat.rank, input->lat.dim_sizes, &input->lat.num_type, &input->lat.n_attr) != -1 &&
                      input->lat.rank == 2 &&
                      (input->Nl_geo = input->lat.dim_sizes[0]) < input->sds.Nl &&
                      (input->Np_geo = input->lat.dim_sizes[1]) < input->sds.Np &&
                      ( (input->sds.Nl + input->Nl_geo - 1) / input->Nl_geo == (input->sds.Np + input->Np_geo - 1) / input->Np_geo  ||
/* The coarse geolocation in L1B has only 270 columns instead of 271 until version 2.1.4 */
                        (input->sds.Nl + input->Nl_geo - 1) / input->Nl_geo == (input->sds.Np + input->Np_geo) / (input->Np_geo + 1) )  &&
                      (input->lon.index = SDnametoindex(input->file.id, strcpy(input->lon.name, "Longitude"))) != -1  &&
                      (input->lon.id = SDselect(input->file.id, input->lon.index)) != -1 &&
                      SDgetinfo(input->lon.id, input->lon.name, &input->lon.rank, input->lon.dim_sizes, &input->lon.num_type, &input->lon.n_attr) != -1 &&
                      input->lon.rank == 2 &&
                      input->lon.dim_sizes[0] == input->Nl_geo  &&
                      input->lon.dim_sizes[1] == input->Np_geo ) {

                     printf("  Found suitable coarse Latitude and Longitude in file.\n");
                     input->lat.file_id  = input->file.id;
                     input->lon.file_id  = input->file.id;
                     input->lat.filename = input->file.name;
                     input->lon.filename = input->file.name;
                     input->lat.edges[1] = input->Np_geo;		/* Same initializations as get_SDS_info() */
                     input->lon.edges[1] = input->Np_geo;
                     input->lat.start[1] = 0;
                     input->lon.start[1] = 0;
       input->xpixsz = input->ypixsz = 1354. / input->sds.Np / DEG2KM;
       printf("  Presumed resolution of input SDS: %g km (%g deg.)\n", input->xpixsz * DEG2KM, input->xpixsz);
                     return 0;

                 } else {		/* No suitable lat and lon in the HDF file - Search MOD02 file */
                     printf("  Unable to find a coarse geolocation either. Searching a MOD02 file instead...\n");
/* Quick fix to use MYD products for Aqua granules */
                     if ( strstr(input->geofile.name, "MYD03")  &&
                        ( substitute(input->geofile.name, "MYD03", "MYD02QKM") != -1  &&
                            printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                            search_file(input->geofile.name, MODISPATH) != -1  ||
                            substitute(input->geofile.name, "MYD02QKM", "MYD02HKM") != -1  &&
                            printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                            search_file(input->geofile.name, MODISPATH) != -1  ||            
                            substitute(input->geofile.name, "MYD02HKM", "MYD021KM") != -1  &&
                            printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                            search_file(input->geofile.name, MODISPATH) != -1  ||            
                            substitute(input->geofile.name, "MYD021KM", "MYD02CRS") != -1  &&
                            printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                            search_file(input->geofile.name, MODISPATH) != -1 ) )
                         printf("  Found L1B file: %s\n", input->geofile.name);
                     else if ( ! strstr(input->geofile.name, "MYD03")  &&
                        ( substitute(input->geofile.name, "MOD03", "MOD02QKM") != -1  &&
                          printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                          search_file(input->geofile.name, MODISPATH) != -1  ||
                          substitute(input->geofile.name, "MOD02QKM", "MOD02HKM") != -1  &&
                          printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                          search_file(input->geofile.name, MODISPATH) != -1  ||            
                          substitute(input->geofile.name, "MOD02HKM", "MOD021KM") != -1  &&
                          printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                          search_file(input->geofile.name, MODISPATH) != -1  ||            
                          substitute(input->geofile.name, "MOD021KM", "MOD02CRS") != -1  &&
                          printf("  Searching %s instead...\n", input->geofile.name) != -1  &&
                          search_file(input->geofile.name, MODISPATH) != -1 ) )
                       printf("  Found L1B file: %s\n", input->geofile.name);
                     else {
                       printf("  Unable to find a MOD02 file either. File %s ignored.\n\n", input->file.name);
                       if (! options.depth)
                         printf("  Option -depth may be considered to perform a recursive search from the specified path.\n");
                       SDendaccess(input->sds.id);
		       if (L1B) SDendaccess(input->sz.id); /*added 07/13 */
                       SDend(input->file.id);
                       return -1;
                     }
                 }
               }
           } else {
               SDendaccess(input->sds.id);
               SDend(input->file.id);
               return -1;
           }
         }

         if ( (input->geofile.id = SDstart(input->geofile.name, DFACC_RDONLY)) == -1 ) {
           printf("  Unable to open geolocation file %s. File %s ignored.\n\n",
                  input->geofile.name, input->file.name);
           SDendaccess(input->sds.id);
	   if (L1B) SDendaccess(input->sz.id); /*added 07/13 */
           SDend(input->file.id);
           return -1;
         }

         input->lat.file_id  = input->geofile.id;
         input->lon.file_id  = input->geofile.id;
         input->lat.filename = input->geofile.name;
         input->lon.filename = input->geofile.name;
         strcpy(input->lat.name, "Latitude");
         strcpy(input->lon.name, "Longitude");

         if ( get_SDS_info(&input->lat) == -1 ||
              get_SDS_info(&input->lon) == -1 ) {
           SDendaccess(input->sds.id);
	   if (L1B) SDendaccess(input->sz.id); /*added 07/13 */
           SDend(input->file.id);
           return -1;
         }

         if (check_geolocation_consistency(*input) == -1) return -1;
         input->Nl_geo = input->lat.dim_sizes[0];
         input->Np_geo = input->lat.dim_sizes[1];

       }

       input->xpixsz = input->ypixsz = 1354. / input->sds.Np / DEG2KM;
       printf("  Presumed resolution of input SDS: %g km (%g deg.)\n", input->xpixsz * DEG2KM, input->xpixsz);
       input->projparam.nzones = 180. / input->xpixsz;	/* Number of rows or zones in the global projection */

   } else {				/* Tile product */

       if ( get_tileID(input->allmetadata, input) == -1 ) {
         SDendaccess(input->sds.id);
	 if (L1B) SDendaccess(input->sz.id); /*added 07/13 */
         SDend(input->file.id);
         return -1;
       }
       input->Nl_geo = input->sds.Nl;
       input->Np_geo = input->sds.Np;

   }

   return 0;

}




int  open_file(HDFfile *file)
{

   printf("Opening file %s...\n", file->name);
   if ( (file->id = SDstart(file->name, DFACC_RDONLY)) == -1 ) {
          printf("Unable to open file %s. File ignored.\n\n", file->name);
          return -1;
   } else return 0;

}




int get_SDS_info(SDS *sds)
{
int k, j;
int32 attr_index, num_type, count;
char attr_name[MAXNAMELENGTH];

   printf("  Searching SDS \"%s\"...\n", sds->name);

   if (    (  (sds->index = SDnametoindex(sds->file_id, sds->name)) == -1 	/* Name not found */
              &&
              ( sscanf(sds->name, "%d", &sds->index) != 1 ||		/* Name is not an index either */
                sds->index-- < -99 )  )					/* Always false - Apply if name was an index */
           ||
           (sds->id = SDselect(sds->file_id, sds->index)) == -1     ) {	/* No such SDS index in the file */
     printf("  Can't find SDS \"%s\" in file %s.\n", sds->name, sds->filename);
     printf("  File ignored.\n\n");
     SDend(sds->file_id);
     sds->id = -1;
     return -1;
   }

   if (SDgetinfo(sds->id, sds->name, &sds->rank, sds->dim_sizes, &sds->num_type, &sds->n_attr) == -1) {
     printf("  Can't get info from SDS \"%s\" in file %s.\n", sds->name, sds->filename);
     printf("  File ignored.\n\n");
     SDendaccess(sds->id);
     SDend(sds->file_id);
     sds->id = -1;
     return -1;
   }
   strcpy(attr_name, "_FillValue");
   attr_index = SDfindattr(sds->id, attr_name);
   sds->fillvalue = NULL;
   if (SDattrinfo(sds->id, attr_index, attr_name, &num_type, &count) == -1)
     printf("  Can't find fill value of SDS \"%s\" in file %s.\n", sds->name, sds->filename);
   else {
     if (num_type != sds->num_type)
       printf("  Data type of fill value is different from SDS data. Fill value is ignored.\n");
     else {
       sds->fillvalue = (void *) malloc(1000 * DFKNTsize(sds->num_type));
       if (SDgetfillvalue(sds->id, sds->fillvalue) != 0) {
         printf("  Can't get fill value from SDS \"%s\" in file %s.\n", sds->name, sds->filename);
         free(sds->fillvalue);
         sds->fillvalue = NULL;
       }
     }
   }
   switch (sds->rank) {					/* Set parameters according to rank */
     case 2:  sds->Nl = sds->dim_sizes[0];
              sds->Np = sds->dim_sizes[1];
              sds->edges[1] = sds->dim_sizes[1];	/* When rank is 2, start[0] and edges[0] will be set */
              sds->start[1] = 0;			/* in read_SDS() according to the subsample of data to read */
              sds->Nplanes = 1;
              sds->maxplanes = 1;
              break;
     case 3:  if (! sds->planeID) {
                printf("  No valid SDS plane specified for this 3D SDS. First one is assumed.\n");
                sds->planeID = (char **) malloc(1 * sizeof(char *));
                sds->planeID[0] = (char *) malloc(10 * sizeof(char));
                sds->Nplanes = 1;
                strcpy(sds->planeID[0], "1");
              }
              if ( sds->dim_sizes[0] < sds->dim_sizes[1]  &&		/* Band sequential format */
                   sds->dim_sizes[0] < sds->dim_sizes[2] )
                sds->format = BSQ;
              else if ( sds->dim_sizes[2] < sds->dim_sizes[0]  &&	/* Pixel interleave format */
                        sds->dim_sizes[2] < sds->dim_sizes[1] )
                sds->format = PIL;
              else {							/* Line interleave format */
                printf("  Line interleave format not supported.\n");
                sds->format = LIL;
/*
                SDendaccess(sds->id);
                SDend(sds->file_id);
                return -1;
*/
              }
              switch (sds->format) {
                case BSQ:
                  sds->Nl = sds->dim_sizes[1];
                  sds->Np = sds->dim_sizes[2];
                  sds->edges[0] = 1;			/* When rank is 3, start[1] and edges[1] will be set */
                  sds->edges[2] = sds->dim_sizes[2];	/* in read_SDS() according to the subsample of data to read */
                  sds->start[0] = atoi(sds->planeID[0]) - 1;
                  sds->start[2] = 0;
                  sds->maxplanes = sds->dim_sizes[0];
                  break;
                case PIL:
                  sds->Nl = sds->dim_sizes[0];	
                  sds->Np = sds->dim_sizes[1];
                  sds->edges[2] = 1;
                  sds->edges[1] = sds->dim_sizes[1];
                  sds->start[2] = atoi(sds->planeID[0]) - 1;
                  sds->start[1] = 0;
                  sds->maxplanes = sds->dim_sizes[2];
                  break;
                case LIL:
                  sds->Nl = sds->dim_sizes[0];	
                  sds->Np = sds->dim_sizes[2];
                  sds->edges[1] = 1;
                  sds->edges[2] = sds->dim_sizes[2];
                  sds->start[1] = atoi(sds->planeID[0]) - 1;
                  sds->start[2] = 0;
                  sds->maxplanes = sds->dim_sizes[1];
                  break;
              }
/*
              printf("%d lines, %d columns, %d planes\n", sds->Nl, sds->Np, sds->maxplanes);
*/
              if (sds->Nplanes == -1) {		/* This is a first step to support multiple SDS */
                printf("  All SDS planes requested for this 3D SDS.\n");
                sds->Nplanes = sds->maxplanes;
                sds->planeID = (char **) malloc(sds->maxplanes * sizeof(char *));
                for (k=0; k<sds->Nplanes; k++) {
                  sds->planeID[k] = (char *) malloc(10 * sizeof(char));
                  sprintf(sds->planeID[k], "%d", k + 1);
                }
              }
              break;
     case 4:  if (! sds->planeID) {
                printf("  No valid SDS plane no. specified for this 4D SDS. First one is assumed.\n");
                sds->planeID = (char **) malloc(1 * sizeof(char *));
                sds->planeID[0] = (char *) malloc(10 * sizeof(char));
                sds->Nplanes = 1;
                strcpy(sds->planeID[0], "1.1");
              }
              sds->maxplanes = sds->dim_sizes[2] * sds->dim_sizes[3];
              if (sds->Nplanes == -1) {		/* This is a first step to support multiple SDS */
                printf("  All SDS planes requested for this 4D SDS.\n");
                sds->Nplanes = sds->maxplanes;
                sds->planeID = (char **) malloc(sds->maxplanes * sizeof(char *));
                for (j=0; j<sds->dim_sizes[2]; j++)
                  for (k=0; k<sds->dim_sizes[3]; k++) {
                    sds->planeID[j * sds->dim_sizes[3] + k] = (char *) malloc(10 * sizeof(char));
                    sprintf(sds->planeID[j * sds->dim_sizes[3] + k], "%d.%d", j + 1, k + 1);
                  }
              }
              sds->Nl = sds->dim_sizes[0];              /* Pixel interleave format only */
              sds->Np = sds->dim_sizes[1];
              sds->edges[1] = sds->dim_sizes[1];	/* When rank is 4, start[0] and edges[0] will be set */
              sds->edges[2] = 1;                        /* in read_SDS() according to the subsample of data to read */
              sds->edges[3] = 1;
              sds->start[1] = 0;
              sscanf(sds->planeID[0], "%d.%d", &sds->start[2], &sds->start[3]);
              sds->start[2]--;
              sds->start[3]--;
              break;
     default: printf("  Rank of SDS \"%s\" in file %s is %d. File ignored.\n\n", sds->name, sds->filename, sds->rank);
              SDendaccess(sds->id);
              SDend(sds->file_id);
              return -1;
   }

   return 0;

}




int  read_SDS(SDS *sds, int firstline, int nlines)
{
int irow, jcol;
int32 newstart[MAX_VAR_DIMS];

    switch (sds->rank) {			/* Set read parameters according to rank */

      case 2:
        sds->edges[0] = nlines;		/* edges[1] and start[1] */
        sds->start[0] = firstline;	/* were already set in get_SDS_info() */
        newstart[0] = sds->start[0];
        newstart[1] = sds->start[1];
        break;
      case 3:
        switch (sds->format) {
          case BSQ:
            sds->edges[1] = nlines;	/* edges[0], start[0], edges[2] and start[2] */
            sds->start[1] = firstline;	/* were already set in get_SDS_info() */
            newstart[0] = sds->start[0];
            newstart[1] = sds->start[1];
            newstart[2] = sds->start[2];
            break;
          case PIL:
            sds->edges[0] = nlines;
            sds->start[0] = firstline;
            newstart[0] = sds->start[0];
            newstart[1] = sds->start[1];
            newstart[2] = 0;
            sds->edges[2] = sds->dim_sizes[2];
            break;
          case LIL:
            sds->edges[0] = nlines;
            sds->start[0] = firstline;
            newstart[0] = sds->start[0];
            newstart[1] = sds->start[1];
            newstart[2] = sds->start[2];
            break;
        }
        break;
      case 4:
        sds->edges[0] = nlines;		/* edges[1], start[1], edges[2], start[2], edges[3] and start[3] */
        sds->start[0] = firstline;	/* were already set in get_SDS_info() */
        newstart[0] = sds->start[0];
        newstart[1] = sds->start[1];
        newstart[2] = 0;			/* Force reading all planes for each pixel */
        newstart[3] = 0;
        sds->edges[2] = sds->dim_sizes[2];
        sds->edges[3] = sds->dim_sizes[3];
        break;
      default:
        printf("  Don't know how to read data with rank %d.\n", sds->rank);
        return -1;

    }

    if (SDreaddata(sds->id, newstart, NULL, sds->edges, sds->data) == -1) {
      printf("  Can't read SDS \"%s\" in file %s. File ignored.\n\n", sds->name, sds->filename);
      SDendaccess(sds->id);
      SDend(sds->file_id);
      return -1;
    }

    if (sds->rank == 4)				/* Reconstruct the requested plane */
      for (irow=0; irow<nlines; irow++)
        for (jcol=0; jcol<sds->Np; jcol++)
          copy_value(sds->data, sds->data, irow * sds->Np + jcol, ( (irow * sds->Np + jcol) * sds->dim_sizes[2] + sds->start[2] ) * sds->dim_sizes[3] + sds->start[3], sds->num_type);

    if ( sds->rank == 3  &&
         sds->format == PIL )
      for (irow=0; irow<nlines; irow++)
        for (jcol=0; jcol<sds->Np; jcol++)
          copy_value(sds->data, sds->data, irow * sds->Np + jcol, (irow * sds->Np + jcol) * sds->dim_sizes[2] + sds->start[2], sds->num_type);

    return 0;

}




int  create_2D_SDS(SDS *sds, unsigned char access_mode)
{
 
     sds->dim_sizes[0] = sds->Nl;
     sds->dim_sizes[1] = sds->Np;
     sds->rank = 2;
 
     if ( access_mode == REPLACE ) {
         if ( (sds->index = SDnametoindex(sds->file_id, sds->name)) != -1  &&
              (sds->id = SDselect(sds->file_id, sds->index)) != -1 )
           printf("  Replacing existing SDS \"%s\" (may fail if SDS was compressed)...\n", sds->name);
         else {
           fprintf(stderr, "  Cannot find or select SDS \"%s\"\n", sds->name);
           return -1;
         }
     } else if ((sds->id = SDcreate(sds->file_id, sds->name, sds->num_type, sds->rank, sds->dim_sizes)) == -1) {
         fprintf(stderr, "  Cannot create SDS \"%s\"\n", sds->name);
         return -1;
     }

     sds->edges[0] = sds->Nl;
     sds->edges[1] = sds->Np;
     sds->start[0] = 0;
     sds->start[1] = 0;

     sds->fillvalue = (void *) malloc(1 * DFKNTsize(sds->num_type));
     setfillvalue(sds->fillvalue, 0, 1, sds->num_type, 0);
     SDsetfillvalue(sds->id, sds->fillvalue);
     free(sds->fillvalue);
 
     return 0;
 
}




int  write_SDS(SDS sds, int firstline, int nlines, unsigned char gzip)
{
comp_info c_info;

   sds.start[0] = firstline;
   sds.edges[0] = nlines;

   if (firstline == 0  &&  nlines == sds.Nl)
     printf("  Writing SDS \"%s\"...\n", sds.name);
   else
     printf("  Writing SDS \"%s\" (line %d to %d)...\n", sds.name, firstline, firstline + nlines - 1);

   if (gzip) {
     c_info.deflate.level = 6;
     SDsetcompress(sds.id, COMP_CODE_DEFLATE, &c_info);
   }

   if (SDwritedata(sds.id, sds.start, NULL, sds.edges, (VOIDP)sds.data) == -1) {
     fprintf(stderr, "  Cannot write SDS \"%s\"\n", sds.name);
     return -1;
   }
   return 0;

}




void write_grid_SDS(int32 file_id, GRID grid, PROJECTION proj)
{
int k, j;
char sds_name[MAXNAMELENGTH];

   for (k=0; k<NGRIDS; k++) {
     if (grid.points[k]) {
       for (j=0; j<grid.Npoints[k]; j++)
         if  (grid.points[k][2 * j] != VECTORUNDEF  ||
              grid.points[k][2 * j + 1] != VECTORUNDEF ) {
           grid.points[k][2 * j]     = proj.col0 + grid.points[k][2 * j]     / proj.pixsz;	/* Unit in SDS is pixels */
           grid.points[k][2 * j + 1] = proj.row0 + grid.points[k][2 * j + 1] / proj.pixsz;
         }
       sprintf(sds_name, "Grid: %s X,Y (km)", grid.type[k]);
       create_and_write_1D_SDS(grid.points[k], grid.Npoints[k], sds_name, DFNT_FLOAT32, file_id);
     }
   }

}




int create_and_write_1D_SDS(void *data, int32 Nitems, char *sds_name, int32 num_type, int32 file_id)
{
int32 sds_id, rank, dim_sizes[2], start[2], edges[2];

     rank = 2;
     dim_sizes[0] = Nitems;
     dim_sizes[1] = 2;
     edges[0] = Nitems;
     edges[1] = 2;
     start[0] = 0;
     start[1] = 0;

     if ((sds_id = SDcreate(file_id, sds_name, num_type, rank, dim_sizes)) == -1) {
         fprintf(stderr, "  Cannot create SDS \"%s\"\n",  sds_name);
         return -1;
     } else {
         printf("  Writing SDS \"%s\"...\n", sds_name);
         if (SDwritedata(sds_id, start, NULL, edges, (VOIDP)data) == -1) {
           fprintf(stderr, "  Cannot write SDS \"%s\"\n", sds_name);
           return -1;
         }
     }
     return 0;
}




char *get_SDS_ID(char *arg, SDS *sds)
{
char *ptr=NULL, *argptr, string[MAXNAMELENGTH];

   argptr = get_next_string(arg, sds->name);

   /* In case of stdin, scanf does not handle blanks in quoted string */
   if ( sds->name[0] == 34 ||		/* If SDS name starts with a quote (" or '), */
        sds->name[0] == 39 ) {		/* then it should ends with an identical quote */

     while ( (ptr = strchr(sds->name + 1, sds->name[0])) == NULL )	/* Search another quote (same as the opening one) */
       if ( scanf("%s", string + 1) != EOF &&				/* Get next field */
            strlen(sds->name) + strlen(string) + 1 < MAXNAMELENGTH ) {	/* Check string size */
           *string = 32;						/* Add a space to separate fields */
           strcat(sds->name, string);					/* Add next field to SDS name */
       } else {
           printf("SDS name specified with unmatched quote. Ignored.\n\n");
           break;
       }

     if (ptr) strcpy(ptr, ptr + 1);		/* Shift string to skip closing quote */
     strcpy(sds->name, sds->name + 1);		/* Shift string to skip opening quote */

   }

   /* ptr points to right after closing quote, if name quoted, or is NULL otherwise */
   /* If name contains dots, it must be a separator, unless quoted */
   if ( ( ( ptr  &&  (ptr = strchr(ptr, 46)) )  ||	/* Look for dot right after closing quote if name quoted */
          (ptr = strchr(sds->name, 46)) )  &&		/* Look for dot in the whole name otherwise */
        (sds->planeID = (char **) realloc(sds->planeID, (sds->Nplanes + 1) * sizeof(char *)))  &&
        (sds->planeID[0] = (char *) malloc(20 * sizeof(char)))  &&
        sscanf(ptr + 1, "%s", sds->planeID[0]) == 1 )  {	/* Get the SDS plane ID (if any) */
     *ptr = '\0';					/* Replace separator by null to terminate string */
     if (strcmp(sds->planeID[0], "all") == 0)
       sds->Nplanes = -1;
     else
       sds->Nplanes = 1;
   }

   return argptr;

}




void get_step_and_offset(int n1, int n2, float *offset, float *step, unsigned char projtype, unsigned char crosstrack)
{
       /* Map i1 into i2:  i2 = offset + step * i1 */
       if (n1 >= n2) *step = 1. / (int) ((float)n1 / n2 + 0.5);		/* Must add 0.5 to handle 1354/271 */
               else  *step = (int) ((n2 + n1 - 1) / n1);
       if (projtype == SWATH  &&  crosstrack)
         *offset = 0;			/* No nesting in SWATH products */
       else
         *offset = (*step - 1) / 2;
}




int check_continuity(double lon1, double lat1, double lon2, double lat2, double xproj1, double xproj2, PROJPARAM *projparam, double pixsz)
{
double latmin, latmax;

    /* Using xproj1 and xproj2 (instead of just lon and lat) is simpler */
    /* to determine if there is an interruption between the 2 points. */
    /* Also that handles correctly the particular case where 1 point */
    /* is exactly on the interruption. */

    lon1 -= projparam->lon_center;
    lon2 -= projparam->lon_center;
    lat1 -= projparam->lat_center;
    lat2 -= projparam->lat_center;

    switch (projparam->type) {
      case INTERSINUS1: if ( (xproj1 + 40) * (xproj2 + 40) < 0  ||
                             (xproj1 - 40) * (xproj2 - 40) < 0 ) return -1;
                        break;
      case INTERSINUS2: if ( (xproj1 + 36) * (xproj2 + 36) < 0  ||
                             MIN(lat1, lat2) < 0  &&  (xproj1 - 55) * (xproj2 - 55) < 0 ) return -1;
                        break;
      case INTERSINUS3: if ( (int)((180 + xproj1) / (ZONEGAP + 360. / projparam->nzones)) != (int)((180 + xproj2) / (ZONEGAP + 360. / projparam->nzones)) )
                          return -1;
                        break;
      case INTERMOLL:   latmin = MIN(lat1, lat2);
                        latmax = MAX(lat1, lat2);
                        /* 36.012652646284 = 40 * 2 * sqrt(2) / PI */
                        /* 18.006326323142 = 20 * 2 * sqrt(2) / PI */
                        /* 72.025305292568 = 80 * 2 * sqrt(2) / PI */
                        if ( latmax > 0  &&  (xproj1 + 36.012652646284) * (xproj2 + 36.012652646284) < 0  ||
                             latmin < 0  &&  (xproj1 + 18.006326323142) * (xproj2 + 18.006326323142) < 0  ||
                             latmin < 0  &&  (xproj1 - 72.025305292568) * (xproj2 - 72.025305292568) < 0 ) return -1;
                        break;
      case GOODE:       latmin = MIN(lat1, lat2);
                        latmax = MAX(lat1, lat2);
                        if ( latmin < 0  &&  (xproj1 + 100) * (xproj2 + 100) < 0  ||
                             latmax > 0  &&  (xproj1 +  40) * (xproj2 +  40) < 0  ||
                             latmin < 0  &&  (xproj1 +  20) * (xproj2 +  20) < 0  ||
                             latmin < 0  &&  (xproj1 -  80) * (xproj2 -  80) < 0 ) return -1;
                        break;
      case GOODE2:      latmin = MIN(lat1, lat2);
                        latmax = MAX(lat1, lat2);
                        if ( latmax > 0  && (xproj1 +  40) * (xproj2 +  40) < 0  ||
                             latmin < 0  && (xproj1 + 100) * (xproj2 + 100) < 0  ||
                             latmin < 0  && (xproj1 +  20) * (xproj2 +  20) < 0  ||
                             latmin < 0  && (xproj1 -  80) * (xproj2 -  80) < 0  ) return -1;
                        break;
      default: break;
    }

/* Draft version - To be refined ? */
      if ( xproj1 * xproj2 < 0  &&
           fabs(xproj1 - xproj2) > 20 * pixsz ) return -1;	/* Segment longer than 20 pixels */
/*
      if ( xproj1 * xproj2 < 0  ||
           lon1 - lon2 >  180  ||
           lon1 - lon2 < -180 ) return -1;
*/

    return 0;

}




void get_projinfo(PROJINFO *projinfo)
{
int projtype;
char *ptr;

   for (projtype=0; projtype<NUMBEROFPROJ; projtype++) {

     switch (projtype) {

       /* Use strdup() to allocate the interruption lists and to make sure */
       /* they all have a different address and can be modified independently */

       case SWATH:         projinfo[projtype].name = strdup("MODIS swath");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = FALSE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case SINUSOIDAL:    projinfo[projtype].name = strdup("Sinusoidal");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case INTEGERSIN:    projinfo[projtype].name = strdup("Integerized Sinusoidal");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case INTERSINUS1:   projinfo[projtype].name = strdup("Interrupted Sinusoidal USGS");
                           projinfo[projtype].lonsym = FALSE;		/* No global symmetry due to lon offset */
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = strdup("-20,-90,90,  60,-90,90");
                           break;
       case INTERSINUS2:   projinfo[projtype].name = strdup("Interrupted Sinusoidal Variant");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = FALSE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = strdup("-26,-90,90,  65,-90,0");
                           break;
       case INTERSINUS3:   projinfo[projtype].name = strdup("Regular Interrupted Sinusoidal");
                           projinfo[projtype].lonsym = FALSE;		/* No global symmetry due to lon offset */
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = FALSE;		/* No global symmetry due to interzone gap */
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;	/* Interruptions vary with no. of zones */
                           projinfo[projtype].Nintlon = 1;	/* Set Nintlon to identify interrupted proj. */
                           break;
       case PLATECARREE:   projinfo[projtype].name = strdup("Plate Carree");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case ROBINSON:      projinfo[projtype].name = strdup("Robinson");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case MOLLWEIDE:     projinfo[projtype].name = strdup("Mollweide");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case INTERMOLL:     projinfo[projtype].name = strdup("Interrupted Mollweide");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = FALSE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = strdup("-100,-90,0,  -40,0,90,  -20,-90,0,  80,-90,0");
                           break;
       case GOODE:         projinfo[projtype].name = strdup("Goode Homolosine");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = FALSE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = strdup("-100,-90,0,  -40,0,90,  -20,-90,0,  80,-90,0");
                           break;
       case GOODE2:        projinfo[projtype].name = strdup("Goode Homolosine Variant");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = FALSE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = strdup("-100,-90,0,  -40,0,60,  -20,-90,0,  -10,75,90,  80,-90,0");
                           break;
       case PERSPECTIVE:   projinfo[projtype].name = strdup("Vertical Perspective");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;		/* No global symmetry due to lon and lat offset */
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = TRUE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case ORTHOGRAPHIC:  projinfo[projtype].name = strdup("Orthographic");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;		/* No global symmetry due to lon and lat offset */
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = TRUE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case GNOMONIC:      projinfo[projtype].name = strdup("Gnomonic");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;		/* No global symmetry due to lon and lat offset */
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = TRUE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case STEREOGRAPHIC: projinfo[projtype].name = strdup("Stereographic");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;		/* No global symmetry due to lon and lat offset */
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = TRUE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case LAMBERT:       projinfo[projtype].name = strdup("Lambert Azimuthal");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = TRUE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case AZIMEQUI:      projinfo[projtype].name = strdup("Azimuthal Equidistant");
                           projinfo[projtype].lonsym = FALSE;		/* No symmetry if not polar */
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = TRUE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case HAMMER:        projinfo[projtype].name = strdup("Hammer-Aitoff");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case ALBERS:        projinfo[projtype].name = strdup("Albers Conical Equal Area");
                           projinfo[projtype].lonsym = FALSE;
                           projinfo[projtype].latsym = FALSE;
                           projinfo[projtype].xprojsym = FALSE;
                           projinfo[projtype].yprojsym = FALSE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case AITOFF:        projinfo[projtype].name = strdup("Aitoff");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = TRUE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case ECKERTIV:      projinfo[projtype].name = strdup("Eckert IV");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;
       case GEOGRAPHIC:   projinfo[projtype].name = strdup("Geographic");
                           projinfo[projtype].lonsym = TRUE;
                           projinfo[projtype].latsym = TRUE;
                           projinfo[projtype].xprojsym = TRUE;
                           projinfo[projtype].yprojsym = TRUE;
                           projinfo[projtype].azimuthal = FALSE;
                           projinfo[projtype].elliptic = FALSE;
                           projinfo[projtype].intlon = NULL;
                           break;

     }

     if ( ptr = projinfo[projtype].intlon ) {
       while (ptr = strchr(ptr, 32)) strcpy(ptr, ptr + 1);	/* Delete all space characters (if any) */
       projinfo[projtype].Nintlon = 0;		
       ptr = projinfo[projtype].intlon;
       while (ptr = strchr(ptr + 1, 44)) {
         projinfo[projtype].Nintlon++;
         if (projinfo[projtype].Nintlon % 3 == 0) *ptr = 32;
       }
       if ((projinfo[projtype].Nintlon + 1) % 3 != 0) {
         printf("Unable to identify interruptions correctly for the %s projection.\n", projinfo[projtype].name);
         projinfo[projtype].intlon = NULL;	/* Erase interruption list */
       }
       projinfo[projtype].Nintlon = (projinfo[projtype].Nintlon + 1) / 3;
     }

   }

}




int update_meta_field(SDS *projmeta, PROJECTION *proj, PRODUCT input, METADATA metadata)
{
int irow, jcol, gnomoproj;
int32 new_num_type;
double lon0, lon, lat;
double xproj, yproj, xarr[4], yarr[4];
BOUNDINGBOX bbox, estimHbox;

  convert_metadata_string(projmeta, &metadata);

  if ( get_gnomonic_Gring(input, xarr, yarr, &lon0, &gnomoproj) == -1 ) return -1;

  get_new_type(metadata.string, 1, metadata.num_type, &projmeta->min, &projmeta->max, &new_num_type);

  if (convert_datatype(projmeta, new_num_type) == -1)
    return -1;

  if ( proj->domain == GLOBAL  &&		/* Global domain requested */
       ! projmeta->data ) {		/* No metadata processed yet */
    bbox.l2 = proj->Nl_glb - 1;
    bbox.p2 = proj->Np_glb - 1;
    bbox.l1 = 0;
    bbox.p1 = 0;
    resize(proj, &bbox);
  }

  estimHbox.l1 = bbox.l1 = proj->Nl_glb;
  estimHbox.p1 = bbox.p1 = proj->Np_glb;
  estimHbox.l2 = bbox.l2 = -1;
  estimHbox.p2 = bbox.p2 = -1;

proj->row0 = proj->Nl_glb / 2. + 0.5 - 1 - proj->allocbox.l1;
proj->col0 = proj->Np_glb / 2. + 0.5 - 1 - proj->allocbox.p1;

  if ( get_estimbox(*proj, input, &estimHbox) == -1 ) return -1;	/* Estimate bounding box */
  printf("             bounding box dimensions: %dx%d\n",
         estimHbox.p2 - estimHbox.p1 + 1, estimHbox.l2 - estimHbox.l1 + 1);

/*
Draft version - Might be optimized with a safety margin method
The projection is resized to the presumed size, which can be much bigger than the actual size
*/
  if ( estimHbox.l1 < proj->allocbox.l1 ||
       estimHbox.l2 > proj->allocbox.l2 ||
       estimHbox.p1 < proj->allocbox.p1 ||
       estimHbox.p2 > proj->allocbox.p2 )
    resize(proj, &estimHbox);
  else if (! projmeta->data) {	/* A larger domain is already allocated -> some data (but no metadata) has been processed */
    if ( alloc_new_array(&projmeta->data, proj->sds.Nl * proj->sds.Np, projmeta->num_type, NULL) == -1 ) return -1;
    projmeta->Nl = proj->sds.Nl;
    projmeta->Np = proj->sds.Np;
  }

/*
  for (irow=0; irow<projmeta->Nl; irow++) {
    for (jcol=0; jcol<projmeta->Np; jcol++) {
*/
  for (irow=estimHbox.l1-proj->allocbox.l1; irow<estimHbox.l2-proj->allocbox.l1; irow++) {
    for (jcol=estimHbox.p1-proj->allocbox.p1; jcol<estimHbox.p2-proj->allocbox.p1; jcol++) {
      yproj = (proj->row0 - irow) * proj->pixsz;
      xproj = (jcol - proj->col0) * proj->pixsz;
      if ( get_lonlat(xproj, yproj, &lon, &lat, &proj->param) != -1 ) {
        if ( ! inside_Gring(lon, lat, xarr, yarr, lon0, gnomoproj) )		/* Pixel is not inside the Gring */
          continue;								/* Skip to next pixel */
        update_box(&bbox, irow + proj->allocbox.l1, jcol + proj->allocbox.p1);
        set_meta_value(projmeta->data, irow * projmeta->Np + jcol, projmeta->num_type, metadata.string, metadata.num_type);
      }
    }
  }
  if (proj->domain != USERBOX) {
    if (bbox.l1 < proj->bbox.l1) proj->bbox.l1 = bbox.l1;
    if (bbox.l2 > proj->bbox.l2) proj->bbox.l2 = bbox.l2;
    if (bbox.p1 < proj->bbox.p1) proj->bbox.p1 = bbox.p1;
    if (bbox.p2 > proj->bbox.p2) proj->bbox.p2 = bbox.p2;
  }

  return 0;

}




int convert_metadata_string(SDS *projmeta, METADATA *metadata)
{
int time, hh,mm,ss,cc;

  printf("  Found %s = ", metadata->fullname);
  switch (metadata->num_type) {
    case DFNT_UCHAR8:
      printf("%s\n", metadata->string);
      hh = mm = ss = cc = -1;
      if ( sscanf(metadata->string, "%d:%d:%d.%d", &hh, &mm, &ss, &cc) == 4  &&	/* Parse a time and convert string into numeric */
           hh >= 0 && hh <= 23  &&
           mm >= 0 && mm <= 59  &&
           ss >= 0 && ss <= 59  &&
           cc >= 0 && cc <= 99 ) {
          time = hh * 360000 + mm * 6000 + ss * 100 + cc;
          memcpy(metadata->string, &time, sizeof(int));
      } else if (string2lutindex(projmeta, metadata->string) == -1) {	/* String was not found in current LUT - LUT will be appended */
          /* Add new string and 7 characters: 4-digit index + colon + space + newline */
          printf("  Adding \"%s\" to LUT of SDS \"%s\"...\n", metadata->string, projmeta->name);
          if ( check_memory((strlen(projmeta->lut) + strlen(metadata->string) + 7 + 1) * sizeof(char)) == -1 )
            return -1;
          projmeta->lut = realloc(projmeta->lut, (strlen(projmeta->lut) + strlen(metadata->string) + 7 + 1) * sizeof(char));
          sprintf(strchr(projmeta->lut, 0), "%4d: %s\n", projmeta->Nlut, metadata->string);	/* Append LUT with current string */
          *(int32 *)metadata->string = projmeta->Nlut++;	/* Convert string into new index */
      }
      metadata->num_type = DFNT_INT32;
      break;
    case DFNT_INT32:
      printf("%d\n", *(int32   *)metadata->string);
      break;
    case DFNT_FLOAT32:
      printf("%g\n", *(float32 *)metadata->string);
      break;
    default: printf("Metadata type %s not supported.\n", typename[metadata->num_type]);
             return -1;
  }

  return 0;

}




int string2lutindex(SDS *projmeta, char *string)
{
char *ptr1, *ptr2;
int index;

  ptr1 = ptr2 = projmeta->lut;	/* Point to first character (should always be a newline character) */
  while ( ptr1  &&					/* ptr1 points to newline character of previous line in LUT */
          (ptr2 = strchr(ptr1 + 1, 10)) ) {			/* ptr2 points to next newline character, or NULL if not found (should not happen) */
    if ( ptr2 - ptr1 - 7 > 0  &&			/* At least 7 characters in a valid line: newline + 4-digit index + colon + space */
         strncmp(ptr1 + 7, string, ptr2 - ptr1 - 7) == 0  &&	/* Compare string from current line to specified string */
         sscanf(ptr1, "%4d: ", &index) == 1 ) {				/* Get index of corresponding line */
      printf("  \"%s\" found in LUT of SDS \"%s\" (index=%d)\n", string, projmeta->name, index);
      *(int32 *)string = index;		/* Convert string into numeric index */
      break;
    }
    ptr1 = ptr2;		/* Point to newline character at the end of current line in LUT */
  }

  if (! ptr1  ||  ! ptr2)	/* String not found in current LUT */
    return -1;
  else
    return 0;

}




int get_gnomonic_Gring(PRODUCT input, double *xarr, double *yarr, double *lon0, int *gnomoproj)
{
int j;
double lon, lat, tan_lat;

    *gnomoproj = EQUATORIAL;					/* Equatorial projection */
    if (input.gringlat[0] >  45) *gnomoproj = NORTHPOLE;	/* North pole projection */
    if (input.gringlat[0] < -45) *gnomoproj = SOUTHPOLE;	/* South pole projection */

    *lon0 = input.gringlon[0] * M_PI / 180;

    for (j=0; j<4; j++) {
      lon = input.gringlon[j] * M_PI / 180;
      lat = input.gringlat[j] * M_PI / 180;
      tan_lat = tan(lat);
      switch (*gnomoproj) {
        case EQUATORIAL: xarr[j] = tan(lon - *lon0);		/* Equatorial projection */
                         yarr[j] = tan_lat / cos(lon - *lon0);
                         break;
        case SOUTHPOLE:  xarr[j] = - sin(lon) / tan_lat;	/* South pole projection */
                         yarr[j] = - cos(lon) / tan_lat;
                         break;
        case NORTHPOLE:  xarr[j] =   sin(lon) / tan_lat;	/* North pole projection */
                         yarr[j] = - cos(lon) / tan_lat;
                         break;
      }
    }

    for (j=0; j<4; j++) {
      if ( CROSS_PRODUCT( xarr[(j + 1) % 4], yarr[(j + 1) % 4],
                          xarr[j],           yarr[j],
                          xarr[(j + 2) % 4], yarr[(j + 2) % 4] ) < 0 ) {
        printf("The Gring points are not sorted clockwise. File ignored.\n");
        return -1;
      }
    }

    return 0;

}




int inside_Gring(double lon, double lat, double *xarr, double *yarr, double lon0, int gnomoproj)
{
int j;
double xproj, yproj, tan_lat, londiff;
static float epsilon=1e-3;	/* Tolerance on cross-product to compensate for the gap between granules */

   lon *= DEG2RAD;
   lat *= DEG2RAD;

   switch (gnomoproj) {
     case EQUATORIAL:					/* Equatorial projection */
       londiff = lon - lon0;
       while(londiff >  M_PI) londiff -= 2 * M_PI;
       while(londiff < -M_PI) londiff += 2 * M_PI;
       if ( fabs(londiff) > 0.3 * M_PI )		/* Must ignore the solution (lon + PI, -lat) */
         return FALSE;
       xproj = tan(lon - lon0);
       yproj = tan(lat) / cos(lon - lon0);
       break;
     case SOUTHPOLE:					/* South pole projection */
       if (lat >= 0) return FALSE;			/* Must ignore the solution (lon + PI, -lat) */
       tan_lat = tan(lat);
       xproj = - sin(lon) / tan_lat;
       yproj = - cos(lon) / tan_lat;
       break;
     case NORTHPOLE:					/* North pole projection */
       if (lat <= 0) return FALSE;			/* Must ignore the solution (lon + PI, -lat) */
       tan_lat = tan(lat);
       xproj =   sin(lon) / tan_lat;
       yproj = - cos(lon) / tan_lat;
       break;
   }

   for (j=0; j<4; j++)
     if ( CROSS_PRODUCT( xarr[(j + 1) % 4], yarr[(j + 1) % 4],
                         xproj, yproj,
                         xarr[j], yarr[j] ) > epsilon  ||
          CROSS_PRODUCT( xarr[(j + 1) % 4], yarr[(j + 1) % 4],
                         xproj, yproj,
                         xarr[(j + 2) % 4], yarr[(j + 2) % 4] ) < -epsilon ) break;

   if (j < 4) return FALSE;
        else  return TRUE;

}




void set_meta_value(void *data, int idx, int32 data_num_type, void *value, int32 value_num_type)
{
   switch (value_num_type) {
     case DFNT_INT32:
       switch (data_num_type) {
         case DFNT_INT8:    ((int8    *) data) [idx] = *(int32 *)value;  break;
         case DFNT_UINT8:   ((uint8   *) data) [idx] = *(int32 *)value;  break;
         case DFNT_INT16:   ((int16   *) data) [idx] = *(int32 *)value;  break;
         case DFNT_UINT16:  ((uint16  *) data) [idx] = *(int32 *)value;  break;
         case DFNT_INT32:   ((int32   *) data) [idx] = *(int32 *)value;  break;
         case DFNT_UINT32:  ((uint32  *) data) [idx] = *(int32 *)value;  break;
         case DFNT_FLOAT32: ((float32 *) data) [idx] = *(int32 *)value;  break;
         default: break;
       }
       break;
     case DFNT_FLOAT32:
       switch (data_num_type) {
         case DFNT_INT8:    ((int8    *) data) [idx] = *(float32 *)value;  break;
         case DFNT_UINT8:   ((uint8   *) data) [idx] = *(float32 *)value;  break;
         case DFNT_INT16:   ((int16   *) data) [idx] = *(float32 *)value;  break;
         case DFNT_UINT16:  ((uint16  *) data) [idx] = *(float32 *)value;  break;
         case DFNT_INT32:   ((int32   *) data) [idx] = *(float32 *)value;  break;
         case DFNT_UINT32:  ((uint32  *) data) [idx] = *(float32 *)value;  break;
         case DFNT_FLOAT32: ((float32 *) data) [idx] = *(float32 *)value;  break;
         default: break;
       }
       break;
   }
}




int check_memory(unsigned long size)
{
  if ( (double)(mallinfo().arena + size) / 1e6 > MAXMEMORY ) {
         printf("%ld Mbytes already allocated. Cannot allocate %ld Mbytes more.\n",
                 mallinfo().arena / 1000000, size / 1000000);
         printf("Maximum authorized memory (%d Mbytes) is exceeded.\n", MAXMEMORY);
         return -1;
  } else return 0;
}




int init_forward_projection(unsigned char projtype, double height, double stdlat1, double stdlat2)
{
  switch (projtype) {
    case SINUSOIDAL:    break;
    case INTEGERSIN:    isinusint(1.0, 0., 0., 0., isinusparam.nzones);
                        break;
    case INTERSINUS1:   intersinint1(1.0, 0., 0.);
                        break;
    case INTERSINUS2:   intersinint2(1.0, 0., 0.);
                        break;
    case INTERSINUS3:   break;
    case PLATECARREE:   break;
    case MOLLWEIDE:     molwforint  (RAD2DEG, 0., 0., 0.);  break;
    case INTERMOLL:     imolwforint (RAD2DEG, 0., 0.);      break;
    case ROBINSON:      robforint   (RAD2DEG, 0., 0., 0.);  break;
    case GOODE:         goodforint  (RAD2DEG, 0., 0.);
                        goodinvint  (RAD2DEG, 0., 0.);
                        break;
    case GOODE2:        goodforint2 (RAD2DEG, 0., 0.);      break;
    case PERSPECTIVE:   gvnspforint  (RAD2DEG, height / DEG2KM, 0., 0., 0., 0.);  break;
    case ORTHOGRAPHIC:  orthforint  (RAD2DEG, 0., 0., 0., 0.);  break;
    case GNOMONIC:      gnomforint  (RAD2DEG, 0., 0., 0., 0.);  break;
    case STEREOGRAPHIC: sterforint (RAD2DEG, 0., 0., 0., 0.);  break;
    case LAMBERT:       lamazforint(RAD2DEG, 0., 0., 0., 0.);  break;
    case AZIMEQUI:      azimforint(RAD2DEG, 0., 0., 0., 0.);  break;
    case HAMMER:        hamforint (RAD2DEG, 0., 0., 0.);  break;
    case ALBERS:        alberforint (RAD2DEG, RAD2DEG, stdlat1, stdlat2, 0., 0., 0., 0.);  break;
    case ECKERTIV:      eckivforint (RAD2DEG, 0., 0., 0.);  break;
    default: printf("Forward projection %s not implemented.\n", projinfo[projtype].name);
             return -1;
  }
  return 0;
}




float fudge(int iscan, int georowsperscan, PRODUCT *input)
{
int georow;
double latA, lonA, latB, lonB, latC, lonC, AB, AC, BC;
float fudgefactor;

/* fudge will be computed to fill the gap between scans iscan and iscan + 1 */
        georow = (iscan + 1) * georowsperscan - 1;
        if ( read_SDS(&input->lat, georow, 1) == -1  ||		/* Load lat and lon for last row of 1st scan */
             read_SDS(&input->lon, georow, 1) == -1  ||
             (lonA = ((float32 *)input->lon.data)[input->Np_geo / 2]) == -999  ||
             (latA = ((float32 *)input->lat.data)[input->Np_geo / 2]) == -999 ) {
          printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow);
          return -1;
        }
        lonA *= DEG2RAD;
        latA *= DEG2RAD;
        georow++;
        if ( read_SDS(&input->lat, georow, 1) == -1  ||		/* Load lat and lon for 1st row of 2nd scan */
             read_SDS(&input->lon, georow, 1) == -1  ||
             (lonB = ((float32 *)input->lon.data)[input->Np_geo / 2]) == -999  ||
             (latB = ((float32 *)input->lat.data)[input->Np_geo / 2]) == -999 ) {
          printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow);
          return -1;
        }
        lonB *= DEG2RAD;
        latB *= DEG2RAD;
        georow++;
        if ( read_SDS(&input->lat, georow, 1) == -1  ||		/* Load lat and lon for 2nd row of 2nd scan */
             read_SDS(&input->lon, georow, 1) == -1  ||
             (lonC = ((float32 *)input->lon.data)[input->Np_geo / 2]) == -999  ||
             (latC = ((float32 *)input->lat.data)[input->Np_geo / 2]) == -999 ) {
          printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow);
          return -1;
        }
        lonC *= DEG2RAD;
        latC *= DEG2RAD;

        AB = acos( sin(latA) * sin(latB) + cos(latA) * cos(latB) * cos(lonB - lonA) );
        AC = acos( sin(latA) * sin(latC) + cos(latA) * cos(latC) * cos(lonC - lonA) );
        BC = acos( sin(latB) * sin(latC) + cos(latB) * cos(latC) * cos(lonC - lonB) );

        fudgefactor = (AC - 2 * BC) / BC / 2. + 0.05 * georowsperscan / 10.;	/* Add 0.05 for a 1km row (~50m) as a safety margin */
printf("AB %g  AC %g  BC %g  Np_geo %d  gap %g  fudge %g  georowsperscan %d\n", EARTHRADIUS*AB, EARTHRADIUS*AC, EARTHRADIUS*BC, input->Np_geo, EARTHRADIUS*(AC - 2 * BC), fudgefactor, georowsperscan);

        if (fudgefactor > 0) {
          printf("  Interscan gap has been detected right after ");
          switch (iscan % 2) {
            case 0: printf("1st"); break;
            case 1: printf("2nd"); break;
          }
          printf(" scan and every other scan (every %dth row).\n", 2 * georowsperscan);
        }

        if (fudgefactor > 0.5) {
          printf("  Found fudge factor greater than 0.5. Ignored.\n");
          fudgefactor = 0.5;
        }

        return MAX(0, fudgefactor);

}




int get_row_bounds(int irow, PRODUCT *input, double *latrowarr, double *lonrowarr, double *xrowarr, double *yrowarr, PROJPARAM *projparam)
{
static double *lat1=NULL, *lat2=NULL, *lon1=NULL, *lon2=NULL;
static int old_georow1, old_georow2, georow1, georow2, old_irow;
int k, jcol, idx, rowsperscan;
double t;
static unsigned char first_row=TRUE;
float fudgefactor;
unsigned char ok=TRUE;
double xproj, yproj;
static float fudgefactor1, fudgefactor2;
double *lon3, *lat3;
double *xarr, *yarr, *lonarr, *latarr;

  if (input->projparam.type == INTEGERSIN ||
      input->projparam.type == GEOGRAPHIC)
  {
    for (k=0; k<2; k++) {		/* Successively compute top and bottom bound of each row */
      /* xproj and yproj are coordinates of pixel bounds, not center, so do not add 0.5 to jcol or irow */
      yproj =   90 - (input->iv * input->sds.Nl + irow + k) * input->ypixsz;	/* Top or bottom edge */
      /* Adjust yproj to make sure it does not fall in the neighbor pixel, i.e. a different projection zone */
      if (yproj != 0)
        yproj += (k - 0.5) * 1e-4 * input->ypixsz;	/* Add or substract 5e-5 of the pixel size */
      for (jcol=0; jcol<input->sds.Np+1; jcol++) {	/* There are input->sds.Np + 1 corner points to compute */
        xproj = -180 + (input->ih * input->sds.Np + jcol    ) * input->xpixsz;	/* Left edge */
        idx = k * (input->sds.Np + 1) + jcol;
        if ( get_lonlat(xproj, yproj, &lonrowarr[idx], &latrowarr[idx], &input->projparam) == -1  ||
             get_xyproj(lonrowarr[idx], latrowarr[idx], &xrowarr[idx], &yrowarr[idx], projparam) == -1 ) {
          xrowarr[jcol] = PROJUNDEF;		/* It is enough to mark xrowarr as bad */
          continue;				/* Skip to next pixel of the current line */
        }
      }
    }
    return 0;
  }

  rowsperscan = (int)(10 * input->sds.Np / 1354. + 0.5);		/* Number of rows per scan */

  if ( first_row  ||
       irow == MAX(0, input->rowmin) ) {
    first_row = TRUE;
    if ( check_memory(4 * input->Np_geo * sizeof(double)) == -1 ) return -1;
    lat1 = (double *) realloc(lat1, input->Np_geo * sizeof(double));
    lon1 = (double *) realloc(lon1, input->Np_geo * sizeof(double));
    lat2 = (double *) realloc(lat2, input->Np_geo * sizeof(double));
    lon2 = (double *) realloc(lon2, input->Np_geo * sizeof(double));
    georow1 = -1;
    georow2 = -1;
    old_irow = -10;
  }

  if ( FUDGE  &&
       first_row ) {
    if ( get_fudgefactor(input, &fudgefactor1, &fudgefactor2) == -1 )
      ok = FALSE;
  }

  for (k=0; k<2; k++) {		/* Successively compute top (k=0) and bottom (k=1) bound of each row */

    old_georow1 = georow1;
    old_georow2 = georow2;

    get_georows(irow + k - 0.5, input, irow / rowsperscan, &georow1, &georow2, &t);

    if ( first_row  ||
         georow1 != old_georow1 ) {	/* georow1 has changed since previous interpolation */

      if (georow1 == old_georow2) {	/* georow1 is the previous georow2 */
          lat3 = lat1;
          lon3 = lon1;
          lat1 = lat2;			/* Swap (lon1,lat1) and (lon2,lat2) addresses */
          lon1 = lon2;
          lat2 = lat3;
          lon2 = lon3;
      } else {
          input->lat.data = lat1;	/* Load lat1 and lon1 */
          input->lon.data = lon1;
          if ( read_SDS(&input->lat, georow1, 1) == -1  ||
               read_SDS(&input->lon, georow1, 1) == -1 ) {
            printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow1);
            ok = FALSE;
          }
          for (jcol=input->Np_geo-1; jcol>=0; jcol--) {
            lat1[jcol] = ((float32 *)lat1)[jcol];
            lon1[jcol] = ((float32 *)lon1)[jcol];
          }
      }

    }

    if ( ok  &&
         ( first_row  ||
           georow2 != old_georow2 ) ) {	/* georow2 has changed since previous interpolation */

      if (georow2 == old_georow1) {	/* georow2 is the previous georow1 */
/* The following case should be impossible because georow2 == old_georow1 should be equivalent to georow1 == old_georow2 */
          if (georow1 != old_georow2) {	/* Make sure (lon1,lat1) and (lon2,lat2) */
            lat3 = lat2;		/* were not already swapped */
            lon3 = lon2;
            lat2 = lat1;		/* Swap (lon1,lat1) and (lon2,lat2) addresses */
            lon2 = lon1;
            lat1 = lat3;
            lon1 = lon3;
          }
      } else {
          input->lat.data = lat2;	/* Load lat2 and lon2 */
          input->lon.data = lon2;
          if ( read_SDS(&input->lat, georow2, 1) == -1  ||
               read_SDS(&input->lon, georow2, 1) == -1 ) {
            printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow2);
            ok = FALSE;
          }
          for (jcol=input->Np_geo-1; jcol>=0; jcol--) {
            lat2[jcol] = ((float32 *)lat2)[jcol];
            lon2[jcol] = ((float32 *)lon2)[jcol];
          }
      }

    }

    if ( ok  &&
         k == 0  &&			/* Computation of the top bound of the current row */
         irow == old_irow + 1  &&	/* The previous row is the precursor of the current row */
         irow % rowsperscan != 0 ) {	/* irow is not the 1st row of a new scan */
      memcpy(lonrowarr, lonrowarr + input->sds.Np + 1, (input->sds.Np + 1) * sizeof(double));	/* The top bound of the current row is */
      memcpy(latrowarr, latrowarr + input->sds.Np + 1, (input->sds.Np + 1) * sizeof(double));	/* the bottom bound of the previous row */
      memcpy(xrowarr, xrowarr + input->sds.Np + 1, (input->sds.Np + 1) * sizeof(double));
      memcpy(yrowarr, yrowarr + input->sds.Np + 1, (input->sds.Np + 1) * sizeof(double));
      continue;			/* Skip to computation of the bottom bound of the current row */
    }

    if (ok) {
      if (FUDGE) {
        if ( k == 0  &&  irow      % (2 * rowsperscan) == rowsperscan  ||
             k == 1  && (irow + 1) % (2 * rowsperscan) == rowsperscan )
          fudgefactor = fudgefactor1;
        else if ( k == 0  &&  irow      % (2 * rowsperscan) == 0  ||
                  k == 1  && (irow + 1) % (2 * rowsperscan) == 0 )
          fudgefactor = fudgefactor2;
        else
          fudgefactor = 0;
      }
/* When there is a gap between 2 scans (every second scan), the bound of both scans is shifted of a fraction of the */
/* distance between the first 2 or the last 2 geolocation rows to fill the gap */
/* This fraction is specified by the fudge parameter FUDGE */
      if (fudgefactor > 0  &&  FUDGE)
        t += copysign(fudgefactor, t);
      lat3 = (double *) malloc(input->Np_geo * sizeof(double));
      lon3 = (double *) malloc(input->Np_geo * sizeof(double));
      get_virtual_row(lon1, lat1, lon2, lat2, t, input->Np_geo, lon3, lat3);	/* Interpolate between rows */
      lonarr = lonrowarr + k * (input->sds.Np + 1);
      latarr = latrowarr + k * (input->sds.Np + 1);
      /* Compute input->sds.Np + 1 pixel bounds from input->sds.Np centers */
      resample_row_geolocation(BOUNDS, lon3, lat3, input, lonarr, latarr);
      free(lat3);
      free(lon3);
      xarr = xrowarr + k * (input->sds.Np + 1);
      yarr = yrowarr + k * (input->sds.Np + 1);
      for (jcol=0; jcol<input->sds.Np+1; jcol++)
        get_xyproj(lonarr[jcol], latarr[jcol], &xarr[jcol], &yarr[jcol], projparam);
    }

  }

  first_row = FALSE;
  old_irow = irow;

  if ( irow == MIN(input->rowmax, input->sds.Nl - 1)  ||	/* Last row of input granule */
       ! ok ) {				/* Read failed */
    free(lat1);
    free(lat2);
    free(lon1);
    free(lon2);
    lat1 = lat2 = lon1 = lon2 = NULL;
    first_row = TRUE;			/* Re-initialize for next granule */
    if (! ok) return -1;		/* Read failed: must skip the input product */
  }

  return 0;

}




int get_row_pixel_centers(int irow, PRODUCT *input, double *lat, double *lon)
{
static double *lat1=NULL, *lat2=NULL, *lon1=NULL, *lon2=NULL;
double *lat3, *lon3;
static int old_georow1, old_georow2, georow1, georow2;
double t;
static unsigned char first_row=TRUE;
unsigned char ok=TRUE;
int jcol, rowsperscan;

  if ( first_row  ||
       irow == MAX(0, input->rowmin) ) {
    first_row = TRUE;
    if ( check_memory(4 * input->Np_geo * sizeof(double)) == -1 ) return -1;
    lat1 = (double *) realloc(lat1, input->Np_geo * sizeof(double));
    lat2 = (double *) realloc(lat2, input->Np_geo * sizeof(double));
    lon1 = (double *) realloc(lon1, input->Np_geo * sizeof(double));
    lon2 = (double *) realloc(lon2, input->Np_geo * sizeof(double));
    georow1 = -1;
    georow2 = -1;
  }

  old_georow1 = georow1;
  old_georow2 = georow2;

  rowsperscan = (int)(10 * input->sds.Np / 1354. + 0.5);		/* Number of rows per scan */

  get_georows(irow, input, irow / rowsperscan, &georow1, &georow2, &t);

  if ( first_row  ||
       georow1 != old_georow1 ) {	/* georow1 has changed since previous interpolation */

    if (georow1 == old_georow2) {	/* georow1 is the previous georow2 */
        lat3 = lat1;
        lon3 = lon1;
        lat1 = lat2;			/* Swap (lon1,lat1) and (lon2,lat2) addresses */
        lon1 = lon2;
        lat2 = lat3;
        lon2 = lon3;
    } else {
        input->lat.data = lat1;		/* Load lat1 and lon1 */
        input->lon.data = lon1;
        if ( read_SDS(&input->lat, georow1, 1) == -1  ||	/* Load lat1 and lon1 */
             read_SDS(&input->lon, georow1, 1) == -1 ) {
          printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow1);
          ok = FALSE;
        }
        for (jcol=input->Np_geo-1; jcol>=0; jcol--) {
          lat1[jcol] = ((float32 *)lat1)[jcol];
          lon1[jcol] = ((float32 *)lon1)[jcol];
        }
    }

    if (ok) {
      input->lat.data = lat2;	/* Load lat2 and lon2 */
      input->lon.data = lon2;
      if ( read_SDS(&input->lat, georow2, 1) == -1  ||	/* Load lat2 and lon2 */
           read_SDS(&input->lon, georow2, 1) == -1 ) {
        printf("  Unable to read row %d of latitude or longitude. End of product skipped.\n", georow2);
        ok = FALSE;
      }
      for (jcol=input->Np_geo-1; jcol>=0; jcol--) {
        lat2[jcol] = ((float32 *)lat2)[jcol];
        lon2[jcol] = ((float32 *)lon2)[jcol];
      }
    }

  }

  if (ok) {
    lat3 = (double *) malloc(input->Np_geo * sizeof(double));
    lon3 = (double *) malloc(input->Np_geo * sizeof(double));
    get_virtual_row(lon1, lat1, lon2, lat2, t, input->Np_geo, lon3, lat3);
    resample_row_geolocation(CENTERS, lon3, lat3, input, lon, lat);
    free(lat3);
    free(lon3);
  }

  first_row = FALSE;

  if ( irow == MIN(input->rowmax, input->sds.Nl - 1)  ||	/* Last row of input granule */
       ! ok ) {				/* Read failed */
    free(lat1);
    free(lat2);
    free(lon1);
    free(lon2);
    lat1 = lat2 = lon1 = lon2 = NULL;
    first_row = TRUE;			/* Re-initialize for next granule */
    if (! ok) return -1;		/* Read failed: must skip the input product */
  }

  return 0;

}




int update_output_pixel(int irow, int jcol, PROJECTION *proj, void *data, PRODUCT *input, int inputidx, int inputrow, int inputcol, BOUNDINGBOX *bbox, BOUNDINGBOX *estimbox, int margin)
{
int idx, k, j;
unsigned char minsenz=FALSE;

  /* irow and jcol are the row and column in the global projection */
  /* So is bbox - No need to add proj->allocbox.l1 and proj->allocbox.p1 */

  /* Note: Although *input is not modified in this routine, it is significantly */
  /* faster to pass the address of this structure instead of the whole big structure */

  update_box(bbox, irow, jcol);			/* Update actual bounds of output projection */

  if (bbox->l2 == -1 || bbox->p2 == -1)		/* So far, no pixel of this product was projected successfully */
    return -1;					/* Skip to next pixel */

  if ( proj->domain != USERBOX ) {
    estimbox->l1 = MIN(estimbox->l1, bbox->l1);	/* Update current extremes of the input product */
    estimbox->l2 = MAX(estimbox->l2, bbox->l2);	/* Those include the actual extremes and are used instead, */
    estimbox->p1 = MIN(estimbox->p1, bbox->p1);	/* to avoid resizing the projection many times */
    estimbox->p2 = MAX(estimbox->p2, bbox->p2);
    if (estimbox->l1 < 0) estimbox->l1 = 0;
    if (estimbox->p1 < 0) estimbox->p1 = 0;
    if (estimbox->l2 > proj->Nl_glb - 1) estimbox->l2 = proj->Nl_glb - 1;
    if (estimbox->p2 > proj->Np_glb - 1) estimbox->p2 = proj->Np_glb - 1;
  }
  if ( ( estimbox->l1 < proj->allocbox.l1 && (estimbox->l1 -= margin) ||	/* If the new product doesn't fit  */
         estimbox->l2 > proj->allocbox.l2 && (estimbox->l2 += margin) ||	/* in the current subwindow        */
         estimbox->p1 < proj->allocbox.p1 && (estimbox->p1 -= margin) ||	/* (i.e. a resize is needed), then */
         estimbox->p2 > proj->allocbox.p2 && (estimbox->p2 += margin) ) &&	/* add some extra room (10 deg.)   */
       resize(proj, estimbox) == -1 )					/* to anticipate a further resize  */
    return -1;		/* Skip to next pixel */

  if ( irow >= proj->allocbox.l1 && irow <= proj->allocbox.l2  &&
       jcol >= proj->allocbox.p1 && jcol <= proj->allocbox.p2 ) {

    idx = (irow - proj->allocbox.l1) * proj->sds.Np + jcol - proj->allocbox.p1;

    /* printf("inputcol is %d\n",inputcol); */
    if ( options.minsenz  &&			/* Min. senz option requested */
         input->projparam.type == SWATH  &&	/* Input in swath format */
         abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz) <= proj->senz[idx] )	/* Candidate input pixel has lower or equal senz */
      minsenz = TRUE;

    switch (proj->sds.num_type) {
      case DFNT_INT8: 
        if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
          return 0;				/* Pixel must not be updated */
        else {
          ((int8 *)proj->sds.data)[idx] = ((int8 *)data)[inputidx];
          if (options.minsenz)
            proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
        }
        break;
/* Draft version: eventually undef must be handled more properly */
/* Avoid undef value to overwrite valid data */
/* Fire mask: high priority = 8, low priority = 1 */
/* Snow cover: low priority = 3,11 */
      case DFNT_UINT8:
        if (SNOW) {
goto bbb;
            if ( ((uint8 *)proj->sds.data)[idx] != fill.uint8  &&	/* Existing valid value */
                 ( ((uint8 *)data)[inputidx] == 1  ||			/* Candidate pixel is a low priority value */
                   ((uint8 *)data)[inputidx] == 3  ||			/* 1=NoDecision, 3=scan angle exceeded */
                   ((uint8 *)data)[inputidx] == 0  ||			/* 0=L1BMissingData */
                   ((uint8 *)data)[inputidx] == 11 ) )			/* 11=night */
              return 0;
            if ( (options.minsenz  &&  ! minsenz)  &&	/* Min. senz option requested and candidate has not lower senz */
                 ((uint8 *)proj->sds.data)[idx] != fill.uint8  &&
                 ((uint8 *)proj->sds.data)[idx] >= 12 )	/* Existing value is not a low priority value (0, 1, 3 or 11) */
              return 0;				/* Pixel must not be updated */
bbb:
            if ( ((uint8 *)proj->sds.data)[idx] != 200  &&	/* Existing snow and sea ice pixels are untouchable */
                 ((uint8 *)proj->sds.data)[idx] != 237  &&	/* 200=sea ice or snow,  237=seaice by both reflectance and IST */
                 ((uint8 *)proj->sds.data)[idx] != 170  &&	/* 170=seaice by reflectance only */
                 ((uint8 *)proj->sds.data)[idx] != 150  &&	/* 150=seaice by IST only */
                 ((uint8 *)proj->sds.data)[idx] != 100  &&	/* 100=LakeIce */
                 ( ((uint8 *)proj->sds.data)[idx] == fill.uint8  ||
                   ((uint8 *)data)[inputidx] >= 12 ) ) {	/* Only fill value has lower priority than 0, 1, 3 or 11 */
              ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
              if (options.minsenz)
                proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
            }
        }
	else if (SNOW_C6) /* added by SS 05/09 for handling C6 fractional snow cover */
	{
		if ( ((uint8 *)proj->sds.data)[idx] != fill.uint8  &&	/* Existing valid value */
             ( ((uint8 *)data)[inputidx] == 201  ||			/* Candidate pixel is a low priority value */
                   ((uint8 *)data)[inputidx] == 200  ||			/* 0=L1BMissingData */
                   ((uint8 *)data)[inputidx] == 211 ) )			/* 11=night */
              return 0;
		  if ( (options.minsenz  &&  ! minsenz)  &&	/* Min. senz option requested and candidate has not lower senz */
                 ((uint8 *)proj->sds.data)[idx] != fill.uint8  &&
                 (((uint8 *)proj->sds.data)[idx] <= 100  ||
				  ((uint8 *)proj->sds.data)[idx] > 211 ))	/* Existing value is not a low priority value (200, 201, 211) */
              return 0;				/* Pixel must not be updated */
	  if ( (
		((uint8 *)proj->sds.data)[idx] == 0  ||	/* Existing snow and sea ice pixels are untouchable */
	        ((uint8 *)proj->sds.data)[idx] > 100 ||
	        ((uint8 *)proj->sds.data)[idx] == fill.uint8
		) /* &&
		( ((uint8 *)data)[inputidx] <= 100 || ((uint8 *)data)[inputidx] >= 211 ) */ )
      {	/* Only fill value has lower priority than 200, 201, 211 */
	    ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
	    if (options.minsenz)
	      proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
	  }
    }
	else if (FIRE)
	{
            if (! COUNT) {
                if ( abs(8 - ((uint8 *)proj->sds.data)[idx]) <= 1  ||	/* Existing fire pixel - Untouchable */
                     (options.minsenz  &&  ! minsenz) )			/* Min. senz option requested and candidate has not lower senz */
                  return 0;
                else {			/* No fire pixel at this location (or no valid data yet) */
                  ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
                  if (options.minsenz)
                    proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
                  if ( MAGNIFY > 0  &&
                       abs(8 - ((uint8 *)data)[inputidx]) <= 1 )		/* New fire pixel */
                    for (k=-MAGNIFY; k<=MAGNIFY; k++)
                      if ( irow + k >= proj->allocbox.l1  &&
                           irow + k <= proj->allocbox.l2 )
                        for (j=-MAGNIFY; j<=MAGNIFY; j++)
                          if ( jcol + j >= proj->allocbox.p1  &&
                               jcol + j <= proj->allocbox.p2 )
                           ((uint8 *)proj->sds.data)[idx + k * proj->sds.Np + j] = ((uint8 *)data)[inputidx];
                }
            } else {
                if ( ((uint8 *)proj->sds.data)[idx] >= 10 ) {		/* Existing fire count */
                    if ( abs(8 - ((uint8 *)data)[inputidx]) <= 1  &&		/* New fire pixel */
                         ((uint8 *)proj->sds.data)[idx] < UCHAR_MAX )	/* Count not saturated */
                      ((uint8 *)proj->sds.data)[idx]++;			/* Increment count */
                    else		/* No new fire pixel - Existing fire count untouchable */
                      return 0;
                } else {		/* No fire count at this location (or no valid data yet) */
                    if ( abs(8 - ((uint8 *)data)[inputidx]) <= 1 )		/* New fire pixel */
                      ((uint8 *)proj->sds.data)[idx] = 10 + 1;		/* Offset to preserve values of fire mask */
                    else {
                      if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
                        return 0;				/* Pixel must not be updated */
                      ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
                      if (options.minsenz)
                        proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
                    }
                }
                if ( MAGNIFY > 0  &&
                     abs(8 - ((uint8 *)data)[inputidx]) <= 1 )		/* New fire pixel */
                  for (k=-MAGNIFY; k<=MAGNIFY; k++)
                    if ( irow + k >= proj->allocbox.l1  &&
                         irow + k <= proj->allocbox.l2 )
                      for (j=-MAGNIFY; j<=MAGNIFY; j++)
                        if ( jcol + j >= proj->allocbox.p1  &&
                             jcol + j <= proj->allocbox.p2 )
                         ((uint8 *)proj->sds.data)[idx + k * proj->sds.Np + j] = ((uint8 *)data)[inputidx];
            }
        } else if (MOD09CRSQA) {
            if ( ((uint8 *)proj->sds.data)[idx] != fill.uint8  &&	/* Valid data exists at this location */
                 ((uint8 *)proj->sds.data)[idx] >= 128 )		/* This is a fire pixel - Untouchable */
              return 0;						/* Pixel must not be updated */
            else {			/* No fire pixel at this location (or no valid data yet) */
              if ( (options.minsenz  &&  ! minsenz) &&		/* Min. senz option requested and candidate has not lower senz */
                   ((uint8 *)data)[inputidx] < 128 ) 		/* Candidate is not a fire pixel */
                return 0;
              ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
              if (options.minsenz)
                proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
              if ( MAGNIFY > 0  &&
                   ((uint8 *)data)[inputidx] >= 128 )		/* New fire pixel */
                for (k=-MAGNIFY; k<=MAGNIFY; k++)
                  if ( irow + k >= proj->allocbox.l1  &&
                       irow + k <= proj->allocbox.l2 )
                    for (j=-MAGNIFY; j<=MAGNIFY; j++)
                      if ( jcol + j >= proj->allocbox.p1  &&
                           jcol + j <= proj->allocbox.p2 ) {
                        ((uint8 *)proj->sds.data)[idx + k * proj->sds.Np + j] = ((uint8 *)data)[inputidx];
                        if (LGID)
                          copy_value(proj->lgid.data, input->lgid.string, idx + k * proj->sds.Np + j, 0, proj->lgid.num_type);
                      }
            }
        } else if (MOD14CRS) {
            if ( ((uint8 *)proj->sds.data)[idx] != fill.uint8  &&	/* Valid data exists at this location */
                 ((uint8 *)proj->sds.data)[idx] >= 1  &&		/* This is a fire pixel - Untouchable */
                 ((uint8 *)proj->sds.data)[idx] <= 25 )
              return 0;						/* Pixel must not be updated */
            else {			/* No fire pixel at this location (or no valid data yet) */
              if ( (options.minsenz  &&  ! minsenz) &&		/* Min. senz option requested and candidate has not lower senz */
                   ((uint8 *)data)[inputidx] & 31 == 0 ) 	/* Candidate is not a fire pixel */
                return 0;
              ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
              if (options.minsenz)
                proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
              if ( MAGNIFY > 0  &&
                   ((uint8 *)data)[inputidx] & 31 != 0 )	/* New fire pixel */
                for (k=-MAGNIFY; k<=MAGNIFY; k++)
                  if ( irow + k >= proj->allocbox.l1  &&
                       irow + k <= proj->allocbox.l2 )
                    for (j=-MAGNIFY; j<=MAGNIFY; j++)
                      if ( jcol + j >= proj->allocbox.p1  &&
                           jcol + j <= proj->allocbox.p2 ) {
                        ((uint8 *)proj->sds.data)[idx + k * proj->sds.Np + j] = ((uint8 *)data)[inputidx] & 31;
                        if (LGID)
                          copy_value(proj->lgid.data, input->lgid.string, idx + k * proj->sds.Np + j, 0, proj->lgid.num_type);
                      }
            }
        }
        else
        {
            if ( CLEAN  &&			/* CLEAN option requested */
             input->sds.fillvalue  &&		/* Fill value is specified in input product */
             testfillvalue(data, inputidx, input->sds.num_type, 0, input->sds.fillvalue) )	/* Candidate input pixel is fill value */
              return 0;	/* Pixel must not be updated */
            else if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
              return 0;				/* Pixel must not be updated */
            else {
              ((uint8 *)proj->sds.data)[idx] = ((uint8 *)data)[inputidx];
              if (options.minsenz)
                proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
            }
        }
        break;
      case DFNT_INT16:
        if ( CLEAN  &&			/* CLEAN option requested */
             input->sds.fillvalue  &&		/* Fill value is specified in input product */
             testfillvalue(data, inputidx, input->sds.num_type, 0, input->sds.fillvalue) )	/* Candidate input pixel is fill value */
          return 0;	/* Pixel must not be updated */
	else if ( L1B &&   /* Added 07/13: check for solar angle */
		  ((int16 *)input->sz.data)[inputidx] >= SOLZENMAX)
	      return 0;  /* Higher solar angle; pixel should not be updated */
        else {
          if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
            return 0;				/* Pixel must not be updated */
          else {
            ((int16 *)proj->sds.data)[idx] = ((int16 *)data)[inputidx];
            if (options.minsenz)
              proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
          }
        }
        break;
      case DFNT_UINT16:
/* Reject faulty data in L1B */
/*
if ( ((uint16 *)data)[inputidx] <= 32767 )
*/
        if ( CLEAN  &&
             input->sds.fillvalue  &&
             testfillvalue(data, inputidx, input->sds.num_type, 0, input->sds.fillvalue) )
          return 0;	/* Pixel must not be updated */
	else if ( L1B &&   /* Added 07/13: check for solar angle */
		  ((int16 *)input->sz.data)[inputidx] >= SOLZENMAX)
	    return 0;  /* Higher solar angle; pixel should not be updated */
        else {
          if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
            return 0;				/* Pixel must not be updated */
          else {
            ((uint16 *)proj->sds.data)[idx] = ((uint16 *)data)[inputidx];
            if (options.minsenz)
              proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
          }
        }
        break;
      case DFNT_INT32:
        if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
          return 0;				/* Pixel must not be updated */
        else {
          ((int32 *)proj->sds.data)[idx] = ((int32 *)data)[inputidx];
          if (options.minsenz)
            proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
        }
        break;
      case DFNT_UINT32:
        if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
          return 0;				/* Pixel must not be updated */
        else {
          ((uint32 *)proj->sds.data)[idx] = ((uint32 *)data)[inputidx];
          if (options.minsenz)
            proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
        }
        break;
      case DFNT_FLOAT32:
       /* Added for MODWATER by SS-02/18/2020 */
      if (WATER)
      {
              if ( (options.minsenz  &&  ! minsenz)  &&	/* Min. senz option requested and candidate has not lower senz */
               ((float32 *)proj->sds.data)[idx] >= -2.0  &&
               (((float32 *)proj->sds.data)[idx] <= 2.0 ))	/* Existing value is not a low priority value (200, 201, 211) */
            return 0;				/* Pixel must not be updated */
	  if ( (
		((float32 *)proj->sds.data)[idx] < -2.0  ||	/* Existing snow and sea ice pixels are untouchable */
	        ((float32 *)proj->sds.data)[idx] > 2.0 ) )
              {	
                   ((float32 *)proj->sds.data)[idx] = ((float32 *)data)[inputidx];
                   if (options.minsenz)
                     proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
              }
      }
      else
      {
              if (options.minsenz  &&  ! minsenz)	/* Min. senz option requested and candidate has not lower senz */
                return 0;				/* Pixel must not be updated */
              else {
                ((float32 *)proj->sds.data)[idx] = ((float32 *)data)[inputidx];
                if (options.minsenz)
                  proj->senz[idx] = abs((inputcol - input->sds.Np / 2. - 0.5) * input->xpixsz);
              }
      }
        break;
      default: break;
    }

    /* We reach this line if pixel was updated */
    if (LGID)
      copy_value(proj->lgid.data, input->lgid.string, idx, 0, proj->lgid.num_type);

  }

  return 0;

}




int cmp_projections(PROJPARAM projparam1, PROJPARAM projparam2)
{

  if ( projparam1.type == projparam2.type  &&
       projparam1.lon_center == projparam2.lon_center  &&
       projparam1.lat_center == projparam2.lat_center  &&
       projparam1.lon_pole == projparam2.lon_pole  &&
       projparam1.lat_pole == projparam2.lat_pole )

    return 0;

  else

    return -1;

}




void set_proj_dimensions(PROJECTION *proj, int *Nrows, int *Ncols)
{

   if (proj->sds.data) {
       *Nrows = proj->sds.Nl;
       *Ncols = proj->sds.Np;
   } else if (proj->meta.data) {
       *Nrows = proj->meta.Nl;
       *Ncols = proj->meta.Np;
   } else {
       if (proj->domain != USERBOX) {
           *Nrows = proj->Nl_glb;
           *Ncols = proj->Np_glb;
proj->row0 = proj->Nl_glb / 2. + 0.5 - 1;
proj->col0 = proj->Np_glb / 2. + 0.5 - 1;
       } else {
           *Nrows = proj->bbox.l2 - proj->bbox.l1 + 1;
           *Ncols = proj->bbox.p2 - proj->bbox.p1 + 1;
proj->row0 = proj->Nl_glb / 2. + 0.5 - 1 - proj->bbox.l1;
proj->col0 = proj->Np_glb / 2. + 0.5 - 1 - proj->bbox.p1;
       }
   }

}




void copy_value(void *data2, void *data1, unsigned long idx2, unsigned long idx1, int32 num_type)
{
    switch (num_type) {
      case DFNT_INT8:    ((int8    *) data2) [idx2] = ((int8    *) data1) [idx1];  break;
      case DFNT_UINT8:   ((uint8   *) data2) [idx2] = ((uint8   *) data1) [idx1];  break;
      case DFNT_INT16:   ((int16   *) data2) [idx2] = ((int16   *) data1) [idx1];
/*
if (((int16   *) data1) [idx1] == -1 || ((int16   *) data1) [idx1] == -9999 )  ((int16   *) data2) [idx2] = 9999;
*/
                         break;
      case DFNT_UINT16:  ((uint16  *) data2) [idx2] = ((uint16  *) data1) [idx1];  break;
      case DFNT_INT32:   ((int32   *) data2) [idx2] = ((int32   *) data1) [idx1];  break;
      case DFNT_UINT32:  ((uint32  *) data2) [idx2] = ((uint32  *) data1) [idx1];  break;
      case DFNT_FLOAT32: ((float32 *) data2) [idx2] = ((float32 *) data1) [idx1];  break;
      default: break;
    }
}




void project_pixel_centers(PROJECTION *proj, PRODUCT input, BOUNDINGBOX *bbox, BOUNDINGBOX *estimbox, int margin)
{
double xproj, yproj, lon, lat;
int irow, jcol, inputrow, inputcol;
double *latarr, *lonarr;
double xmin, xmax, ymin, ymax;
int inputrowmin, inputrowmax;

  if (proj->domain == USERBOX)
    get_xybounds(proj, &xmin, &xmax, &ymin, &ymax);

  latarr = (double *) malloc(input.sds.Np * sizeof(double));
  lonarr = (double *) malloc(input.sds.Np * sizeof(double));

  inputrowmin = input.sds.Nl;
  inputrowmax = -1;

  for (inputrow=MAX(0,input.rowmin); inputrow<=MIN(input.rowmax,input.sds.Nl-1); inputrow++) {		/* Read product one line at a time */

    if ( read_and_convert_line(&input, inputrow, 1, proj) == -1 ) {
      printf("  Unable to read row %d successfully. End of product skipped.\n", inputrow);
      break;		/* Drop the end of the product */
    }

    if ( input.projparam.type == SWATH  &&		/* Projection must be computed based on geolocation file */
         get_row_pixel_centers(inputrow, &input, latarr, lonarr) == -1 )
      break;

    for (inputcol=0; inputcol<input.sds.Np; inputcol++) {

      if (input.projparam.type == SWATH) {
          if ( get_xyproj(lonarr[inputcol], latarr[inputcol], &xproj, &yproj, &proj->param) == -1 )
            continue;					/* Skip to next pixel of the current line */
      } else {						/* input.projparam.type is INTEGERSIN */
          xproj = -180 + (input.ih * input.sds.Np + inputcol + 0.5) * input.xpixsz;
          yproj =   90 - (input.iv * input.sds.Nl + inputrow + 0.5) * input.ypixsz;
          if (proj->param.type != input.projparam.type) {	/* If input and output projections are different */
            if ( get_lonlat(xproj, yproj, &lon, &lat, &input.projparam) == -1 ||
                 get_xyproj(lon, lat, &xproj, &yproj, &proj->param) == -1 )
              continue;					/* Skip to next pixel of the current line */
          }
      }

      if ( proj->domain == USERBOX  &&
           ( xproj > xmax  ||
             xproj < xmin  ||
             yproj > ymax  ||
             yproj < ymin ) )
        continue;			/* Skip to next pixel if pixel does not fall into the requested domain */

      irow = - yproj / proj->pixsz + proj->row0 + 0.5;
      jcol =   xproj / proj->pixsz + proj->col0 + 0.5;
      update_output_pixel(irow + proj->allocbox.l1, jcol + proj->allocbox.p1, proj, input.sds.data, &input, inputcol, inputrow, inputcol, bbox, estimbox, margin);
      if (inputrowmin > inputrow) inputrowmin = inputrow;
      if (inputrowmax < inputrow) inputrowmax = inputrow;
  
    }

  }

  if (inputrowmin <= inputrowmax)
    printf("  Input rows actually used: %d-%d\n", inputrowmin, inputrowmax);

  free(latarr);
  free(lonarr);

}




void get_virtual_row(double *lon1, double *lat1, double *lon2, double *lat2, double t, int ncols, double *lon3, double *lat3)
{
int geocol;

  for (geocol=0; geocol<ncols; geocol++) {

/* Draft version - UNDEF value must not be hard-coded eventually */
    if ( lon1[geocol] == 0  &&  lat1[geocol] == 0  ||	/* Filter (0,0) because of MOD09CRS problem */
         lon2[geocol] == 0  &&  lat2[geocol] == 0  ||
         lon1[geocol] == -999  ||
         lon2[geocol] == -999  ||
         lat1[geocol] == -999  ||
         lat2[geocol] == -999 ) {
       lon3[geocol] = PROJUNDEF;		/* It is enough to mark lon3 as bad */
       continue;
    }
    lon3[geocol] = (1 - t) * lon1[geocol] + t * lon2[geocol];
    lat3[geocol] = (1 - t) * lat1[geocol] + t * lat2[geocol];
    if (lon1[geocol] - lon2[geocol] > 180) lon3[geocol] += t * 360;	/* Interpolation across the dateline */
    if (lon2[geocol] - lon1[geocol] > 180) lon3[geocol] -= t * 360;	/* Change lon3 but not lon1 and lon2 */
    if (lon3[geocol] >  180) lon3[geocol] -= 360;
    if (lon3[geocol] < -180) lon3[geocol] += 360;

  }

}




void resample_row_geolocation(unsigned char output, double *oldlon, double *oldlat, PRODUCT *input, double *newlon, double *newlat)
{
int jcol, geocol1, geocol2, npoints;
double u, coloffset;

  switch (output) {
    case BOUNDS:
      npoints = input->sds.Np + 1;
      coloffset = -0.5;
      break;
    case CENTERS:
      npoints = input->sds.Np;
      coloffset = 0;
      break;
  }
  for (jcol=0; jcol<npoints; jcol++) {
    geocol1 = floor(input->coloffset + (jcol + coloffset) * input->colstep);
    geocol2 = geocol1 + 1;
    if (geocol1 < 0) {
      geocol1 = 1;
      geocol2 = 0;
    }
    if (geocol2 > input->Np_geo - 1) {
      geocol1 = input->Np_geo - 1;
      geocol2 = input->Np_geo - 2;
    }
    if ( oldlon[geocol1] == PROJUNDEF  ||
         oldlon[geocol2] == PROJUNDEF ) {
      newlon[jcol] = PROJUNDEF;		/* It is enough to mark newlon as bad */
      continue;				/* Skip to next pixel in current row */
    }
    u = (input->coloffset + (jcol + coloffset) * input->colstep - geocol1) / (geocol2 - geocol1);
    newlon[jcol] = (1 - u) * oldlon[geocol1] + u * oldlon[geocol2];
    newlat[jcol] = (1 - u) * oldlat[geocol1] + u * oldlat[geocol2];
    if (oldlon[geocol1] - oldlon[geocol2] >= 180)	/* Interpolation across the dateline */
      newlon[jcol] += u * 360;
    if (oldlon[geocol2] - oldlon[geocol1] >= 180)
      newlon[jcol] -= u * 360;
    if (newlon[jcol] >  180) newlon[jcol] -= 360;
    if (newlon[jcol] < -180) newlon[jcol] += 360;
  }

}




void nearest_pixel_interpol(PROJECTION *proj, PRODUCT input, BOUNDINGBOX *bbox, BOUNDINGBOX *estimbox, int margin)
{
int k, irow, jcol;
double xproj, yproj, *xpixarr, *ypixarr;
double xpixarrmin, xpixarrmax, ypixarrmin, ypixarrmax;
int inputrow, inputcol;
int inputrowmin, inputrowmax;
int offset[6];
BOUNDINGBOX pixelbbox;
double xmin, xmax, ymin, ymax;
double *latarr, *lonarr, *xarr, *yarr;
static float epsilon=1e-8;	/* Tolerance on cross-product */

  latarr = (double *) malloc(2 * (input.sds.Np + 1) * sizeof(double));
  lonarr = (double *) malloc(2 * (input.sds.Np + 1) * sizeof(double));
  xarr   = (double *) malloc(2 * (input.sds.Np + 1) * sizeof(double));
  yarr   = (double *) malloc(2 * (input.sds.Np + 1) * sizeof(double));

  if (proj->domain == USERBOX)
    get_xybounds(proj, &xmin, &xmax, &ymin, &ymax);

  offset[0] = 0;
  offset[1] = 1;
  offset[2] = input.sds.Np + 2;
  offset[3] = input.sds.Np + 1;
  offset[4] = offset[0];	/* Allow to use k + 1 instead of (k + 1) % 4 */
  offset[5] = offset[1];	/* Allow to use k + 2 instead of (k + 2) % 4 */

  inputrowmin = input.sds.Nl;
  inputrowmax = -1;

  for (inputrow=MAX(0,input.rowmin); inputrow<=MIN(input.rowmax,input.sds.Nl-1); inputrow++) {		/* Read product one line at a time */

    if ( read_and_convert_line(&input, inputrow, 1, proj) == -1 ) {
      printf("  Unable to read row %d successfully. End of product skipped.\n", inputrow);
      break;		/* Drop the end of the product */
    }

    if ( get_row_bounds(inputrow, &input, latarr, lonarr, xarr, yarr, &proj->param) == -1 )
      break;					/* Skip end of product */

    for (inputcol=0; inputcol<input.sds.Np; inputcol++) {

      xpixarr = xarr + inputcol;
      ypixarr = yarr + inputcol;
     
      for (k=0; k<4; k++)
        if (xpixarr[offset[k]] == PROJUNDEF)
          break;
      if (k < 4) continue;	/* Skip to next pixel of current line */

      xpixarrmin = ypixarrmin =  DBL_MAX;
      xpixarrmax = ypixarrmax = -DBL_MAX;
      if ( proj->domain == USERBOX ) {
        for (k=0; k<4; k++) {
          if (xpixarr[offset[k]] < xpixarrmin) xpixarrmin = xpixarr[offset[k]];
          if (xpixarr[offset[k]] > xpixarrmax) xpixarrmax = xpixarr[offset[k]];
          if (ypixarr[offset[k]] < ypixarrmin) ypixarrmin = ypixarr[offset[k]];
          if (ypixarr[offset[k]] > ypixarrmax) ypixarrmax = ypixarr[offset[k]];
        }
        if ( xpixarrmin > xmax  ||
             xpixarrmax < xmin  ||
             ypixarrmin > ymax  ||
             ypixarrmax < ymin )
          continue;		/* Skip to next pixel if pixel bbox does not intersect the requested domain */
      }

      for (k=0; k<4; k++)	/* Need to check all 4 angles to reject flat or clockwise or butterfly quadrilateral */
        if ( CROSS_PRODUCT( xpixarr[offset[k + 1]], ypixarr[offset[k + 1]],
                            xpixarr[offset[k]],     ypixarr[offset[k]],
                            xpixarr[offset[k + 2]], ypixarr[offset[k + 2]] ) < 0 )
          break;
      if (k < 4) continue;

      for (k=0; k<4; k++)
        if ( check_continuity( lonarr[inputcol + offset[k]],     latarr[inputcol + offset[k]],
                               lonarr[inputcol + offset[k + 1]], latarr[inputcol + offset[k + 1]],
                               xpixarr[offset[k]], xpixarr[offset[k + 1]], &proj->param, proj->pixsz) == -1 )
          break;
      if (k < 4) continue;

      if (xpixarrmax < xpixarrmin)		/* Pixel bounding box is not computed yet (i.e. no USERBOX option) */
        for (k=0; k<4; k++) {
          if (xpixarr[offset[k]] < xpixarrmin) xpixarrmin = xpixarr[offset[k]];
          if (xpixarr[offset[k]] > xpixarrmax) xpixarrmax = xpixarr[offset[k]];
          if (ypixarr[offset[k]] < ypixarrmin) ypixarrmin = ypixarr[offset[k]];
          if (ypixarr[offset[k]] > ypixarrmax) ypixarrmax = ypixarr[offset[k]];
        }
      pixelbbox.l1 = - ypixarrmax / proj->pixsz + proj->row0 + proj->allocbox.l1 + 0.5;
      pixelbbox.l2 = - ypixarrmin / proj->pixsz + proj->row0 + proj->allocbox.l1 + 0.5;
      pixelbbox.p1 =   xpixarrmin / proj->pixsz + proj->col0 + proj->allocbox.p1 + 0.5;
      pixelbbox.p2 =   xpixarrmax / proj->pixsz + proj->col0 + proj->allocbox.p1 + 0.5;

/* pixelbbox must be computed with respect to global projection because of this loop */
/* projection may be resized in the middle of this loop */
      for (irow=pixelbbox.l1; irow<=pixelbbox.l2; irow++) {
        yproj = ((proj->Nl_glb / 2. - 0.5) - irow) * proj->pixsz;
        for (jcol=pixelbbox.p1; jcol<=pixelbbox.p2; jcol++) {
          xproj = (jcol - (proj->Np_glb / 2. - 0.5)) * proj->pixsz;
          for (k=0; k<3; k++)		/* If point is inside 3 angles it must be inside the 4th one too */
            if ( CROSS_PRODUCT( xpixarr[offset[k + 1]], ypixarr[offset[k + 1]],
                                xproj, yproj,
                                xpixarr[offset[k]], ypixarr[offset[k]] ) > epsilon  ||
                 CROSS_PRODUCT( xpixarr[offset[k + 1]], ypixarr[offset[k + 1]],
                                xproj, yproj,
                                xpixarr[offset[k + 2]], ypixarr[offset[k + 2]] ) < -epsilon )
              break;
          if (k == 3) {
            update_output_pixel(irow, jcol, proj, input.sds.data, &input, inputcol, inputrow, inputcol, bbox, estimbox, margin);
            if (inputrowmin > inputrow) inputrowmin = inputrow;
            if (inputrowmax < inputrow) inputrowmax = inputrow;
          }
        }
      }
    }

  }

  if (inputrowmin <= inputrowmax)
    printf("  Input rows actually used: %d-%d\n", inputrowmin, inputrowmax);

  free(latarr);
  free(lonarr);
  free(xarr);
  free(yarr);

}




void get_xybounds(PROJECTION *proj, double *xmin, double *xmax, double *ymin, double *ymax)
{

/*
  *xmin = (proj->bbox.p1 - proj->col0 - 0.5) * proj->pixsz;
  *xmax = (proj->bbox.p2 - proj->col0 + 0.5) * proj->pixsz;
  *ymin = (proj->row0 - proj->bbox.l2 - 0.5) * proj->pixsz;
  *ymax = (proj->row0 - proj->bbox.l1 + 0.5) * proj->pixsz;
*/
  *xmin = (proj->bbox.p1 - (proj->Np_glb / 2. + 0.5 - 1) - 0.5) * proj->pixsz;
  *xmax = (proj->bbox.p2 - (proj->Np_glb / 2. + 0.5 - 1) + 0.5) * proj->pixsz;
  *ymin = ((proj->Nl_glb / 2. + 0.5 - 1) - proj->bbox.l2 - 0.5) * proj->pixsz;
  *ymax = ((proj->Nl_glb / 2. + 0.5 - 1) - proj->bbox.l1 + 0.5) * proj->pixsz;

}




void get_georows(double fractrow, PRODUCT *input, int iscan, int *georow1, int *georow2, double *t)
{
int georowsperscan;

  georowsperscan = (int)(10 * input->Np_geo / 1354. + 0.5);	/* Number of geolocation rows per scan */

  *georow1 = floor(input->rowoffset + fractrow * input->rowstep);
  *georow2 = *georow1 + 1;
  if (*georow1 < iscan * georowsperscan) {
    *georow1 = iscan * georowsperscan + 1;
    *georow2 = iscan * georowsperscan;
  }
  if (*georow2 > (iscan + 1) * georowsperscan - 1) {
    *georow1 = (iscan + 1) * georowsperscan - 1;
    *georow2 = (iscan + 1) * georowsperscan - 2;
  }

  *t = (input->rowoffset + fractrow * input->rowstep - *georow1) / (*georow2 - *georow1);

}




int get_fudgefactor(PRODUCT *input, float *fudgefactor1, float *fudgefactor2)
{
int georowsperscan;

  if (! FUDGE) {
    *fudgefactor1 = *fudgefactor2 = 0;
    return 0;
  }

  georowsperscan = (int)(10 * input->Np_geo / 1354. + 0.5);	/* Number of geolocation rows per scan */

  if (georowsperscan < 2)
    printf("  Cannot detect interscan gap with less than 2 rows per scan.\n");
  else {
    input->lat.data = (float32 *) malloc(input->Np_geo * sizeof(float32));
    input->lon.data = (float32 *) malloc(input->Np_geo * sizeof(float32));
    if ( (*fudgefactor1 = fudge(0, georowsperscan, input)) < 0  ||
         (*fudgefactor2 = fudge(1, georowsperscan, input)) < 0 ) {
      printf("  Unable to detect where the interscan gap is. Gap will not be filled.\n");
      if (*fudgefactor1 < 0) *fudgefactor1 = 0;
      if (*fudgefactor2 < 0) *fudgefactor2 = 0;
/*
      free(input->lon.data);
      free(input->lat.data);
      return -1;
*/
    }
    free(input->lon.data);
    free(input->lat.data);
  }
  if ( FUDGEFACTOR1 >= 0 ) {
    printf("  Fudge factor between 1st and 2nd scan forced to %g\n", FUDGEFACTOR1);
    *fudgefactor1 = FUDGEFACTOR1;
  }
  if ( FUDGEFACTOR2 >= 0 ) {
    printf("  Fudge factor between 2nd and 3rd scan forced to %g\n", FUDGEFACTOR2);
    *fudgefactor2 = FUDGEFACTOR2;
  }

  return 0;

}




int resize_data(void **data, int newNl, int newNp, int curNl, int curNp, int Hoffset, int Voffset, int32 num_type, unsigned char subset, void *fillvalue)
{
void *buffer;
int irow, sizeoftype;

  /* Either current subwindow has to be expanded and then Hoffset>=0 and Voffset>=0 */
  /* Or current subwindow has to be shrunk and then Hoffset<=0 and Voffset<=0 */
/*
  if ( Hoffset * Voffset < 0 ) {
    fprintf(stderr, "Unexpected offset combination in resize_data() routine.\n");
    return -1;
  }
*/

  buffer = NULL;
  if ( alloc_new_array(&buffer, newNl * newNp, num_type, fillvalue) == -1 )	/* Allocate the new subwindow */
    return -1;

  if (*data) {			/* Only if data already exits */
    sizeoftype = DFKNTsize(num_type);
    if (! subset)			/* Expand current subwindow */
      for (irow=0; irow<curNl; irow++)
        /* Copy the old subwindow (curNl * curNp) into the new one (newNl * newNp) */
        /* data and buffer are now of the same type, then memcpy can be used */
        /* Use generic char type and sizeoftype in memcpy, instead of explicit actual type for all possible cases */
        memcpy((char *) buffer + ((irow + Voffset) * newNp + Hoffset) * sizeoftype,
               (char *) *data + irow * curNp * sizeoftype,
               curNp * sizeoftype);
    else					/* Subset current subwindow - Hoffset<=0 and Voffset<=0 */
      for (irow=0; irow<newNl; irow++)
        memcpy((char *) buffer + irow * newNp * sizeoftype,
               (char *) *data + ((irow - Voffset) * curNp - Hoffset) * sizeoftype,
               newNp * sizeoftype);

    free(*data);			/* Free the old subwindow */
printf("arena %.1f Mbytes", (double) mallinfo().arena / 1e6);
printf(" (%.1f Mbytes freed)\n", curNl * curNp * sizeoftype/1e6);
  }

  *data = buffer;			/* Replace the old subwindow with the new one */

  return 0;

}
